// Copyright (C) 1997 - 1999 Jan Kreft <Kreft@cardiff.ac.uk>

#import "RandSource.h"
#import "FeedQuad.h"
#import "GeckoControlSwarm.h"
#import "Stat.h"
#import <stdlib.h>
#import <math.h>
#import <objectbase.h>
#import <simtools.h>
#import "Animal.h"

@implementation Animal

Animal* recycleAnimals = 0;

-recycle
{
    [species tallyDeath: self];

    hook = recycleAnimals;
    recycleAnimals = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    Animal* togo;
    while (recycleAnimals) {
        togo = recycleAnimals;
        recycleAnimals = recycleAnimals->hook;
        [togo->energy drop];
        [togo drop];
    }
    return self;
}

-reuse
{
    Animal* result = recycleAnimals;
    recycleAnimals = recycleAnimals->hook;
    result->hook    = nil;
    return result;
}

-init
{
    [energy zero];
    wannax 	= 0.0;
    wannay 	= 0.0;
    birthday 	= [geckoControl getCurrentTime];
    kidcount 	= 0;
    firstbornon = 0;
    atecount    = 0;
    scratchpad 	= 0;
    scaleradius = 1.0;	
    scalev      = 1.0/M_PI;	
    
    return self;
}

-allocEnergy
{
    energy = [Metab create];
    return self;
}

+create 
{
    id result;
    
    if (recycleAnimals) {
        result = [recycleAnimals reuse];
    } else {
        result = [Animal createBegin: globalZone];
        [result allocEnergy];
    }
    [result init];
    
    return result;					
}

-copymom
    // create an agent matching mom in all heritable respects
{
    Animal* baby;
    Quad* newQuad;

    baby = [Animal create];
    baby->species   = species;
    [energy copyTo: baby->energy];
    baby->color     = color;

    [species incrementCensus];
    [species tallyBirth: self];
    
    baby->breedsize   = breedsize;
    baby->kidsize     = kidsize;
    baby->minsize     = minsize;
    baby->scaleradius = scaleradius;
    baby->scalev      = scalev;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
}

-addSpeciesParams
{
    GRate rate;
    
    rate = [energy getTaxRate: [energy getIam]];
    [species addParam: "TaxRate" Vary: 0 DValue: rate 
          Message: M(setTaxRate:) Owner: self];

    rate = [energy getTaxExponent: [energy getIam]];
    [species addParam: "TaxExp" Vary: 0 DValue: rate
          Message: M(setTaxExp:) Owner: self];

    rate = [energy getConvertFrom: [energy getIeat] To: [energy getIam]];
    [species addParam: "Convert" Vary: 0 DValue: rate 
          Message: M(setConvertRate:) Owner: self];

    [species addParam: "MinSize" Vary: 0 DValue: minsize 
          Message: M(setMinSize:) Owner: self];

    [species addParam: "KidSize" Vary: 0 DValue: kidsize 
          Message: M(setKidSize:) Owner: self
          Offset: [species getParam: "MinSize"]];

    [species addParam: "BreedSize" Vary: 0 DValue: breedsize  
          Message: M(setBreedSize:) Owner: self 
          Offset: [species getParam: "KidSize"]];
   
   [species addParam: "ScaleR" Vary: 0 DValue: scaleradius 
          Message: M(setScaleR:) Owner: self];
   [species addParam: "ScaleV" Vary: 0 DValue: scalev 
          Message: M(setScaleV:) Owner: self];

    return self;   
}

-liveGrandma: (Species*) s
{
    [energy addConvertFrom: sunwater To: carbos Rate: .1];
    [energy addTax: carbos Rate: .0480 Exponent: .75];
    breedsize = 11.13;  // orig 11
    kidsize = 2.95;
    minsize = 3.0;
    scaleradius = 1.0;
    scalev      = 1.0/M_PI; // back-compatibility default
    
    species = s;
    [self setGColor: [species getGColor]];
    [self addSpeciesParams];
    
    return self;
}

-resetBreedSize: (GCoord) bs KidSize: (GCoord) ks MinSize: (GCoord) ms
      TaxRate: (GRate) tr TaxExp: (GRate) te
    // although this works on any Animal, its purpose is resetting grandma
{
    [energy replaceTax: [energy getIam] Rate: tr Exponent: te];
    breedsize = bs;
    kidsize = ks;
    minsize = ms;
    return self;    
}

-resetConvertFrom: (ResType) rfrom To: (ResType) rto Rate: (GRate) newrate
    // although this works on any Animal, its purpose is resetting grandma
{
    [energy setConvertFrom: rfrom To: rto Rate: newrate];
    return self;
}

-setBreedSize: (GCoord) bs 
{
    breedsize = bs; 
    return self;
}
-setKidSize: (GCoord) ks {kidsize = ks; return self;}
-setMinSize: (GCoord) ms { minsize = ms; return self;}
-setScaleR: (GCoord) sr { scaleradius = sr; return self;}
-setScaleV: (GCoord) v { scalev = v; return self;}

-setConvertRate: (GRate) nr
    // cheats rather than parse what was there before.
{
    [energy setConvertFrom: sunwater To: carbos Rate: nr];
    return self;
}

// NOTE:  this isn't right--only works if getIam is valid.
-setTaxRate: (GRate) tr 
{ 
    ResType me = [energy getIam];
    [energy replaceTax: me
            Rate: tr 
            Exponent: [energy getTaxExponent: me]];
    return self;
}

-setTaxExp: (GRate) te 
{
    ResType me = [energy getIam];
    [energy replaceTax: me
            Rate: [energy getTaxRate: me] 
            Exponent: te];
    return self;
}


-setX: (GCoord) cx Y: (GCoord) cy Radius: (GCoord) r
{
    x = cx;
    y = cy;
    radius = r;
    [energy setStores: [energy getIam] 
            Amount: radius2volume(r, scaleradius, scalev)];
    [self getMcr: myquad->mcr];

    return self;
}

-getX: (GCoord*) cx Y: (GCoord*) cy Radius: (GCoord*) r
{ 
    *cx = x; 
    *cy = y; 
    *r  = radius; 
    return self; 
}

-getX: (GCoord*) cx Y: (GCoord*) cy
{ *cx = x; *cy = y; return self; }

-(GCoord) getX {return x;}
-(GCoord) getY {return y;}
-(GCoord) getGX {return x;}
-(GCoord) getGY {return y;}

-setMetab: (Metab*) m { energy = m; return self; }

-setGColor: (int) colorid { color = colorid; return self; }
	
-(const char*) getInstanceName { return [species getParamName]; } 
-(char*) getName { return [species getParamName]; } 

-(int) getGColor { return color; }	

-setQuad: (Quad*) q { myquad = q; return self; }

-(GCoord) getRadius { return radius; }
-(GCoord) getBreedRadius { return breedsize; }
-(GCoord) getMinRadius { return minsize; }
-(GCoord) getKidRadius { return kidsize; }
-(GCoord) getScaleR { return scaleradius; }
-(GCoord) getScaleV { return scalev; }
-(Species*) getSpecies { return species; }
	
// stat support--eminently overrideable

-statsFinalize { return self; }

-(int) getAgeBucket { 
printf("Animal-getagebucket\n");
    return ([geckoControl getCurrentTime] - birthday) / [geckoControl getStatCycle];
}
-(int) getDefaultAgeCount { return 5; }
-(int) getStageBucket
{
    if (kidcount > 0)
        return 2;
    if (atecount < 3)
        return 0;
    return 1;
}
-(int) getDefaultStages { return 3;}
-(int) getSizeBucket {
    ResQuant v, smallv;
    GCoord smallr, mediumr;

    smallv = radius2volume([self getMinRadius], scaleradius, scalev);
    smallv += radius2volume([self getKidRadius], scaleradius, scalev);
    smallr = volume2radius(smallv, scaleradius, scalev);
    
    v = radius2volume([self getBreedRadius], scaleradius, scalev);
    v -= smallv;
    mediumr = volume2radius(v, scaleradius, scalev);
    
    if (radius <= smallr)
        return 0;
    if (radius <= mediumr)
        return 1;
    return 2;
}
-(int) getDefaultSizes { return 3;}

-calcSpecific: (id) specStats
    // calls BucketDistrib calc by default.
{
    [specStats calc];
    return self;
}
-(BOOL) observeSpecific: (id) specStats
    // false--take default.
{ return 0; }

-(int) getSpecificNumBuckets { return 1;}
-(int) getSpecificBucket { return 0;}
-(int) getSpecificCounter: (int) anEvent { return -1;}
-(int) getSpecificSum: (int) anEvent { return -1;}
-(double) getSpecificAmount: (int) anEvent { return 0;}
-(int) getSpecificCounts { return 1;}
-(int) getSpecificSums { return 1;}
-(int) getSpecificDeriveds { return 1;}

-(GCoord) getMoveRadius { return ( 2.0 * (radius - lastr) );  }
    // for a plant, can move by however much I last grew by
    // the 2.0 factor because grow diameter by 2.0 * growradius

-(GCoord) getChildRadius
    // distance at which to place kid
{
    return( (radius+kidsize) * exp([randSource evenRand]) );
}

-(GCoord) getPerceiveRadius { return radius; }

-getDirectionX: (GCoord*) dx Y: (GCoord*) dy
    // convert wannas to unit vector pointing same direction
{
    ptToward(x, y, x + wannax, y + wannay, 1.0, dx, dy);
    *dx -= x;
    *dy -= y;
    wannax = *dx;
    wannay = *dy;
    return self;
}

-(double) getDirectionAngle
    // convert wannas to unit vector pointing same direction
{
    ResQuant wnorm;
    double angle = M_PI * 2.0 * [randSource evenRand];
    
    wnorm = sqrt(wannax * wannax + wannay * wannay); // hypotenuse...
    if (wnorm > 0.0) {
        wnorm = fabs(wannax / wnorm);
        if (wnorm > 1.0) 
            wnorm = 1.0;
        angle = acos(wnorm);// first quadrant angle...
        if (wannax < 0.0) 
            angle = M_PI - angle;
        if (wannay < 0.0)
            angle = (2.0 * M_PI) - angle;
        wannax = cos(angle);
        wannay = sin(angle);
    } 
    return angle;
}

-setDirectionX: (GCoord) dx Y: (GCoord) dy
{
    wannax = dx; wannay = dy; return self;
}

-addDirectionX: (GCoord) dx Y: (GCoord) dy Weight: (double) weight
    // convert dx, dy to unit vector before weighing in
{
    GCoord ux, uy;
    ptToward(x, y, x+dx, y+dy, 1.0, &ux, &uy);
    ux -= x;
    uy -= y;
    wannax += ux * weight; 
    wannay += uy * weight; 
    return self;
}

-setScratchpad: (GArea) s { scratchpad = s; return self; }

-(GArea) getScratchpad { return scratchpad; }

-getQuad {return myquad;}		

-(GArea) getArea { return (M_PI * radius * radius); }		

-(BOOL) competesWith: (Animal*) other
    // if we eat the same thing, we compete.
{
    ResType meat, sheat; // me eat, she eat...
    ResType i;

    meat = [energy getIeat];
    sheat = [other->energy getIeat];
    if (meat) {
        if (meat == sheat) 
            return 1;
        else
            return 0;
    } else {
        for (i=nullres+1; i <= NumRes; i++) {
            if ( [energy doIeat: i] && [other->energy doIeat: i])
                return 1;
        }
    }
    return 0;
}	

-(BOOL) attacks: (Animal*) other
    // if I eat what she converts food into, I want to attack her.
{
    ResType meat, sheis; // me eat, she is...
    ResType i;

    meat = [energy getIeat];
    sheis = [other->energy getIam];
    
    if ((meat) && (sheis)) {
        if (meat == sheis) 
            return 1;
        else
            return 0;
    } else {
        for (i=nullres+1; i <= NumRes; i++) {
            if ( [energy doIeat: i] && [other->energy amI: i])
                return 1;
        }
    }
    return 0;
}	

-(BOOL) eats: (ResType) res
{
    if ([energy doIeat: res])
        return 1;
    else
        return 0;
}	

-(BOOL) eatList: (ResList*) list
    // same as attack, but for landscape quad
{
    ResType meat; // me eat
    ResType i;

    meat = [energy getIeat];
    
    if (meat) {
        if (list->list[meat] > 0.0)
            return 1;
        else
            return 0;
    } else {
        for (i=1; i <= NumRes; i++) {
            if ( [energy doIeat: i] && (list->list[i] > 0.0))
                return 1;
        }
    }
    return 0;
}	

-(BOOL) makes: (ResType) res
{
    if ([energy amI: res])
        return 1;
    else
        return 0;
}	

-(BOOL) hiddenExposed: (Mcr*) exposed 
{
    [exposed setMcr: myquad->mcr]; 
    return 0;
}

-(void) getMcr: (Mcr*) m
{
    m->ux = x + radius;
    m->uy = y + radius;
    m->lx = x - radius;
    m->ly = y - radius;
}

-(ResQuant) getVolume
{
    return [energy getStores: [energy getIam]];
}

-feedRes: (ResType) res Amount: (ResQuant) dinner
{
    [energy feedRes: res Amount: dinner];
    if (dinner > 0.0) 
        atecount++;
    return self;
}

-stealRes: (ResType) res Amount: (ResQuant) dinner
{
    ResQuant stores = [energy getStores: res];
    stores -= dinner;
    [energy setStores: res Amount: stores];
    
    [species tallyLossSteal: self Amount: dinner];
    
    return self;
}

-(ResQuant) stealAllRes: (ResType) res
{
    ResQuant stores = [energy getStores: res];
    [energy setStores: res Amount: 0];
    return stores;
}

-metabolize
{	
    lastr = radius; // save off what size we were...will reset at grow step
    [energy metabolize];
    return self;
}

-resetRadius
    // set size based on current stores
{	
    ResType Iam, i;
    ResQuant vol;

    Iam = [energy getIam];
    if (Iam) {
        radius = volume2radius([energy getStores: Iam], scaleradius, scalev);
    } else {
        // well, we have no -good- definition.  So let's add 'em up....
        vol = 0;
        for (i = nullres+1; i <=NumRes; i++) vol += [energy getStores: i];
        radius = volume2radius(vol, scaleradius, scalev);
    }
    return self;
}
	
-paytax
{
    ResQuant netgrowth;
    
    LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];
    ResList* taxes = [energy taxes];
    [self resetRadius];
    
    netgrowth = [energy getStores: [energy getIam]];	// current
    netgrowth -= radius2volume(lastr, scaleradius, scalev);	// last
    [species tallyGrowth: self Amount: netgrowth];
    
    [landRoot payTaxX: x Y: y List: taxes];

    [taxes drop];

    if (radius < minsize) {
        [self die]; 
        return nil;
    } else {
        return self;
    }       
}

-die
{
    LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];
    ResList* taxes = [energy getStores];
    
    [species tallyLossStarve: self Amount: taxes->list[[energy getIam]]];
    
    [landRoot payTaxX: x Y: y List: taxes];
    // do NOT drop taxes, it's a pointer, not a copy

    // note that recycle, called by delData, does the census
    [myquad delData];

    return nil;
}

-(ResList*) kidshare
    // find initial stores for kid--can be no general def'n if multi-am....
{
    ResList* share;
    ResQuant vol;

    share = [[ResList create] zero];
    vol   = radius2volume(kidsize, scaleradius, scalev);
    share->list[[energy getIam]] = vol;
    return share;
}

-(Quad*) getRoot
{
    Quad* myRoot;
    myRoot = myquad->parent;
    while ((myRoot->type != sessile) && (myRoot->type != mobile))
        myRoot = myRoot->parent;
    return myRoot;
}


-(BOOL) reproduce
    // False if didn't.  
{
    Animal* baby;
    GCoord distance;
    ResList* kidgets;
    double angle;
    Quad* myroot;
    
    if (radius < breedsize)
        return 0;

    angle = 2.0 * M_PI * [randSource evenRand];
    // record a birth
    if (kidcount == 0)
        firstbornon = [geckoControl getCurrentTime];
    kidcount++;

    // make the kid
    baby         = [self copymom];
    distance     = [self getChildRadius];
    baby->x      = x + ( distance * cos(angle) );
    baby->y      = y + ( distance * sin(angle) );
    baby->wannax = cos(angle);
    baby->wannay = sin(angle);
    baby->radius = kidsize;

    // transfer resources 
    kidgets = [self kidshare];
    [energy debit: kidgets];
    [self resetRadius];
    baby->energy = [baby->energy setStores:  kidgets];

    // hang on quad tree
    [baby getMcr: baby->myquad->mcr];
    myroot = [self getRoot];
// GINGER:
if ((myroot == nil) || ((myroot->type != sessile) && (myroot->type != mobile))) {
    printLog("***ERROR: adding baby to :");
    [myroot printQuad]; printLog("\n");
}
    [myroot addNode: baby->myquad];

    [kidgets drop];
    return 1;
}

-hunt {return self; }

-move
{
    GCoord scratchx, scratchy;
    GCoord delta = [self getMoveRadius];
    if (delta == 0.0)
        return self;
    
    [self getDirectionX: &scratchx Y: &scratchy]; // ignore, because reset wanna's
    x += fabs(delta) * wannax;
    y += fabs(delta) * wannay;
    wannax = wannay = 0.0;	// done with this intention.
    [self getMcr: myquad->mcr];
    [myquad growMcr];
    return self;
}

-step 
    // One step in the lifecycle of a plant.
    // Note: returns nil if dies--higher level routine deletes, because
    // higher level routine is traversing the tree.
{
    [self metabolize];
 
    if (![self paytax]) 
        return nil;
    if (![self reproduce]) {
        [self hunt];
    }
    [self move];

    return self;
}

-print
{
    printLog("%s at (", [species getParamName]);
    printLogD("%3f,", x);
    printLogD("%3f),", y);
    printLogD(" size %2f", radius);
    return self;
}

-println
{
    [self print]; 
    printLog("\n");
    return self;
}

-printv
{
    [self println];
    printLogD("breedsize = %3f, ", breedsize);
    printLogD("kidsize = %3f, ", kidsize);
    printLogD("minsize = %3f, ", minsize);
    printLogD("scaler = %3f, ",scaleradius);
    printLogD("scalev = %3f\n",scalev);
    printLog("birthday = %d, firstbornon = %d, kidcount = %d\n",
        birthday, firstbornon, kidcount);
    [energy printv];
    printLog("\n");
    return self;
}

@end
