// Copyright (C) 1997 - 1999 Jan Kreft <Kreft@cardiff.ac.uk>

#import "Critter.h"

typedef enum {oxygen, ammonia, nitrite, nitrate} BacillusSub;

@interface Bacillus: Critter { // Critter inherits SwarmObjects
  // all these instance vars are params set by the protocol file bacillus
  int process; // identifier for process, one species carries out one process
  BacillusSub sub1;    // identifier for type of substrate
  BacillusSub sub2;
  GRate y1, y2;
  GRate vmax;
  GRate km1, km2, ki1;
  GRate maintenanceR;
  EpsType eps; // eps switch, see EpsType for possible settings
  GRate epsProdCoeff, epsProdConst, epsDensity;
  
  // all these instance vars are variables
  GRate myS1, myS2; // my subConcns
  GRate uptakeRate;
  ResQuant dryMass, lastDryMass, growthRate;
  ResQuant epsMass;
  GCoord epsRadius;
  int  myGrandma; // which Grandma do I stem from
  char myName[64];  // identifier for cell and lineage, my grandma 0, first division gives 00 and 01, etc.
}
// creating ...
-recycle;			// add to recycling list -of this type-
-reuse;			        // scrounge recycling list -of this type-
-recycleClear;                  // clear out recycling list
+create;
-copymom;			  // self=mom, returns baby
-liveGrandma: (Species*) s;	  // initialize/reinit a default Bacillus
-setX: (GCoord) cx Y: (GCoord) cy Radius: (GCoord) r;
-setAgentsFromFile: (char*) filename; // initialize/reinit from file, an alternative to the above

// mutating ...
-(int) deviateFromIParam: (char*) ThisIParam bySigma: (double) sigma;
-(double) deviateFromDParam: (char*) ThisDParam bySigma: (double) sigma;
-(int) varyIValue: (int) iv bySigma: (double) sigma;
-(double) varyDValue: (double) dv bySigma: (double) sigma;

// setting params ...
-addSpeciesParams;
-setProcess: (int) p;
-setSub1: (BacillusSub) bs;
-setSub2: (BacillusSub) bs;
-setShoveRadius: (GCoord) r;
-setConvertBack: (GRate) cb;
-setBreedSize: (GCoord) bs;
-setY1: (GRate) yd;
-setY2: (GRate) yd;
-setMaintenanceR: (GRate) m;
-setDensity: (GRate) d;
-setVmax: (GRate) mur;
-setKm1: (GRate) k;
-setKm2: (GRate) k;
-setKi1: (GRate) k;
-setEps: (EpsType) e;
-setEpsProdCoeff: (GRate) p;
-setEpsProdConst: (GRate) p;
-setEpsDensity: (GRate) d;
-(BOOL) haveEps; // returns true if we have eps, Diffuse wants to know
-normalize; // normalize Km and Ki values

// working ...
-(GRate) doTheLocomotion; // batch job calling move
-move;		  // based on shoves from neighbors

-(int) getProcess;
-(EpsType) getEps;
-(GRate) getMyS1;
-(GRate) getMyS2;
-(GRate) getRateForReaction: (int) r Sub1: (GRate) s1 Sub2: (GRate) s2;
-grow;

-(ResQuant) getGrowthRate;
-(ResQuant) getDryMass;
-(ResQuant) getVolume; // dummy method

-update;

-checkDeath;
-die;

-checkVolume;
-(BOOL) reproduce;		// false if didn't -choose- to.

-step;
// printing ...
-printv;
-printSpheresMatlab;
-printFed;
@end
