// Copyright (C) 1997 - 1999 Jan Kreft <Kreft@cardiff.ac.uk>

#import <objectbase.h>
#import <simtools.h>
#import "RandSource.h"
#import "Critter.h"
#import "Bacillus.h"
#import "LandRoot.h"
#import "GeckoControlSwarm.h"
#import "Stat.h"
#import "ResourceQuad.h"
#import "Diffuse.h"
#import "Base.h"
#import "Eps.h"

@implementation Bacillus

// global vars valid for all Bacilli
Bacillus* recycleBacilluss = 0;

static LandRoot* landRoot;
static DiffuseQuad* actualSite;
static Quad* mobileRoot;
static Mcr* landMcr;

// global parameters valid for all Bacilli, set by protocol!
static GRate density;                   // the density of the cells (dryMass)
static GCoord shoveRadius;		// multiplier on radius for shoving purposes
static GRate convertBack;		// when die, multiplier on mass->substrate

static BOOL someHaveEps = FALSE; // do we have eps at all?

// geometry info we get from DiffuseQuad, set via Protocol
static BOOL   periodicY; // if we have a plate, we have periodicY boundaries
static BOOL   biofilm;   // should be set from protocol via modified "Plate" directive, make plate and periodicY independent
static GCoord carrierLayer; // height of the carrier
static GCoord lengthX;   // length of the non-periodic coordinate
static GCoord lengthY;   // length of the periodic coordinate
static GCoord lengthZ;   // length of the third dimension, also with periodic boundaries
static GCoord cellSize;  // length of a lattice cell (all sides equal)
static int numSubs, numSpecies; // something else that Diffuse knows

// stats for yields, per species and per substrate
static GRate** popFed; // popFed[species][sub]

static GRate KaNH3;  // Acidity constant for Ammonia
static GRate KbHNO2; // Basicity constant for Nitrite

-recycle
{
    [species tallyDeath: self];
    hook = recycleBacilluss;
    recycleBacilluss = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    Bacillus* togo;
    while (recycleBacilluss) {
	togo = recycleBacilluss;
	recycleBacilluss = recycleBacilluss->hook;
	[togo drop];
    }
    return self;
}

-reuse
{
    Bacillus* result = recycleBacilluss;
    recycleBacilluss = recycleBacilluss->hook;
    result->hook    = nil;
    return result;
}

+create 
{
    id result;

    if (recycleBacilluss) {
	result = [recycleBacilluss reuse];
    } else {
	result = [Bacillus createBegin: globalZone];
    }
    return result;					
}

-copymom {
    // create an agent matching mom in all heritable respects

    Bacillus* baby;
    Quad* newQuad;

    baby = [Bacillus create];
    baby->species   = species;
    //baby->color     = color; // inherit color
    [baby setGColor: [species getGColor]]; // set to default

    [species incrementCensus];
    [species tallyBirth: self];

    // here come our parameters
    // mutate params here if we also want diverse grandmas
    // vary hereditary params, mom and baby equal
    baby->breedsize    = breedsize;
    baby->kidsize      = kidsize; 
    baby->minsize      = minsize;
    baby->process      = process;
    baby->myGrandma    = myGrandma;
    strcpy(baby->myName, myName); // copy string myName into baby->myName
    baby->sub1         = sub1;
    baby->sub2         = sub2;
    baby->vmax         = vmax;
    baby->km1          = km1;
    baby->ki1          = ki1;
    baby->km2          = km2;
    baby->y1           = y1;
    baby->y2           = y2;
    baby->maintenanceR = maintenanceR;
    baby->eps          = eps;
    baby->epsProdCoeff = epsProdCoeff;
    baby->epsProdConst = epsProdConst;
    baby->epsDensity   = epsDensity;
    baby->epsMass      = epsMass; // if we don't have eps, this will be zero and should stay zero
    baby->epsRadius    = epsRadius;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
}

-liveGrandma: (Species*) s {
  // Initialize/reinit a default Bacillus which is as empty as possible
  // This is called very early, before setX:Y:Radius:,
  // so early that some things set here get overwritten later
  // That's why we do our initializations later in setX:Y:Radius:

  //printLog("liveGrandma in Bacillus was called\n");

  species = s;

  [self addSpeciesParams];

  return self;
}

-setAgentsFromFile: (char*) filename {
  // This is called in GeckoModelSwarm -hangCritters:Num:Root:Where:
  // when agents are to be read in from file. It will be called only once per run.

  int i, j, tokenCount;
  int thisProcess;
  Bacillus* thisOne;
  FILE* file;
  int num = 0;
  int numread;
  //char *line;
  char *got; // result from fgets
  char line[4096]; // line read in
  const char delimiters[] = " \t\n\r";
  char *token;

  file = fopen(filename, "r");
  if (file == NULL) {
    geckoError(SourceMessage, "\n\nERROR: Could not open agent-list file %s for reading!\n\n\n", filename);    
  }

  //numread = getline(&line, &num, file);
  got = fgets(line, sizeof(line), file);

  //while (numread > 1) { // while we have a non-empty line
  while (got) { // while we have a non-empty line

    token = strtok (line, delimiters);
    tokenCount = 1;

    thisProcess = atoi(token);
    thisOne = [[[geckoControl getModelSwarm] getSpeciesHandle: thisProcess] copymom];

    while (token != NULL) {
      token = strtok (NULL, delimiters);
      if (token != NULL) {
	tokenCount++;
	switch (tokenCount) {
	case 2:
	  // We throw away random draws that would have been used if we were not reading from file
	  thisOne->x = atof(token); break;
	case 3:
	  thisOne->y = atof(token); break;
	case 4:
	  thisOne->z = atof(token); break;
	case 5:
	  thisOne->radius = atof(token);
	  thisOne->dryMass = density * radius2sphere( thisOne->radius );
	  break;
	case 6:
	  thisOne->myGrandma = atoi(token); break;
	case 7:
	  if (strlen(thisOne->myName) < strlen(token)) {
	    strcpy(thisOne->myName, token);
	  } else { printLog("Warning, myName too short"); } break;
	case 8:
	  thisOne->growthRate = atof(token);
	  thisOne->lastDryMass = thisOne->dryMass - thisOne->growthRate;  
	  break;
	case 9:
	  thisOne->vmax = atof(token); break;
	case 10:
	  thisOne->epsRadius = atof(token);
	  if (thisOne->epsRadius > 0)
	    thisOne->epsMass = thisOne->epsDensity * (radius2sphere(thisOne->epsRadius) - (thisOne->dryMass/density));
	  else
	    thisOne->epsMass = 0;
	  break;
	default:
	  printLog("Warning, unknown case in file init\n"); break;
	}
      }	
    }
    // to be done for every agent/line-read-in
    thisOne->birthday 	= [geckoControl getCurrentTime];
    thisOne->kidcount 	= 0;
    //[thisOne setGColor: 8]; // we want a different grandma, but protocol-set color for all other bacilli
    // 0 black, 6 yellow, 8 red, 9 orange, rest hard to see on blue background
    [thisOne getMcb: thisOne->myquad->mcr]; // after this, we have an mcr and can ask for our site
    landRoot = [[geckoControl getModelSwarm] getLandRoot];
    actualSite = [landRoot firstQuadProducing: 1 Over: [thisOne->myquad getMcr]];
    mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
    landMcr    = [actualSite getMcr];
    periodicY = [actualSite getPeriodicY];
    if (periodicY) biofilm = TRUE;
    carrierLayer = [actualSite getCarrierLayer];
    lengthX   = [actualSite getLengthX];
    lengthY   = [actualSite getLengthY];
    lengthZ   = [actualSite getLengthZ];
    cellSize  = [actualSite getCellSize];
    [thisOne update];
    [thisOne normalize];
    [mobileRoot addNode: thisOne->myquad];
    
    got = fgets(line, sizeof(line), file);
    //numread = getline(&line, &num, file);
  }
  
  // to be done once at the end
  numSubs = [actualSite getNumSubs];
  numSpecies = [actualSite getNumSpecies];
  popFed = (GRate**) [globalZone alloc: numSpecies * sizeof(GRate*)];
  for (i=0; i < numSpecies; i++) popFed[i] = (GRate*) [globalZone alloc: numSubs * sizeof(GRate)];
  for (i=0; i < numSpecies; i++) {
    for (j=0; j < numSubs; j++)
      popFed[i][j] = 0.0;
  }

  fclose(file);
  return self;
}

-setX: (GCoord) cx Y: (GCoord) cy Radius: (GCoord) r {
  // This is called in GeckoModelSwarm -hangCritters: grandma Num: num Root: root Where: where
  // to make the very first Bacillus (Grandma) with r being grandma->kidsize
  
  static BOOL precise = FALSE;
  static BOOL random  = FALSE;
  static int thisRun = -1;
  static int grandmaCount;
  int i, j;
  char buffer[256];
  GRate tmp;
  
  if (random) { // random grandma switch
    // random size for grandma, but must be > minsize and < breedvol, don't change starting point r
    do {
      tmp = [self varyDValue: r bySigma: 0.4];
    } while ((tmp <= minsize) || (tmp >= sphere2radius(breedsize)));
  } else {
    tmp = r;
  }

  if (precise) {
    x = cx;
    y = cy;
    // z comes later
    radius = tmp;
  } else { // reduce precision to 6 digits, same as in printing using %g
    sprintf(buffer, "%g", cx); x = atof(buffer);
    sprintf(buffer, "%g", cy); y = atof(buffer);
    // z comes later
    sprintf(buffer, "%g", tmp); radius = atof(buffer);
  }

  epsMass = 0.0; // EPS stuff only ever changed if eps flag is TRUE
  epsRadius = 0.0;
  growthRate = 0;

  // this is a good place for initializations since also called at
  // start of second run and only called at start of runs
  birthday 	= [geckoControl getCurrentTime];
  kidcount 	= 0;
  dryMass = density * radius2sphere(radius);
  lastDryMass = dryMass - growthRate;  
  //[self setGColor: 8]; // we want a different grandma, but protocol-set color for all other bacilli
  // 0 black, 6 yellow, 8 red, 9 orange, rest hard to see on blue background
  [self getMcb: myquad->mcr]; // after this, we have an mcr and can ask for our site
  landRoot = [[geckoControl getModelSwarm] getLandRoot];
  actualSite = [landRoot firstQuadProducing: 1 Over: [myquad getMcr]];
  mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
  landMcr    = [actualSite getMcr];
  periodicY = [actualSite getPeriodicY];
  if (periodicY) biofilm = TRUE;
  carrierLayer = [actualSite getCarrierLayer];
  lengthX   = [actualSite getLengthX];
  lengthY   = [actualSite getLengthY];
  lengthZ   = [actualSite getLengthZ];
  cellSize  = [actualSite getCellSize];

  [self update];

  // Now we can set z randomly in the range [0,lengthZ]
  if (precise) {
    z = lengthZ * [randSource evenRand];
  } else {
    sprintf(buffer, "%g", lengthZ * [randSource evenRand]); z = atof(buffer);
  }

  // Out of bounds check, if we get warnings here, the values in the protocol file should be changed
  if (biofilm) { // cells should not be within the carrier nor in the topmost layer which is always bulk liquid
    if (x <= carrierLayer) {
      x = carrierLayer + EPSILON; printLog("Warning, x < cellSize, change birtharea for %d in protocol\n", process);
    }
    if (x >= (lengthX-cellSize)) {
      x = lengthX - cellSize - EPSILON;
      printLog("Warning, x > (lengthX-cellSize), change birtharea for %d in protocol\n", process);
    }
  }
  if (periodicY) {
    if (y < 0)        y += lengthY;
    if (y >= lengthY) y -= lengthY;
  }
  // periodicZ always true so far
  if (z < 0)        z += lengthZ;
  if (z >= lengthZ) z -= lengthZ;

  [self getMcb: myquad->mcr]; // have to repeat after changing z
  
  [self normalize];
  
  if (thisRun != [geckoControl getNumRun]) { // executed only at the beginning of each run
    thisRun = [geckoControl getNumRun];
    grandmaCount = 0;
    numSubs = [actualSite getNumSubs];
    numSpecies = [actualSite getNumSpecies];
    popFed = (GRate**) [globalZone alloc: numSpecies * sizeof(GRate*)];
    for (i=0; i < numSpecies; i++) popFed[i] = (GRate*) [globalZone alloc: numSubs * sizeof(GRate)];
    for (i=0; i < numSpecies; i++) {
      for (j=0; j < numSubs; j++)
	popFed[i][j] = 0.0;
    }
  }
  
  grandmaCount++;
  myGrandma = grandmaCount;
  myName[0]   = '0';
  myName[1]   = '\0';

  return self;
}

-normalize { // normalize Km and Ki values

  Param* param;

  // Start from scratch
  param = [species getParam: "Km1"];
  km1 = [param getPartDValue];
  param = [species getParam: "Ki1"];
  ki1 = [param getPartDValue];
  param = [species getParam: "Km2"];
  km2 = [param getPartDValue];

  // First convert Km and Ki for nitrogen substrate from neutral to ionized
  // This follows from K(NH4)/K(NH3) = C(NH4)/C(NH3)

  KaNH3  = exp(6344/(273+celsius));  // Acidity constant
  KbHNO2 = exp(-2300/(273+celsius)); // Basicity constant
  
  // only sub1, the nitrogen substrate, has to be converted into neutral, unionized form
  switch (sub1) {
  case 1: // ammonia
    km1 = km1 * KaNH3 / pow(10,pH);
    ki1 = ki1 * KaNH3 / pow(10,pH);
    break;
  case 2: // nitrite
    km1 = km1 * (pow(10,pH) * KbHNO2);
    ki1 = ki1 * (pow(10,pH) * KbHNO2);
    break;
  default: printLog("error in switch statement in Bacillus\n"); break;
  }

  // We normalize subConcn by dividing Kms and Kis and all subConcn in DiffuseQuad by the external subConcn
  km1 /= [actualSite getExtConcn: sub1 ];
  ki1 /= [actualSite getExtConcn: sub1 ];
  km2 /= [actualSite getExtConcn: sub2 ];

  return self;

}

-setBreedSize: (GCoord) bs {

    ResQuant kidvol, minvol;
    
    breedsize = bs;
    
    // Note that breedsize has the meaning of Volume in contrast to
    // standard Gecko where all sizes are radii (see Critter.h)

    kidvol   = 0.75 * breedsize;
    kidsize  = sphere2radius(kidvol);
    minvol   = 0.2 * breedsize;
    minsize  = sphere2radius(minvol);

    return self;
}

-setProcess: (int) p { process = p; return self; }
-setSub1: (BacillusSub) bs { sub1 = bs; return self; }
-setSub2: (BacillusSub) bs { sub2 = bs; return self; }
-setShoveRadius: (GCoord) r {shoveRadius = r; return self;}
-setConvertBack: (GRate) cb {convertBack = cb; return self;}
-setY1: (GRate) yd {y1 = yd; return self;}
-setY2: (GRate) yd {y2 = yd; return self;}
-setMaintenanceR: (GRate) m {maintenanceR = m * [[geckoControl getModelSwarm] getStepSize]; return self;}
-setDensity: (GRate) d {density = d; return self;}
-setVmax: (GRate) mur {vmax = mur * [[geckoControl getModelSwarm] getStepSize]; return self;}
-setKm1: (GRate) k {km1 = k; return self;}
-setKi1: (GRate) k {ki1 = k; return self;}
-setKm2: (GRate) k {km2 = k; return self;}
-setEps: (EpsType) e {eps = e; if (eps != NoEps) someHaveEps = TRUE; return self;}
-setEpsProdCoeff: (GRate) p {epsProdCoeff = p; return self;}
-setEpsProdConst: (GRate) p {epsProdConst = p * [[geckoControl getModelSwarm] getStepSize]; return self;}
-setEpsDensity: (GRate) d {epsDensity = d; return self;}
-(BOOL) haveEps { return someHaveEps; }

-addSpeciesParams {
  [species addParam: "Density" Vary: 0 DValue: density
	   Message: M(setDensity:) Owner: self];
  
  [species addParam: "ConvertBack" Vary: 0 DValue: convertBack
	   Message: M(setConvertBack:) Owner: self];
  
  [species addParam: "BreedSize" Vary: 0 DValue: breedsize
	   Message: M(setBreedSize:) Owner: self];
  [species addParam: "ShoveRadius" Vary: 0 DValue: shoveRadius
	   Message: M(setShoveRadius:) Owner: self];
  
  [species addParam: "Process" Vary: 0 IValue: process 
	   Message: M(setProcess:) Owner: self];
  [species addParam: "Sub1" Vary: 0 IValue: sub1
	   Message: M(setSub1:) Owner: self];
  [species addParam: "Sub2" Vary: 0 IValue: sub2
	   Message: M(setSub2:) Owner: self];

  [species addParam: "Y1" Vary: 0 DValue: y1
	   Message: M(setY1:) Owner: self];
  [species addParam: "Y2" Vary: 0 DValue: y2
	   Message: M(setY2:) Owner: self];
  [species addParam: "MaintenanceR" Vary: 0 DValue: maintenanceR 
	   Message: M(setMaintenanceR:) Owner: self];
  [species addParam: "Vmax" Vary: 0 DValue: vmax
	   Message: M(setVmax:) Owner: self];
  [species addParam: "Km1" Vary: 0 DValue: km1
	   Message: M(setKm1:) Owner: self];
  [species addParam: "Ki1" Vary: 0 DValue: ki1
	   Message: M(setKi1:) Owner: self];
  [species addParam: "Km2" Vary: 0 DValue: km2
	   Message: M(setKm2:) Owner: self];

  [species addParam: "Eps" Vary: 0 IValue: eps
	   Message: M(setEps:) Owner: self];
  [species addParam: "EpsProdCoeff" Vary: 0 DValue: epsProdCoeff
	   Message: M(setEpsProdCoeff:) Owner: self];
  [species addParam: "EpsProdConst" Vary: 0 DValue: epsProdConst
	   Message: M(setEpsProdConst:) Owner: self];
  [species addParam: "EpsDensity" Vary: 0 DValue: epsDensity
	   Message: M(setEpsDensity:) Owner: self];

  return self;   
}

-(int) deviateFromIParam: (char*) ThisIParam bySigma: (double) sigma {
  // changing a value from the default, no change of the default itself!
  // the "I" method varies ints
  Param* param = [species getParam: ThisIParam];
  int result;
  double deviation;
  double cv = sigma; // coefficient of variation = relative standard deviation

  do {
    // restrict change to a range of 2 * relative standard deviation
    do deviation = [randSource gaussRand];
    while ((deviation >= 2.0) || (deviation <= -2.0));
    result = (int) (0.5 + [param getPartIValue] * (1.0 + cv * deviation));
  } while (!samesign(result, [param getPartIValue]));

  return result;   
}

-(double) deviateFromDParam: (char*) ThisDParam bySigma: (double) sigma {
  // changing a value from the default, no change of the default itself!
  // the "D" method varies doubles
  Param* param = [species getParam: ThisDParam];
  double result;
  double deviation;
  double cv = sigma; // coefficient of variation = relative standard deviation

  do {
    // restrict change to a range of 2 * relative standard deviation
    do deviation = [randSource gaussRand];
    while ((deviation >= 2.0) || (deviation <= -2.0));
    result = [param getPartDValue] * (1.0 + cv * deviation);
  } while (!samesign(result, [param getPartDValue]));

  return result;   
}

-(int) varyIValue: (int) iv bySigma: (double) sigma {
  // change the current value, new value may be inherited or not
  // the "I" method varies ints
  int val = iv;
  int result;
  double deviation;
  double cv = sigma; // coefficient of variation = relative standard deviation

  do {
    // restrict change to a range of 2 * relative standard deviation
    do deviation = [randSource gaussRand];
    while ((deviation >= 2.0) || (deviation <= -2.0));
    result = (int) (0.5 + val * (1.0 + cv * deviation));
  } while (!samesign(result, val));

  return result;   
}

-(double) varyDValue: (double) dv bySigma: (double) sigma {
  // change the current value, new value may be inherited or not
  // the "D" method varies doubles
  double val = dv;
  double result;
  double deviation;
  double cv = sigma; // coefficient of variation = relative standard deviation

  do {
    // restrict change to a range of 2 * relative standard deviation
    do deviation = [randSource gaussRand];
    while ((deviation >= 2.0) || (deviation <= -2.0));
    result = val * (1.0 + cv * deviation);
  } while (!samesign(result, val));

  return result;   
}

-move {
  // OK, so I lied.  In theory, it shoves neighbors.  In practice,
  // each finds neighbors to be shoved by.  Would be more efficient
  // doing as a batch once-through process, but first do it right
  // then make it faster if and when.  As of the moment, the distance
  // shoved is simply the vector sum of overlap radii.
  
  Quad* myRoot;
  Critter* neighbor;
  Quad* nq;
  GCoord overlapRadius, distRadius;
  
  static BOOL firsttime = 1;
  static Mcr* shovewin;		// equiv-area square
  
  timeval_t now = [geckoControl getCurrentTime];

  if (firsttime) {
    firsttime = 0;
    shovewin = [Mcr create];
  }
  
#ifndef MUTUAL
  wannax = wannay = wannaz = 0.0;
#endif MUTUAL

  // pick around for neighbors inside shovewin
  myquad->visible = 0;
  
  // if we are producing eps, our distance to neighbors is explicitly modelled by the total volume taken up by
  // EPS and biomass, else we multiply our radius by a calibrated factor to avoid overlap between cells
  switch (eps) {
  case NoEps:
    distRadius = shoveRadius * radius;
    [shovewin circleToBrickX: x Y: y Z: z Radius: distRadius];
    break;
  case Capsule:
    distRadius = shoveRadius * epsRadius;
    [shovewin circleToBrickX: x Y: y Z: z Radius: distRadius];
    break;
  case Slime:
    distRadius = shoveRadius * epsRadius;
    [shovewin circleToBrickX: x Y: y Z: z Radius: distRadius];
    break;
  case SelfBinding: case AllBinding:  case SlimersBinding:
    distRadius = shoveRadius * epsRadius;
    // need larger shovewin if we also deal with attractive forces
    [shovewin circleToBrickX: x Y: y Z: z Radius: 3*distRadius];
    break;
  }

  myRoot = [self getRoot];
  nq = [myquad pickWin: shovewin Root: myRoot Invisibles: 0]; 
  while (nq) {
    // get one shove
    neighbor = nq->data;
    overlapRadius = distRadius;
    overlapRadius -= distance3D(x, y, z, neighbor->x, neighbor->y, neighbor->z) - neighbor->radius;

    if (overlapRadius > 0) { // cells too close
#ifdef MUTUAL
      [neighbor addDirectionX: neighbor->x - x Y: neighbor->y - y Z: neighbor->z - z Weight: overlapRadius];
#endif MUTUAL
#ifdef SQUARE
      // squaring the overlapRadius inhibits gliding and wiggling by placing less weight on slight overlaps
      [self addDirectionX: x - neighbor->x Y: y - neighbor->y Z: z - neighbor->z Weight: overlapRadius*overlapRadius];
#else SQUARE
      [self addDirectionX: x - neighbor->x Y: y - neighbor->y Z: z - neighbor->z Weight: overlapRadius];
#endif SQUARE
    }
    else { // cells too far
      
      switch (eps) {
      default: break; // do nothing
      case SelfBinding: // neighbor same species? Note that we are not eps.
	if (process == [neighbor getProcess])
	  // attractive force has to be negative, weak, slowly rising at the beginning, and fading in the distance
	  [self addDirectionX: x - neighbor->x Y: y - neighbor->y Z: z - neighbor->z Weight:
		  -0.01*(overlapRadius*overlapRadius)/((-overlapRadius+1)*(-overlapRadius+1)*(-overlapRadius+1))];
	break;
      case AllBinding: // neighbor any species, that is, not eps?
	if ([neighbor getProcess] >= 0)
	  // attractive force has to be negative, weak, slowly rising at the beginning, and fading in the distance
	  [self addDirectionX: x - neighbor->x Y: y - neighbor->y Z: z - neighbor->z Weight:
		  -0.01*(overlapRadius*overlapRadius)/((-overlapRadius+1)*(-overlapRadius+1)*(-overlapRadius+1))];
	break;
      case SlimersBinding:
	if ([neighbor getEps] != NoEps)
	  // attractive force has to be negative, weak, slowly rising at the beginning, and fading in the distance
	  [self addDirectionX: x - neighbor->x Y: y - neighbor->y Z: z - neighbor->z Weight:
		  -0.01*(overlapRadius*overlapRadius)/((-overlapRadius+1)*(-overlapRadius+1)*(-overlapRadius+1))];
	break;
      }

    }
    
    // next
    nq = [nq pickWin: shovewin Root: myRoot Invisibles: 0]; 
  }
  
  myquad->visible = 1;

  if (biofilm) {
    // Then get a shove from the concrete wall called carrier as well (none if we don't sit on carrier)
    // carrier is always at the low x side
    overlapRadius = [actualSite getOverlapWithCarrierAtX: (x - distRadius) Y: y ];
    if (overlapRadius > 0) {
      [self addDirectionX: 1 Y: 0 Z: 0 Weight: overlapRadius];
    }
  }
  
  // now do it
  x += wannax;
  y += wannay;
  z += wannaz;

  // Out of bounds check
  if (biofilm) { // cells should not be within the carrier nor in the topmost layer which is always bulk liquid
    if (x <= carrierLayer) {
      x = carrierLayer + EPSILON; printLog("Warning, in move for %d at %d x < cellSize\n", process, now);
    }
    if (x >= (lengthX-cellSize)) {
      x = lengthX - cellSize - EPSILON; printLog("Warning, in move for %d at %d x > (lengthX-cellSize)\n", process, now);
    }
  }
  if (periodicY) {
    if (y < 0)        y += lengthY;
    if (y >= lengthY) y -= lengthY;
  }
  // periodicZ always true so far
  if (z < 0)        z += lengthZ;
  if (z >= lengthZ) z -= lengthZ;
  
  [self getMcb: myquad->mcr];
  [myquad growMcr];
  
#ifdef MUTUAL
  wannax = wannay = wannaz = 0.0;
#endif MUTUAL

  return self;
}

-(GRate) doTheLocomotion {

  int count = 0;
  Quad* q;
  BOOL relaxed = FALSE;
  GRate minPorosity;

  while (!relaxed) {
    count++;

    //traverse the agent quadding tree and tell everyone to move, including Eps
    
    q = [mobileRoot pickWin: mobileRoot->mcr Root: mobileRoot Invisibles: 0];
    
    while (q) {
      
      [q->data move]; // this does the actual work
      
      q = [q pickWin: mobileRoot->mcr Root: mobileRoot Invisibles: 0];
    }

    minPorosity = [actualSite calcPorosity];
    if (minPorosity > 0.26) {
      relaxed = TRUE;
    }
    if (count > 1000) { // We are probably stuck in a traffic jam, no hope ever to get out
      geckoError(SourceMessage, "Warning: We made more than 1000 loco_loops: Overcrowded?\n");
    }
  }

  if (count > 1) printLog("loco_loop %d\n", count);
    
  return minPorosity;
}

// Diffuse and GeckoModelSwarm wants to know
-(int) getProcess { return process; }

-(EpsType) getEps { return eps; }

// for our own use in printSubConcnHet, asking each agent for their values
-(GRate) getMyS1 { return myS1; }
-(GRate) getMyS2 { return myS2; }

-(GRate) getRateForReaction: (int) r Sub1: (GRate) s1 Sub2: (GRate) s2 {
  GRate rate;//, s1n;

  switch (r) { // for which reaction should we supply the rate? (We may have more than 4 cases in future)
  case 1:
    // Nitroso: ammonia consumption, s1 is ammonia and s2 is oxygen
    rate = - dryMass * vmax * (s1/(km1+s1+(s1*s1/ki1))) * (s2/(km2+s2));
    break;
  case 2:
    // Nitroso: oxygen consumption, s1 is ammonia and s2 is oxygen
    rate = - dryMass * (y1/y2) * vmax * (s1/(km1+s1+(s1*s1/ki1))) * (s2/(km2+s2));
    break;
  case 11:
    // Nitro: nitrite consumption, s1 is nitrite and s2 is oxygen
    rate = - dryMass * vmax * (s1/(km1+s1+(s1*s1/ki1))) * (s2/(km2+s2));
    break;
  case 12:
    // Nitro: oxygen consumption, s1 is nitrite and s2 is oxygen
    rate = - dryMass * (y1/y2) * vmax * (s1/(km1+s1+(s1*s1/ki1))) * (s2/(km2+s2));
    break;
  }
  return rate;
}

-grow {
  // inquire about all relevant subConcn's
  //myS1 = [actualSite  getConcnSub: (BacillusSub) sub1 X: (GCoord) x Y: (GCoord) y];
  //myS2 = [actualSite  getConcnSub: (BacillusSub) sub2 X: (GCoord) x Y: (GCoord) y];
  myS1 = [actualSite  avgSub: (BacillusSub) sub1 X: (GCoord) x Y: (GCoord) y Radius: (GCoord) radius];
  myS2 = [actualSite  avgSub: (BacillusSub) sub2 X: (GCoord) x Y: (GCoord) y Radius: (GCoord) radius];
  
  // grow: take up sub, metabolize sub taken up completely, and maintain
  // growth independent substrate consumption = maintenance rate
  // maintenance is debited from dryMass, hence it's independent of substrates
    
  // The meaning of rate is amount per timestep, as the timestep is implicit, rates are actually amounts
  uptakeRate = dryMass * vmax * (myS1/(km1+myS1+(myS1*myS1/ki1))) * (myS2/(km2+myS2));

  popFed[process-1][sub1]  += uptakeRate;
  popFed[process-1][sub2]  += uptakeRate * (y1/y2);

  // this is not specific growth rate but dx/dt
  // also, this is not per min, but per timestep, in -printv it is per min
  growthRate = uptakeRate * y1 - dryMass * maintenanceR;

  // EPS production is a function of growth a la Luedeking-Piret expression
  if (eps != NoEps) {
    growthRate = (uptakeRate * y1 - dryMass * (maintenanceR + 1.214286*epsProdConst))/(1 + 1.214286*epsProdCoeff);
    epsMass += epsProdCoeff * growthRate + dryMass * epsProdConst;
  }

  lastDryMass = dryMass;
  dryMass += growthRate;
  
  return self;
}

-(ResQuant) getGrowthRate {return growthRate;}

-(ResQuant) getDryMass {return dryMass;}

-(ResQuant) getVolume {
  // dummy method, this method is called by GeckoModelSwarm to calc "Biomass"
  return dryMass;
}

-update {
  // dryMass and lastDryMass are set outside and before -update
  // in -grow, -reproduce, and for grandma in -setX:Y:Radius:
  // -update only deals with dependent variables

  static BOOL firsttime = TRUE;
  static Eps* theOne; // We save off this handle to the first real Eps blob for future messages to Eps
  GCoord distance, xi, yi, zi;
  double alpha, beta;
  timeval_t now = [geckoControl getCurrentTime];
  //Quad* q;
  
  radius = sphere2radius(dryMass / density);
  if ((radius * CubertPiOver6) > cellSize)
    printLog("Warning, for %d at %d (radius * CubertPiOver6) > cellSize\n", process, now); 
  [species tallyGrowth: self Amount: growthRate]; // not really used, and not checked
  
  switch (eps) {
  default: break;
  case Capsule:
    // epsRadius is from cube with total volume (dryMass and EPS)
    //epsRadius = pow( ((dryMass / density) + (epsMass / epsDensity)) , 1.0/3.0 ) / 2.0;
    epsRadius = sphere2radius( (dryMass / density) + (epsMass / epsDensity) );
    break;
  case Slime: case SelfBinding: case AllBinding: case SlimersBinding:
    if ((epsMass / epsDensity) > breedsize) {
      // We want eps blob and excreter adjacent with the same center (volume not mass) as excreter before excretion,
      // both placed in a random direction
      distance = pow( ((dryMass / density) + (epsMass / epsDensity)) , 1.0/3.0 ) / 2.0;
      alpha = 2.0 * M_PI * [randSource evenRand];
      beta  = 2.0 * M_PI * [randSource evenRand];
      // shift eps blob
      xi      = x + ( distance * cos(alpha) * cos(beta) );
      yi      = y + ( distance * sin(alpha) * cos(beta) );
      zi      = z + ( distance * sin(beta)              );
      // now I move, too....
      x -= ( distance * cos(alpha) * cos(beta) );
      y -= ( distance * sin(alpha) * cos(beta) );
      z -= ( distance * sin(beta)              );
      // Out of bounds check
      if (biofilm) { // cells should not be within the carrier nor in the topmost layer which is always bulk liquid
	if (x <= carrierLayer) {
	  x = carrierLayer + EPSILON; printLog("Warning, for %d at %d x < cellSize\n", process, now);
	}
	if (x >= (lengthX-cellSize)) {
	  x = lengthX - cellSize - EPSILON; printLog("Warning, for %d at %d x > (lengthX-cellSize)\n", process, now);
	}
	if (xi <= carrierLayer) {
	  xi = carrierLayer + EPSILON; printLog("Warning, for %d at %d xi < cellSize\n", process, now);
	}
	if (xi >= (lengthX-cellSize)) {
	  xi = lengthX - cellSize - EPSILON; printLog("Warning, for %d at %d xi > (lengthX-cellSize)\n", process, now);
	}
      }
      if (periodicY) {
	if (yi < 0)       yi += lengthY;
	if (yi >= lengthY) yi -= lengthY;
	if (y < 0)       y += lengthY;
	if (y >= lengthY) y -= lengthY;
      }
      // periodicZ always true so far
      if (zi < 0)       zi += lengthZ;
      if (zi >= lengthZ) zi -= lengthZ;
      if (z < 0)       z += lengthZ;
      if (z >= lengthZ) z -= lengthZ;

      if (firsttime) {
	firsttime = FALSE;
	[mobileRoot squashQuad];
	theOne = [[[geckoControl getModelSwarm] getEpsHandle] copymom];
	[theOne setX: xi Y: yi Z: zi Radius: epsMass];
	[mobileRoot addNode: theOne->myquad];
	[mobileRoot subQuad];
	[mobileRoot growMcr];
      }
      else {
	// This is more flexible but slower, needed only if Eps including theOne dies
	//        q = [mobileRoot pickWin: landMcr Root: mobileRoot Invisibles: 1];
	//        while ([q->data getProcess] != -1) { // only Eps returns -1
	//  	    q = [q pickWin: landMcr Root: mobileRoot Invisibles: 1];
	//        }
	//        [q->data excreteEpsAmount: epsMass X: xi Y: yi Z: zi];
	[theOne excreteEpsAmount: epsMass X: xi Y: yi Z: zi];
      }
      epsMass = 0.0;
    }
    // epsRadius is from cube with total volume (dryMass and EPS)
    //epsRadius = pow( ((dryMass / density) + (epsMass / epsDensity)) , 1.0/3.0 ) / 2.0;
    epsRadius = sphere2radius( (dryMass / density) + (epsMass / epsDensity) );
    break;
  }
  return self;
}

-checkDeath {
  
  if (radius < minsize) {
    [self die]; 
    return nil;
  } else {
    return self;
  }       
}

-die {
  static BOOL firsttime = TRUE;
  static Eps* theOne; // We save off this handle to the first real Eps blob for future messages to Eps
  timeval_t now = [geckoControl getCurrentTime];
  
  if (eps != NoEps) {
    if (firsttime) {
      firsttime = FALSE;
      [mobileRoot squashQuad];
      theOne = [[[geckoControl getModelSwarm] getEpsHandle] copymom];
      [theOne setX: x Y: y Z: z Radius: epsMass];
      [mobileRoot addNode: theOne->myquad];
      [mobileRoot subQuad];
      [mobileRoot growMcr];
    }
    else {
      // This is more flexible but slower, needed only if Eps including theOne dies
      //        q = [mobileRoot pickWin: landMcr Root: mobileRoot Invisibles: 1];
      //        while ([q->data getProcess] != -1) { // only Eps returns -1
      //  	    q = [q pickWin: landMcr Root: mobileRoot Invisibles: 1];
      //        }
      //        [q->data excreteEpsAmount: epsMass X: x Y: y Z: z];
      [theOne excreteEpsAmount: epsMass X: x Y: y Z: z];
    }
  }
  
  [species tallyLossStarve: self Amount: dryMass]; 
  
  // should increase the appropriate sub in the sub matrix, don't want landroot to deal with this
  [landRoot payTaxX: x Y: y Resource: sunwater Amount: dryMass * convertBack];
  
  // note that recycle, called by delData, does the census
  [myquad delData];
  
  printLog("Death of %d at %d\n", process, now);
  
  return nil;
}

-checkVolume {
// descriptive Donachie model
  ResQuant breedVol;
  ResQuant gph; // generations per hour
  
  gph = growthRate * 60 / (dryMass * [[geckoControl getModelSwarm] getStepSize] * log(2));
  //printLogD("gph %g\n", gph);

  if (gph > 0) {
    // breedsize in Protocol is (Vu of Donachie = 0.4 fl) * (some factor):
    // see lab book page 147
    breedVol = breedsize * pow(2.0,gph); 
    // if want to make vol measurements imprecise
    breedVol = [self varyDValue: breedVol bySigma: 0.1];
    //printLog("min\t%d\t", [geckoControl getCurrentTime]);
    //printLogD("breedVol\t%g\n", breedVol);
    if ((dryMass / density) >= breedVol) [self reproduce];
  }
  return self;
}

-(BOOL) reproduce {
  // False if didn't.

  Bacillus* baby;
  GCoord distance;
  double alpha, beta;
  Quad* myRoot;
  timeval_t now = [geckoControl getCurrentTime];
  //    BOOL found = FALSE;
  //    Quad* q;

  // mutate params here if we don't want diverse grandmas
  // but changes once every cell cycle, baby and mom equal
  // vary
  //vmax = [self varyDValue: vmax bySigma: 0.1];
  // deviate
  vmax = [self deviateFromDParam: "Vmax" bySigma: 0.1] * [[geckoControl getModelSwarm] getStepSize]; 
  //printLog("min\t%d\t", [geckoControl getCurrentTime]);
  //printLogD("vmax\t%g\n", vmax / [[geckoControl getModelSwarm] getStepSize]);

  // make the kid 
  baby = [self copymom];

  // life stats, kidcount and birthday are inherited from Critter, BUT
  // kidcount is used as a generation counter here and 
  // birthday has the meaning of time of last division
  // kidcount and birthday for mom = kid, since binary fission is: mum splits into two daughters, no mum left
  kidcount++;
  baby->kidcount = kidcount;
  birthday = [geckoControl getCurrentTime];
  baby->birthday = birthday;
  //printLog("lastDivision %d\t generation %d\n", birthday, kidcount);

  // Without checking, we will get seg faults, maybe long after the limit was exceeded!!!
  if (strlen(myName) < sizeof(myName) - 1) {
  strcat(myName, "0"); // lineage string is expanded by one digit
  strcat(baby->myName, "1");
  } else printLog("Warning, for %d at %d length of myName trimmed to the limit of %d\n", process, now, sizeof(myName)-1);

  // update variables, don't call update (update should only be called once per step)
  dryMass = dryMass / 2.0;
  lastDryMass = lastDryMass / 2.0;
  baby->dryMass = dryMass;
  baby->lastDryMass = lastDryMass;
  radius = sphere2radius(dryMass / density);
  baby->radius = sphere2radius(baby->dryMass / density);

  if (eps != NoEps) {
    
    epsMass /= 2.0;
    //epsRadius = pow( ((dryMass / density) + (epsMass / epsDensity)) , 1.0/3.0 ) / 2.0;
    epsRadius = sphere2radius( (dryMass / density) + (epsMass / epsDensity) );
    baby->epsMass = epsMass;
    baby->epsRadius = epsRadius;
  }

  // We want kid and mum directly adjacent with the same center as mum before reproduce,
  // both placed in a random direction
  distance = radius;

  if ((2*radius) > lengthZ) {
    printLog("Warning, for %d at %d ", process, now);
    printLogD("diam: %g > sliceThickness\n", 2*radius);
  }

  alpha = 2.0 * M_PI * [randSource evenRand];
  beta  = 2.0 * M_PI * [randSource evenRand];
  // shift kid
  baby->x      = x + ( distance * cos(alpha) * cos(beta) );
  baby->y      = y + ( distance * sin(alpha) * cos(beta) );
  baby->z      = z + ( distance * sin(beta)              );
  // now I move, too....
  x -= ( distance * cos(alpha) * cos(beta) );
  y -= ( distance * sin(alpha) * cos(beta) );
  z -= ( distance * sin(beta)              );

  // Out of bounds check
  if (biofilm) { // cells should not be within the carrier nor in the topmost layer which is always bulk liquid
    if (baby->x <= carrierLayer) {
      baby->x = carrierLayer + EPSILON;
      printLog("Warning, for %d at %d baby->x < cellSize\n", process, now);
    }
    if (baby->x >= (lengthX-cellSize)) {
      baby->x = lengthX - cellSize - EPSILON;
      printLog("Warning, for %d at %d baby->x > (lengthX-cellSize)\n", process, now);
    }
    if (x <= carrierLayer) {
      x = carrierLayer + EPSILON;
      printLog("Warning, for %d at %d x < cellSize\n", process, now);
    }
    if (x >= (lengthX-cellSize)) {
      x = lengthX - cellSize - EPSILON;
      printLog("Warning, for %d at %d x > (lengthX-cellSize)\n", process, now);
    }
  }
  if (periodicY) {
    if (baby->y < 0)       baby->y += lengthY;
    if (baby->y >= lengthY) baby->y -= lengthY;
    if (y < 0)       y += lengthY;
    if (y >= lengthY) y -= lengthY;
  }
  // periodicZ always true so far
  if (baby->z < 0)       baby->z += lengthZ;
  if (baby->z >= lengthZ) baby->z -= lengthZ;
  if (z < 0)       z += lengthZ;
  if (z >= lengthZ) z -= lengthZ;
  
  // hang on quad tree
  [baby getMcb: baby->myquad->mcr];
  myRoot = [self getRoot];
  [myRoot addNode: baby->myquad];

  [myquad growMcr];

  return 1;
}

-step {
  // One step in the lifecycle of a Bacillus.
  // Note: returns nil if dies--higher level routine deletes, because
  // higher level routine is traversing the tree.

  // print before everything else since most things are printed only when it's the first cell's turn
  //if ([geckoControl getCurrentTime] % [geckoControl getStatCycle] == 0) [self printv]; 

  [self grow];

  [self update];

  [self move]; // if this doesn't suffice, Diffuse will call doTheLocomotion
  
  if (![self checkDeath]) // death if radius < minsize
    return nil;
  
  [self checkVolume]; // are we big enough to trigger division?

  return self;
}

-printv {
  // most printing is done from Diffuse while the growth step is frozen, here we only print when we try things out
  // some things should be printed for all cells, some not
  // some things should be printed often, some not

  static int lastRound = -1;
  static int lastRun = -1;
  timeval_t now = [geckoControl getCurrentTime];
  int thisRun = [geckoControl getNumRun];

  // this is printed only for one cell per round
  if ((lastRound != now) || (lastRun != thisRun)) {
    // this is printed at lower frequency ([geckoControl getStatCycle])
    if (now % [geckoControl getStatCycle] == 0) {

    }
    
    // this is printed every printstep

    lastRound = now;
    lastRun = thisRun;
  }
  
  // this is printed for all cells
  
  // this is printed at lower frequency ([geckoControl getStatCycle])
  if (now % [geckoControl getStatCycle] == 0) {

  }
  return self;
}

-printSpheresMatlab {
  // Replacement for snapshots, prints all agents for import into Matlab or conversion into POV-Ray file
  char filename[256];
  FILE *sphereFile;
  static int lastRound = -1;
  static int lastRun = -1;
  timeval_t now = [geckoControl getCurrentTime];
  int thisRun = [geckoControl getNumRun];
  int timeunit =  [[geckoControl getModelSwarm] getStepSize];

  sprintf(filename, "%s_%03ld_%06ld.ml", [geckoControl getFileStem], thisRun, now);

  if ((lastRound != now) || (lastRun != thisRun)) { // do this only once at the beginning of each round
    sphereFile = fopen(filename, "w"); // overwrite, don't append
    fclose(sphereFile);
    lastRound = now;
    lastRun = thisRun;
  }

  sphereFile = fopen(filename, "a"); // append, don't overwrite

  fprintf(sphereFile, "%d\t", process); // numerical identifier for species
  fprintf(sphereFile, "%g\t", x);  // x coordinate of center
  fprintf(sphereFile, "%g\t", y);  // y coordinate of center
  fprintf(sphereFile, "%g\t", z);  // z coordinate of center
  fprintf(sphereFile, "%g\t", radius);       // radius
  fprintf(sphereFile, "%d\t", myGrandma);    // number of grandma
  fprintf(sphereFile, "%s\t", myName);       // myName, string of 0 and 1
  fprintf(sphereFile, "%g\t", growthRate / timeunit);   // growth rate
  fprintf(sphereFile, "%g\t", vmax / timeunit);         // vmax
  fprintf(sphereFile, "%g\n", epsRadius);    // epsRadius

  fclose(sphereFile);
  return self;
}

-printFed { // print the amount of substrate taken up, per substrate and species, useful for yields
  int i, j;

  printBac("%d\t%d\t", [geckoControl getNumRun], [geckoControl getCurrentTime]);

  for (i=0; i < numSpecies; i++)
    for (j=0; j < numSubs; j++)
      printBacD("%g\t", popFed[i][j]);
  
  printBac("\n");

  return self;
}

@end
