// vfb, redesign, 951221

#import <math.h>
#import <time.h>
#import "RandSource.h"
#import "Base.h"
#import "Colors.h"
#import "Stat.h" 
#import <simtools.h>

// Used in constructing agents' equivalent area squares or equivalent volume cubes in Mcr and Diffuse
GCoord SqrtPiOver2 = 0.886226925;
GCoord CubertPiOver6 = 0.805995977;
// Used in Diffuse for converting substrate concentrations from ionized to unionized
GRate pH = 7.0; // pH
GRate celsius = 30.0; // temperature in degrees celsius

char GeckoVersion[128] = "Gecko 0.6, new BacSim 0.2: multi-species, multi-substrate, 2.5D, ADI";
int arbColorMax = 20; // pref. not multiple of 3
// Note the highlight color in BLT graph is blue, making blue kinda useless
// for a trace.  Yellow is impossible to see...
char* arbColorStrings[20] = {
                   "snow3",
                   "turquoise1", 
                   "turquoise3", 
                   "aquamarine1", 
                   "SeaGreen1", 
                   "SeaGreen3", 
                   "DarkOliveGreen3", 
                   "LightGoldenrod2", 
                   "gold3", 
                   "chocolate3", 
                   "salmon1", 
                   "orange1", 
                   "firebrick1", 
                   "DeepPink1", 
                   "HotPink1", 
                   "pink1", 
                   "maroon1", 
                   "magenta1", 
                   "MediumOrchid2",
                   "MediumPurple1"
                   };

GCoord distance(GCoord x1, GCoord y1, GCoord x2, GCoord y2)
{
    GCoord a = x2 - x1;
    GCoord b = y2 - y1;
    return(sqrt(a*a + b*b));
}  

GCoord distance3D(GCoord x1, GCoord y1, GCoord z1, GCoord x2, GCoord y2, GCoord z2)
{
    GCoord a = x2 - x1;
    GCoord b = y2 - y1;
    GCoord c = z2 - z1;
    return( sqrt(a*a + b*b + c*c) );
}  

BOOL ptToward(GCoord fromx, GCoord fromy, GCoord tox, GCoord toy,
              GCoord dist, GCoord *ptx, GCoord *pty)
    // Returns false if from/to points are coincident or dist = 0, but still 
    // sets ptx and pty (to fromx, fromy).
    // But if it works, the purpose is to project a point at distance dist from
    // point fromx, fromy, in the direction of point tox, toy.  We don't care
    // whether we overshoot, that's just the direction to go in.  (Or opposite,
    // in the case of a negative dist.)
{
    
    GCoord wnorm = distance(fromx, fromy, tox, toy);
    *ptx = fromx;
    *pty = fromy;
    if (wnorm == 0.0)
        return 0;
    dist = dist / wnorm;
    *ptx += (tox - fromx) * dist;
    *pty += (toy - fromy) * dist;
    return 1;
}

BOOL ptToward3D(GCoord fromx, GCoord fromy, GCoord fromz, GCoord tox, GCoord toy, GCoord toz,
              GCoord dist, GCoord *ptx, GCoord *pty, GCoord *ptz)
    // Returns false if from/to points are coincident or dist = 0, but still 
    // sets ptx, pty and ptz (to fromx, fromy, fromz).
    // But if it works, the purpose is to project a point at distance dist from
    // point (fromx, fromy, fromz) in the direction of point (tox, toy, toz).  We don't care
    // whether we overshoot, that's just the direction to go in.  (Or opposite,
    // in the case of a negative dist.)
{
    
    GCoord wnorm = distance3D(fromx, fromy, fromz, tox, toy, toz);
    *ptx = fromx;
    *pty = fromy;
    *ptz = fromz;
    if (wnorm == 0.0)
        return 0;
    dist = dist / wnorm;
    *ptx += (tox - fromx) * dist;
    *pty += (toy - fromy) * dist;
    *ptz += (toz - fromz) * dist;
    return 1;
}

double angleToward(GCoord fromx, GCoord fromy, GCoord tox, GCoord toy)
    // angle of vector pointing this way
{
    GCoord wnorm;
    double angle = M_PI * 2.0 * [randSource evenRand];//random default
    GCoord dx = tox - fromx;
    GCoord dy = toy - fromy;
    
    wnorm = sqrt(dx * dx + dy * dy); // hypotenuse...
    if (wnorm > 0.0) {
      wnorm = fabs(dx / wnorm); // Ankathete / Hypotenuse
        if (wnorm > 1.0) 
            wnorm = 1.0;
        angle = acos(wnorm); // first quadrant angle...
        if (dx < 0.0) // second and third quadrant
            angle = M_PI - angle;
        if (dy < 0.0) // forth and third quadrant, result for third is aggregate, M_PI + angle
            angle = (2.0 * M_PI) - angle;
    } 
    return angle;
}

GArea radius2area(GCoord r)
{
    GArea result = M_PI * r * r;
    return result;
}

ResQuant radius2volume(GCoord r, GCoord scaler, GCoord scalev)
    // The volume of a critter is a (maybe) sphere,
    // and the volume of the sphere is (4/3)* pi * r**3.
    // Radius = (r according to volume) * scaler.
    // Scalev = %spherical volume (nonsphere).
{
    ResQuant result;
    // unscale r
    r = r / scaler;
    result = scalev * (4.0 / 3.0) * M_PI * r * r * r;
    return result;
}

GCoord volume2radius(ResQuant v, GCoord scaler, GCoord scalev)
    // The volume of a critter is a (maybe) sphere,
    // and the volume of the sphere is (4/3)* pi * r**3.
    // Radius = (r according to volume) * scaler.
    // Scalev = %spherical volume (nonsphere).
{
    GCoord result;
    result = (GCoord) pow( (v * 0.75 / (M_PI * scalev)), 0.333333 );
    result = scaler * result;
    return result;
}

ResQuant radius2sphere(GCoord r) {
  // The volume of the sphere is (4/3)* pi * r**3.
  ResQuant result = (4.0 / 3.0) * M_PI * r * r * r;
  return result;
}

GCoord sphere2radius(ResQuant v) {
  // The volume of the sphere is (4/3)* pi * r**3.
  GCoord result = (GCoord) pow( (v * 0.75 / M_PI), 0.333333333333333333 );
  return result;
}

char* timeString()
{
    // Courtesy Rick Riolo:
    const int TimeDateSz = 25;
    struct tm   *ts;
    time_t      bintime;
    static char  startDate[25];
    static char  startTime[25];
    static char  startBoth[25];

    time(&bintime);
    ts = localtime( &bintime );
    strftime( startDate, TimeDateSz-1, "%D", ts );
    strftime( startTime, TimeDateSz-1, "%H:%M", ts );
    sprintf(startBoth, "%s  %s", startDate, startTime);
    return startBoth;
}

char* colorName(int color)
{
    char* colorname;
    switch (color) {
        case 1: colorname = ColorName1; break;
        case 2: colorname = ColorName2; break;
        case 3: colorname = ColorName3; break;
        case 4: colorname = ColorName4; break;
        case 5: colorname = ColorName5; break;
        case 6: colorname = ColorName6; break;
        case 8: colorname = ColorName8; break;
        case 9: colorname = ColorName9; break;
        case 10:colorname = ColorName10; break;
        default:colorname = ColorName0; 
    }
    return colorname;
}
