// vfb, rewrite, 960116.
// vfb, orig, 951020:

#import <math.h>	
#import <objectbase.h>	
#import <simtools.h>	
#import "RandSource.h"
#import "GeckoControlSwarm.h"
#import "Grazer.h"
#import "Carnie.h"

@implementation Carnie

Carnie* recycleCarnies = nil;
Mcr* huntwin;
BOOL spidersReproducedFlag = 0;

-recycle
{
    [species tallyDeath: self];
    
    [species decrementCensus];
    hook            = recycleCarnies;
    recycleCarnies  = self;
    return nil;
}

-recycleClear
{
    Carnie* togo;
    while (recycleCarnies) {
        togo = recycleCarnies;
        recycleCarnies = recycleCarnies->hook;
        [togo->energy drop];
        [togo drop];
    }
    return self;
}

-reuse
{
    Carnie* result  = recycleCarnies;
    recycleCarnies  = recycleCarnies->hook;
    result->hook    = nil;
    return result;
}

-init
{
    double angle = 2 * M_PI * [randSource evenRand];
    [super init];
    smell_radius    = 1.0;
    sleeptime = 0;
    sleepleft = 0;
    last_ate        = [geckoControl getCurrentTime];
    hibernaterule   = 10000; // never, ish.
    hibernatetax    = 0.0;
    wannax = cos(angle);
    wannay = sin(angle);
    return self;
}

+create 
{
    id result;
    
    if (recycleCarnies)
        result = [recycleCarnies reuse];
    else {
        result = [Carnie createBegin: globalZone];
        [result allocEnergy];
    }

    [result init];
    return result;					
}

-addSpeciesParams
{
    [super addSpeciesParams];
    
    [species addParam: "SmellRadius" Vary: 0 DValue: smell_radius 
          Message: M(setSmellRadius:) Owner: self];

    [species addParam: "Sleep" Vary: 0 IValue: sleeptime
          Message: M(setSleep:) Owner: self];

    [species addParam: "HibRule" Vary: 0 IValue: hibernaterule 
          Message: M(setHibRule:) Owner: self];

    [species addParam: "HibTax" Vary: 0 DValue: hibernatetax 
          Message: M(setHibTax:) Owner: self];

    return self;   
}

-liveGrandma: (Species*) s
{
    // as per Hairston&Hairston -- forest lowest at .56
    [energy addConvertFrom: proteins To: cproteins Rate: .56];
    [energy addTax: cproteins Rate: .127 Exponent: .75]; // .01
    breedsize = 7.18; 
    kidsize = 4.04;
    minsize = 2.52;
    scaleradius  = 1.0;
    scalev  = 1.0/M_PI;

    smell_radius = 2.09;
    hibernatetax = .0072;
    sleeptime    = 6;	
    hibernaterule= 15;

    species = s;
    [self setGColor: [species getGColor]];
    [self addSpeciesParams];
    
    return self;
}

-resetSmellRadius: (GRate) sr Sleep: (int) s HibRule: (int) hr 
      HibTax: (GRate) ht
{
    smell_radius  = sr;
    hibernatetax  = ht;
    sleeptime     = s;	
    hibernaterule = hr; 
    return self;
}

-setSmellRadius: (GRate) sr {smell_radius  = sr;return self;}
-setSleep: (int) s {sleeptime     = s;	return self;}
-setHibRule: (int) hr {hibernaterule = hr; return self;}
-setHibTax: (GRate) ht {hibernatetax  = ht; return self;}
-setConvertRate: (GRate) nr
    // cheats rather than parse what was there before.
{
    [energy setConvertFrom: proteins To: cproteins Rate: nr];
    return self;
}

-copymom
    // create an agent matching mom in all heritable respects
{
    Carnie* baby;
    Quad* newQuad;

    baby = [Carnie create];
    baby->species   = species;
    [energy copyTo: baby->energy];
    [species incrementCensus];
    [species tallyBirth: self];
    baby->color     = color;
    baby->breedsize = breedsize;
    baby->kidsize   = kidsize;
    baby->minsize   = minsize;
    baby->scaleradius  = scaleradius;
    baby->scalev       = scalev;
    baby->smell_radius = smell_radius;
    baby->hibernatetax = hibernatetax;
    baby->hibernaterule = hibernaterule;
    baby->sleeptime = sleeptime;
    baby->sleepleft = 0;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setParent: nil];
    [newQuad setChild: nil];
    [baby setQuad: newQuad];

    return baby;
}

-(GCoord) getMoveRadius 
    // how far can I move? without hopping, too slow to react?
{ 
    return ( radius );
}

-(GCoord) getPerceiveRadius 
    // how far can I smell?
{ 
    return ( radius * smell_radius );
}

-(Critter*) findCritterR: (GCoord) distance 
            Attack: (BOOL) attack Root: (Quad*) huntroot
    // Look for prey--kinda arbitrary logic, so localize.
{
    Quad * other, * bestother;
    id odata, bdata;
    myquad->visible = 0;

    bestother = nil;
    
    // find first candidate        
    [huntwin circleToSquare: x Y: y Radius: distance];
    other = [huntroot pickWin: huntwin Root: huntroot Invisibles: 0];
    
    if (other) do {
        odata = other->data;
        if (attack) {
            if ([self attacks: odata]) {
                if ((bestother)) {
                    bdata = bestother->data;
                    if ([bdata getRadius] > [odata getRadius])
                        bestother = other;
                } else
                    bestother = other;
            }
        } else {
            if (![self attacks: odata]) {
                if ((bestother)) {
                    if ([randSource evenRand] > 0.5)
                        bestother = other;
                } else
                    bestother = other;
            }
        }
        
        other = [other pickWin: huntwin Root: huntroot Invisibles: 0];
    } while ((other));
    
    myquad->visible = 1;
    
    if (bestother) {
        return bestother->data;
    } else
        return nil;
}

-move
{
    GCoord scratchx, scratchy;
    GCoord delta = [self getMoveRadius];

    if (sleepleft > 0)
        return self;
    
    if (delta == 0.0)
        return self;
    
    [self getDirectionX: &scratchx Y: &scratchy]; // ignore, because reset wanna's

    x += fabs(delta) * wannax;
    y += fabs(delta) * wannay;

    [self getMcr: myquad->mcr];
    [myquad growMcr];

    return self;
}

-hunt
    // Pretty complex behavior for carnies.  Priorities:
    // 	1. try to eat (any gopper will do)
    // 	2. move away from a fellow Carnie.
    // 	3. move toward something to eat (as is, any gopper)
    // 	4. if not on a plant, move toward a plant
{
    ResQuant dinner;
    ResType meat = [energy getIeat];
    id plant;
    id gopper;
    id carnie;
    BOOL randommove = 1;
    Quad* mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
    Quad* plantRoot  = [[geckoControl getModelSwarm] getPlantRoot];
    
    // Can eat a gopper?
    gopper = [self findCritterR: radius Attack: 1 Root: mobileRoot]; 
    if ((gopper)) {
        dinner = [gopper stealAllRes: meat];
        [self feedRes: meat Amount: dinner];
        [self setDirectionX: [gopper getGX] - x 
                          Y: [gopper getGY] - y];
        [gopper step];
        sleepleft = sleeptime;  
        last_ate = [geckoControl getCurrentTime];
        color = CarnieDormantColor;
        randommove = 0;
    
    // away from same species?
    } else {
        carnie = [self findCritterR: [self getPerceiveRadius] 
                Attack: 0 Root: mobileRoot];
        if ( (carnie) && ([carnie getSpecies] == [self getSpecies]) ) {
            [ self setDirectionX: x - [carnie getGX]  
                               Y: y - [carnie getGY]];
            randommove = 0;

        // Can smell a gopper to head towards?
        } else {
            gopper = [self findCritterR: [self getPerceiveRadius] 
                        Attack: 1 Root: mobileRoot];
            if ((gopper)) {
                [self setDirectionX: [gopper getGX] - x 
                                  Y: [gopper getGY] - y]; // head thataway
                randommove = 0;


           // well, what else is in the neighborhood?
           // if I'm not on a plant, better move toward one
           } else {
                plant = [self findCritterR: radius Attack: 0 Root: plantRoot];
                if (!plant) {
                    plant = [self findCritterR: [self getPerceiveRadius] 
                            Attack: 0 Root: plantRoot];
                    if ((plant)) {
                        [self setDirectionX: [plant getGX] - x 
                                          Y: [plant getGY] - y];
                        randommove = 0;
                    }
                }
            }
        }
    } 
    return self;    
}

-(BOOL) hibernating
{
    if ( ([geckoControl getCurrentTime] - last_ate) > hibernaterule ) {
        color = CarnieDormantColor;
        return 1;
    }

/*
        if ( radius > kidsize ) {
            color = CarnieDormantColor;
            return 1;
        } else {
            return 0;
        }
    }
*/
    return 0;
}

-paytaxRate: (GRate) rate
{
    ResType me = [energy getIam];
    GRate savetax = [energy getTaxRate: me];
    GRate saveexp = [energy getTaxExponent: me];
    [energy replaceTax: me Rate: rate Exponent: saveexp]; 
    [super paytax];
    [energy replaceTax: me Rate: savetax Exponent: saveexp]; 
    return self;
}

-(BOOL) reproduce
{
    BOOL answer = [super reproduce];
    if (answer)
        spidersReproducedFlag = 1;
    return answer;    
}

-step 
    // Override Critter-step because of sleep/wakeup and plantcycle.
{
    Critter* other;
    Quad* mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
    int carniecycle = [[geckoControl getModelSwarm] getCarniecycle];
    
    if (([geckoControl getCurrentTime] % carniecycle) != 0)
        return self;

    // sleep step
    if ( (sleepleft > 0) || ([self hibernating]) ) {

        // first check if gopper stepped on me and woke me up.
        other = [self findCritterR: [self getPerceiveRadius] 
                 Attack: 1 Root: mobileRoot];
        
        if ((other)) {
            while (sleepleft > 0) {
                if (![self paytax])
                    return nil;
                sleepleft--;
            }
            // This isn't true, but makes me stop hibernating.
            last_ate = [geckoControl getCurrentTime];
        } else {
            if ([self hibernating]) {
                [self paytaxRate: hibernatetax];
                return self;
            } else {
                [self paytax];
                sleepleft--;
                return self;
            }
        }
    }
    
    // awake step
    color = CarnieColor;

    [super step];
    
    [self hibernating];
    
    return self;
}

-printv
{
    [super printv];
    printLogD("smell_radius = %7f, ", smell_radius);
    printLog("sleeptime = %d, hibrule = %d, ", sleeptime, hibernaterule);
    printLogD("hibernatetax = %7f\n\n", hibernatetax);
    return self;
}

@end
