// vfb, orig, 951020:
// 	mockup only, for testing Quads

#import "RandSource.h"
#import "Critter.h"
#import "FeedQuad.h"
#import "GeckoControlSwarm.h"
#import "Stat.h"
#import <stdlib.h>
#import <math.h>
#import <objectbase.h>
#import <simtools.h>

@implementation Critter

Critter* recycleCritters = 0;

-recycle
{
    [species tallyDeath: self];

    hook = recycleCritters;
    recycleCritters = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    Critter* togo;
    while (recycleCritters) {
        togo = recycleCritters;
        recycleCritters = recycleCritters->hook;
        [togo drop];
    }
    return self;
}

-reuse
{
    Critter* result = recycleCritters;
    recycleCritters = recycleCritters->hook;
    result->hook    = nil;
    return result;
}

-init
{
  z = 0.0;
  wannax 	= 0.0;
  wannay 	= 0.0;
  wannaz 	= 0.0;
  birthday 	= [geckoControl getCurrentTime];
  kidcount 	= 0;
  return self;
}

+create 
{
    id result;
    
    if (recycleCritters) {
        result = [recycleCritters reuse];
    } else {
        result = [Critter createBegin: globalZone];
    }
    [result init];
    
    return result;					
}

-copymom
    // create an agent matching mom in all heritable respects
{
    Critter* baby;
    Quad* newQuad;

    baby = [Critter create];
    baby->species   = species;
    baby->color     = color;

    [species incrementCensus];
    [species tallyBirth: self];
    
    baby->breedsize   = breedsize;
    baby->kidsize     = kidsize;
    baby->minsize     = minsize;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
}

-liveGrandma: (Species*) s
{
    breedsize = 11.13;  // orig 11
    kidsize = 2.95;
    minsize = 3.0;
    
    species = s;
    [self setGColor: [species getGColor]];
    
    return self;
}

-setX: (GCoord) cx Y: (GCoord) cy Radius: (GCoord) r {
  x = cx;
  y = cy;
  radius = r;
  [self getMcr: myquad->mcr];
  return self;
}

-setX: (GCoord) cx Y: (GCoord) cy Z: (GCoord) cz Radius: (GCoord) r {
  x = cx;
  y = cy;
  z = cz;
  radius = r;
  [self getMcb: myquad->mcr];
  return self;
}

-getX: (GCoord*) cx Y: (GCoord*) cy Radius: (GCoord*) r
{ 
    *cx = x; 
    *cy = y; 
    *r  = radius; 
    return self; 
}

-getX: (GCoord*) cx Y: (GCoord*) cy Z: (GCoord*) cz Radius: (GCoord*) r
{ 
    *cx = x; 
    *cy = y; 
    *cz = z; 
    *r  = radius; 
    return self; 
}

-getX: (GCoord*) cx Y: (GCoord*) cy { *cx = x; *cy = y; return self; }
-getX: (GCoord*) cx Y: (GCoord*) cy Z: (GCoord*) cz { *cx = x; *cy = y; *cz = z; return self; }

-(GCoord) getX {return x;}
-(GCoord) getY {return y;}
-(GCoord) getZ {return z;}
-(GCoord) getGX {return x;}
-(GCoord) getGY {return y;}
-(GCoord) getGZ {return z;}

-setGColor: (int) colorid { color = colorid; return self; }
	
-(const char*) getInstanceName { return [species getParamName]; } 
-(char*) getName { return [species getParamName]; } 

-setAgentsFromFile: (char*) filename { return self; } // no-op

// Diffuse and GeckoModelSwarm wants to know, by default we return -2
-(int) getProcess { return -2; }

-(EpsType) getEps { return NoEps; } // only eps producers return in the affirmative

-(GRate) doTheLocomotion { return -1.0; } // no-op

-(int) getGColor { return color; }	

-setQuad: (Quad*) q { myquad = q; return self; }

-(GCoord) getRadius { return radius; }
-(GCoord) getBreedRadius { return breedsize; }
-(GCoord) getMinRadius { return minsize; }
-(GCoord) getKidRadius { return kidsize; }

-(Species*) getSpecies { return species; }
	
// stat support--eminently overrideable

-statsFinalize { return self; }

-(int) getAgeBucket {
printf("critter-getagebucket\n");
    return ([geckoControl getCurrentTime] - birthday) / [geckoControl getStatCycle];
}
-(int) getDefaultAgeCount { return 5; }
-(int) getStageBucket
{
    if (kidcount > 0)
        return 2;
    return 1;
}
-(int) getDefaultStages { return 3;}
-(int) getSizeBucket {
    ResQuant v, smallv;
    GCoord smallr, mediumr;

    smallv = radius2sphere([self getMinRadius]);
    smallv += radius2sphere([self getKidRadius]);
    smallr = sphere2radius(smallv);
    
    v = radius2sphere([self getBreedRadius]);
    v -= smallv;
    mediumr = sphere2radius(v);
    
    if (radius <= smallr)
        return 0;
    if (radius <= mediumr)
        return 1;
    return 2;
}
-(int) getDefaultSizes { return 3;}

-calcSpecific: (id) specStats
    // calls BucketDistrib calc by default.
{
    [specStats calc];
    return self;
}
-(BOOL) observeSpecific: (id) specStats
    // false--take default.
{ return 0; }

-(int) getSpecificNumBuckets { return 1;}
-(int) getSpecificBucket { return 0;}
-(int) getSpecificCounter: (int) anEvent { return -1;}
-(int) getSpecificSum: (int) anEvent { return -1;}
-(double) getSpecificAmount: (int) anEvent { return 0;}
-(int) getSpecificCounts { return 1;}
-(int) getSpecificSums { return 1;}
-(int) getSpecificDeriveds { return 1;}

-(GCoord) getPerceiveRadius { return radius; }

-getDirectionX: (GCoord*) dx Y: (GCoord*) dy
    // convert wannas to unit vector pointing same direction
{
    ptToward(x, y, x + wannax, y + wannay, 1.0, dx, dy);
    *dx -= x;
    *dy -= y;
    wannax = *dx;
    wannay = *dy;
    return self;
}

-getDirectionX: (GCoord*) dx Y: (GCoord*) dy Z: (GCoord*) dz
    // convert wannas to unit vector pointing same direction
{
  ptToward3D(x, y, z, x + wannax, y + wannay, z + wannaz, 1.0, dx, dy, dz); // in Base.m
  *dx -= x;
  *dy -= y;
  *dz -= z;
  wannax = *dx;
  wannay = *dy;
  wannaz = *dz;
  return self;
}

-(double) getDirectionAngle
    // convert wannas to unit vector pointing same direction
{
    ResQuant wnorm;
    double angle = M_PI * 2.0 * [randSource evenRand];
    
    wnorm = sqrt(wannax * wannax + wannay * wannay); // hypotenuse...
    if (wnorm > 0.0) {
        wnorm = fabs(wannax / wnorm);
        if (wnorm > 1.0) 
            wnorm = 1.0;
        angle = acos(wnorm);// first quadrant angle...
        if (wannax < 0.0) 
            angle = M_PI - angle;
        if (wannay < 0.0)
            angle = (2.0 * M_PI) - angle;
        wannax = cos(angle);
        wannay = sin(angle);
    } 
    return angle;
}

-setDirectionX: (GCoord) dx Y: (GCoord) dy
{
    wannax = dx; wannay = dy; return self;
}

-setDirectionX: (GCoord) dx Y: (GCoord) dy Z: (GCoord) dz
{
    wannax = dx; wannay = dy; wannaz = dz; return self;
}

-addDirectionX: (GCoord) dx Y: (GCoord) dy Weight: (double) weight
    // convert dx, dy to unit vector before weighing in
{
    GCoord ux, uy;
    ptToward(x, y, x+dx, y+dy, 1.0, &ux, &uy);
    ux -= x;
    uy -= y;
    wannax += ux * weight; 
    wannay += uy * weight; 
    return self;
}

-addDirectionX: (GCoord) dx Y: (GCoord) dy Z: (GCoord) dz Weight: (double) weight
    // convert dx, dy, dz to unit vector before weighing in
{
    GCoord ux, uy, uz;
    ptToward3D(x, y, z, x+dx, y+dy, z+dz, 1.0, &ux, &uy, &uz);
    ux -= x;
    uy -= y;
    uz -= z;
    wannax += ux * weight; 
    wannay += uy * weight; 
    wannaz += uz * weight; 
    return self;
}

-getQuad {return myquad;}		

-(GArea) getArea { return (M_PI * radius * radius); }		

-(BOOL) hiddenExposed: (Mcr*) exposed 
{
    [exposed setMcr: myquad->mcr]; 
    return 0;
}

-(void) getMcr: (Mcr*) m
{
    m->ux = x + radius;
    m->uy = y + radius;
    m->lx = x - radius;
    m->ly = y - radius;
}

-(void) getMcb: (Mcr*) m
{
    m->ux = x + radius;
    m->uy = y + radius;
    m->uz = z + radius;
    m->lx = x - radius;
    m->ly = y - radius;
    m->lz = z - radius;
}

-(ResQuant) getVolume {return radius2sphere(radius);}

-(Quad*) getRoot
{
    Quad* myRoot;
    myRoot = myquad->parent;
    while ((myRoot->type != sessile) && (myRoot->type != mobile))
        myRoot = myRoot->parent;
    return myRoot;
}

-print
{
    printLog("%s at (", [species getParamName]);
    printLogD("%3f,", x);
    printLogD("%3f,", y);
    printLogD("%3f),", z);
    printLogD(" size %2f", radius);
    return self;
}

-println
{
    [self print]; 
    printLog("\n");
    return self;
}

-printv
{
    [self println];
    printLogD("breedsize = %3f, ", breedsize);
    printLogD("kidsize = %3f, ", kidsize);
    printLogD("minsize = %3f, ", minsize);
    printLog("birthday = %d, kidcount = %d\n", birthday, kidcount);
    printLog("\n");
    return self;
}

@end
