// vfb, 961015, orig. Puffball

#import "random.h"
#import "Animal.h"
#import "CritterZoo.h"
#import "LandRoot.h"
#import "GeckoControlSwarm.h"
#import "Stat.h"
#import "RandSource.h"
#import <stdlib.h>
#import <math.h>
#import <objectbase.h>
#import <simtools.h>

@implementation Puffball

Puffball* recyclePuffballs = 0;

-init
{
    [super init];
    color   = LastCritterColor;
    kidcost = 1.0;
    
    return self;
}

-recycle
{
    [species tallyDeath: self];

    hook = recyclePuffballs;
    recyclePuffballs = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    Puffball* togo;
    while (recyclePuffballs) {
        togo = recyclePuffballs;
        recyclePuffballs = recyclePuffballs->hook;
        [togo->energy drop];
        [togo drop];
    }
    return self;
}

-reuse
{
    Puffball* result = recyclePuffballs;
    recyclePuffballs = recyclePuffballs->hook;
    result->hook    = nil;
    return result;
}

+create 
{
    id result;
    
    if (recyclePuffballs) {
        result = [recyclePuffballs reuse];
    } else {
        result = [Puffball createBegin: globalZone];
        [result allocEnergy];
    }
    [result init];
    
    return result;					
}

-copymom
    // create an agent matching mom in all heritable respects
{
    Puffball* baby;
    Quad* newQuad;

    baby = [Puffball create];
    baby->species   = species;
    [energy copyTo: baby->energy];
    baby->color     = color;

    [species incrementCensus];
    [species tallyBirth: self];
    
    baby->breedsize = breedsize;
    baby->kidsize   = kidsize;
    baby->minsize   = minsize;
    baby->kidcost   = kidcost;
    baby->scaleradius = scaleradius;
    baby->scalev = scalev;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
}

-liveGrandma: (Species*) s
    // initialize/reinit a default puffball
{
    [energy addConvertFrom: sunwater To: carbos Rate: .1];
    [energy addTax: carbos Rate: .0480 Exponent: .75];
    breedsize = 11.13;  // orig 11
    kidsize = 2.95;
    minsize = 3.0;
    scaleradius = 1.0;
    scalev = 1.0/M_PI;
    kidcost = 2.0;
    
    species = s;
    [self setGColor: [species getGColor]];

    [self addSpeciesParams];
    
    return self;
}

-addSpeciesParams
{
    [super addSpeciesParams];
    [species addParam: "KidCost" Vary: 0 DValue: kidcost
        Message: M(setKidCost:) Owner: self];
    return self;   
}

-(GCoord) getChildRadius
    // distance at which to place kid
{
    GCoord distance;
    distance = kidsize; // min
    distance += 2.0 * breedsize * exp(1.0) * [randSource evenRand];
    return(distance);
}
-setKidCost: (GCoord) kc {kidcost=kc; return self;}
-(GCoord) getKidCost {return kidcost;}

-(BOOL) reproduce
    // false if didn't -choose- to.
{
    Puffball* baby;
    GCoord distance;
    ResList* kidgets, *kidcosts;
    double angle;
    Quad* myRoot;
    ResQuant myvol, kidvol;
    int kidi;
    
    if (radius < breedsize)
        return 0;

    kidgets  = [self kidshare];
    kidcosts = [self kidshare];
    kidcosts->list[[energy getIam]] *= kidcost;
    
    myvol = [energy getStores: [energy getIam]];
    kidvol = kidgets->list[[energy getIam]];
    kidcount = myvol / (kidvol * kidcost);
    firstbornon = [geckoControl getCurrentTime];		// not that it matters....

    for (kidi = 0; kidi < kidcount; kidi++) {
        // make a kid
        angle = 2.0 * M_PI * [randSource evenRand];
        baby         = [self copymom];
        distance     = [self getChildRadius];
        baby->x      = x + ( distance * cos(angle) );
        baby->y      = y + ( distance * sin(angle) );
        baby->wannax = cos(angle);
        baby->wannay = sin(angle);
        baby->radius = kidsize;
    
        // transfer resources 
        kidgets = [self kidshare];
        [energy debit: kidcosts];
        [self resetRadius];
        baby->energy = [baby->energy setStores:  kidgets];
    
        // hang on quad tree
        [baby getMcr: baby->myquad->mcr];
        myRoot = myquad->parent;
        while ((myRoot->type != sessile) && (myRoot->type != mobile))
            myRoot = myRoot->parent;
        [myRoot addNode: baby->myquad];
    }

    [kidgets drop];
    [kidcosts drop];
    return 1;
}

-printv
{
    [self println];
    printLogD("breedsize = %3f, ", breedsize);
    printLogD("kidsize = %3f, ", kidsize);
    printLogD("minsize = %3f, ", minsize);
    printLogD("scaler = %3f, ",scaleradius);
    printLogD("scalev = %3f, ",scalev);
    printLogD("kidcost = %3f\n",kidcost);
    printLog("birthday = %d, firstbornon = %d, kidcount = %d\n",
        birthday, firstbornon, kidcount);
    [energy printv];
    printLog("\n");
    return self;
}

@end // Puffball

@implementation TimeLimitCritter

TimeLimitCritter* recycleTimeLimitCritters = 0;

-init
{
    [super init];
    color     = LastCritterColor;
    adultTime = 100;
    
    return self;
}

-recycle
{
    [species tallyDeath: self];

    hook = recycleTimeLimitCritters;
    recycleTimeLimitCritters = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    TimeLimitCritter* togo;
    while (recycleTimeLimitCritters) {
        togo = recycleTimeLimitCritters;
        recycleTimeLimitCritters = recycleTimeLimitCritters->hook;
        [togo->energy drop];
        [togo drop];
    }
    return self;
}

-reuse
{
    TimeLimitCritter* result = recycleTimeLimitCritters;
    recycleTimeLimitCritters = recycleTimeLimitCritters->hook;
    result->hook    = nil;
    return result;
}

+create 
{
    id result;
    
    if (recycleTimeLimitCritters) {
        result = [recycleTimeLimitCritters reuse];
    } else {
        result = [TimeLimitCritter createBegin: globalZone];
        [result allocEnergy];
    }
    [result init];
    
    return result;					
}

-copymom
    // create an agent matching mom in all heritable respects
{
    TimeLimitCritter* baby;
    Quad* newQuad;

    baby = [TimeLimitCritter create];
    baby->species   = species;
    [energy copyTo: baby->energy];
    baby->color     = color;

    [species incrementCensus];
    [species tallyBirth: self];
    
    baby->breedsize = breedsize;
    baby->kidsize   = kidsize;
    baby->minsize   = minsize;
    baby->scaleradius = scaleradius;
    baby->scalev = scalev;
    baby->adultTime = adultTime;
    baby->maxAge    = maxAge;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
}

-liveGrandma: (Species*) s
    // initialize/reinit a default TimeLimitCritter
{
    [energy addConvertFrom: sunwater To: carbos Rate: .1];
    [energy addTax: carbos Rate: .0480 Exponent: .75];
    breedsize   = 100000.0;  // until adultTime
    kidsize     = 2.95;
    minsize     = 3.0;
    scaleradius = 1.0;
    scalev = 1.0/M_PI;
    adultTime   = 150;	     // comparable to br 11.0
    maxAge      = 0;
    
    species = s;
    [self setGColor: [species getGColor]];

    [self addSpeciesParams];
    
    return self;
}

-addSpeciesParams
{
    [super addSpeciesParams];
    [species addParam: "AdultTime" Vary: 0 IValue: adultTime
        Message: M(setAdultTime:) Owner: self];
    [species addParam: "MaxAge" Vary: 0 IValue: maxAge
          Message: M(setMaxAge:) Owner: self];
    return self;   
}

-setAdultTime: (int) at {adultTime = at; return self;}
-(int) getAdultTime { return adultTime; }
-setMaxAge: (int) ma { maxAge = ma; return self;}
-(int) getMaxAge { return maxAge;}

-(BOOL) reproduce
    // False if didn't.  For now, select random angle for kid.
{
    TimeLimitCritter* baby;
    GCoord distance;
    ResList* kidgets;
    double angle;
    Quad* myRoot;
    
    if ([geckoControl getCurrentTime] == (adultTime + birthday)) {
        breedsize = radius;
    } else if ([geckoControl getCurrentTime] < (adultTime + birthday)) {
        return 0;
    }
    
    if (radius < breedsize)
        return 0;

    angle = 2.0 * M_PI * [randSource evenRand];

    // record a birth
    if (kidcount == 0)
        firstbornon = [geckoControl getCurrentTime];
    kidcount++;

    // make the kid
    baby         = [self copymom];
    distance     = [self getChildRadius];
    baby->x      = x + ( distance * cos(angle) );
    baby->y      = y + ( distance * sin(angle) );
    baby->wannax = cos(angle);
    baby->wannay = sin(angle);
    baby->radius = kidsize;

    // transfer resources 
    kidgets = [self kidshare];
    [energy debit: kidgets];
    [self resetRadius];
    baby->energy = [baby->energy setStores:  kidgets];

    // hang on quad tree
    [baby getMcr: baby->myquad->mcr];
    myRoot = myquad->parent;
    while ((myRoot->type != sessile) && (myRoot->type != mobile))
        myRoot = myRoot->parent;
    [myRoot addNode: baby->myquad];

    [kidgets drop];
    return 1;
}

-maxAgeCheck
{
    if (maxAge > 0) {
        if ((birthday + maxAge) <= [geckoControl getCurrentTime]) {
            [self die];
            return 0;
        }
    }
    return self;
}

-step 
    // One step in the lifecycle of a TimeLimitCritter.
    // Note: returns nil if dies--higher level routine deletes, because
    // higher level routine is traversing the tree.
{
    [self metabolize];
 
    if (![self paytax]) 
        return nil;
    if (![self maxAgeCheck])
        return nil;
    if (![self reproduce]) {
        [self hunt];
    }
    [self move];

    return self;
}

-printv
{
    [self println];
    printLogD("breedsize = %3f, ", breedsize);
    printLogD("kidsize = %3f, ", kidsize);
    printLogD("minsize = %3f, ", minsize);
    printLogD("scaler = %3f, ",scaleradius);
    printLogD("scalev = %3f, ",scalev);
    printLog("adultTime = %3d, ",adultTime);
    printLog("maxAge = %3d\n",maxAge);
    printLog("birthday = %d, firstbornon = %d, kidcount = %d\n",
        birthday, firstbornon, kidcount);
    [energy printv];
    printLog("\n");
    return self;
}

@end // TimeLimitCritter

@implementation NoReverseTLC

NoReverseTLC* recycleNoReverseTLCs = 0;

-init
{
    [super init];
    color      = LastCritterColor;
    structural = 0.75;
    
    return self;
}

-recycle
{
    [species tallyDeath: self];

    hook = recycleNoReverseTLCs;
    recycleNoReverseTLCs = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    NoReverseTLC* togo;
    while (recycleNoReverseTLCs) {
        togo = recycleNoReverseTLCs;
        recycleNoReverseTLCs = recycleNoReverseTLCs->hook;
        [togo->energy drop];
        [togo drop];
    }
    return self;
}

-reuse
{
    NoReverseTLC* result = recycleNoReverseTLCs;
    recycleNoReverseTLCs = recycleNoReverseTLCs->hook;
    result->hook    = nil;
    return result;
}

+create 
{
    id result;
    
    if (recycleNoReverseTLCs) {
        result = [recycleNoReverseTLCs reuse];
    } else {
        result = [NoReverseTLC createBegin: globalZone];
        [result allocEnergy];
    }
    [result init];
    
    return result;					
}

-copymom
    // create an agent matching mom in all heritable respects
{
    NoReverseTLC* baby;
    Quad* newQuad;

    baby = [NoReverseTLC create];
    baby->species   = species;
    [energy copyTo: baby->energy];
    baby->color     = color;

    [species incrementCensus];
    [species tallyBirth: self];
    
    baby->breedsize = breedsize;
    baby->kidsize   = kidsize;
    baby->minsize   = minsize;
    baby->adultTime = adultTime;
    baby->structural = structural;
    baby->scaleradius = scaleradius;
    baby->scalev = scalev;
    baby->curmin    = minsize;

    newQuad = [Quad create];
    [newQuad setData: baby Type: critter];
    [newQuad setSibling: nil];
    [newQuad setChild: nil];
    [newQuad setParent: nil];
    [baby setQuad: newQuad];

    return baby;
}

-liveGrandma: (Species*) s
    // initialize/reinit a default NoReverseTLC
{
    [energy addConvertFrom: sunwater To: carbos Rate: .1];
    [energy addTax: carbos Rate: .0480 Exponent: .75];
    breedsize   = 100000.0;  // until adultTime
    kidsize     = 2.95;
    minsize     = 3.0;
    curmin      = minsize;
    scaleradius = 1.0;
    scalev = 1.0/M_PI;
    adultTime   = 150;	     // comparable to br 11.0
    structural  = 0.75;
    
    species = s;
    [self setGColor: [species getGColor]];

    [self addSpeciesParams];
    
    return self;
}

-addSpeciesParams
{
    [super addSpeciesParams];
    [species addParam: "Structural" Vary: 0 DValue: structural
        Message: M(setStructural:) Owner: self];
    return self;   
}

-setStructural: (double) st {structural = st; return self;}
-(double) getStructural { return structural; }

-printv
{
    [self println];
    printLogD("breedsize = %3f, ", breedsize);
    printLogD("kidsize = %3f, ", kidsize);
    printLogD("minsize = %3f, ", minsize);
    printLogD("scaler = %3f, ",scaleradius);
    printLogD("scalev = %3f, ",scalev);
    printLog("adultTime = %3d, ",adultTime);
    printLogD("structural = %3f\n",structural);
    printLog("birthday = %d, firstbornon = %d, kidcount = %d\n",
        birthday, firstbornon, kidcount);
    [energy printv];
    printLog("\n");
    return self;
}

-paytax
{
    ResQuant netgrowth, curvol;
    GCoord newmin;
    
    LandRoot* landRoot = [[geckoControl getModelSwarm] getLandRoot];
    ResList* taxes = [energy taxes];
    [self resetRadius];
    
    curvol = netgrowth = [energy getStores: [energy getIam]];	// current
    netgrowth -= radius2volume(lastr, scaleradius, scalev);	// last
    [species tallyGrowth: self Amount: netgrowth];

    if (netgrowth > 0) {
        curvol *= structural;	// new minimum volume
        newmin = volume2radius(curvol, scaleradius, scalev);
        if (newmin > curmin) {
            curmin = newmin;
        }
    }
    
    [landRoot payTaxX: x Y: y List: taxes];

    [taxes drop];

    if (radius < curmin) {
        [self die]; 
        return nil;
    } else {
        return self;
    }       
}

@end // NoReverseTLC

