// Copyright (C) 1997- 1999 Jan Kreft <Kreft@cardiff.ac.uk>

// Redesign of Diffusion, using ADI diffusion-reaction algorithm
// Only used by bacteria, other critters are fed by the land or eat critters.
// We have lattices for substrates, reactions, biomasses, etc.
// Each cycle, diffusion steps first, then tells Bacilli to move, then Bacilli step
// The printing is done in between to insure synchrony

#include <objectbase.h>				
#include <collections.h>
#import "Base.h"
#import "Mcr.h"
#import "Quad.h"
#import "Res.h"
#import "Colors.h"
#import "Stat.h"
#import "Site.h"
#import "Bacillus.h"

@interface DiffuseQuad : ResourceQuad {
  // defining variables

  int numSubs; // number of substrates
  int numSpecies; // number of species

  // we need one var per substrate (numSubs)
  GRate*  extConcn; // array of subConcns in the bulk liquid (external), doesn't change, open system
  GRate  extConcn0; // set by Protocol
  GRate  extConcn1;
  GRate  extConcn2;
  GRate  extConcn3;
  GRate*  diffC;    // array of diffCs
  GRate  diffC0;    // set by Protocol
  GRate  diffC1;
  GRate  diffC2;
  GRate  diffC3;
  
  GCoord cellSize;      // only square/cubic grid, size of one grid element
  int	 cellsX;
  int	 cellsY;
  int    boundaryLayer; // the width of the hydrodynamic boundary layer in microns
  int    carrierLayer;  // the width of the carrier layer in microns
  GCoord sliceThickness;// the width of the third dimension in microns (third dimension always periodic)

  BOOL   haveInitMcr;	        // must be false as not currently in use
  BOOL   plate;                 // plate geometry, bottom with carrier, top bulk liquid, for biofilms
    
  int    numColors;		// how many colors used to display density
  GRate  highColorDensity;	// for density >= this, use high color
  GRate  lowColorDensity;	// for density < this, use low color
  
  // working variables

  int front;        // front of the biomass
  int countDiff;    // reset before each cycle
  int totDiffSteps; // total number of diffusion steps made this run

  // normalized for ADI, we need one var per substrate (numSubs)
  GRate* D;
  GRate* Ar;
  GRate* dt;
  // for conversion
  GRate* max;
  GRate* maxlast;
  // arguments for tridag, reuseable, hence only one set in total
  GRate* aa;
  GRate* bb;
  GRate* cc;
  GRate* rr;
  GRate* uu;
  int jj;
  GRate* gam; // internal to tridag
  GRate* hh;  // internal to tridagPeriodic
  GRate* vv;  // internal to tridagPeriodic

  BOOL   relaxed;      // has substrate field relaxed to the steady state?
  BOOL*  relaxedSub;   // for individual substrates
  BOOL   stable;       // is the diffusion algorithm stable?

  // matrices
  GRate*** s;		// substrate concn matrix, one per substrate (numSubs)
  GRate*** d;           // effective diffusion coefficient matrix, one per substrate (numSubs)
  GRate*** r;           // reaction rate matrix, one per substrate (numSubs)
  GRate*** x;           // biomass matrix, one per species (numSpecies)
  GRate*** g;           // growth rate matrix, one per species (numSpecies)
  GRate**  xtot;        // total biomass matrix, all species and epsMass
  GRate**  xbac;        // total biomass matrix, all species
  GRate**  xeps;        // total epsMass matrix
  GRate**  por;         // porosity matrix
  GRate**  tmp;	        // temporary buffer matrix for s, only one needed
  GRate**  delta;	// save off differences between s step n and s step n+1
  // these two could be combined but would be a lot of work, I mean a lot
  char**  c;            // carrier = 2, biofilm = 1, all liquid = 0
  char**  cs;           // carrier = 2, (biofilm + boundary layer) = 1, bulk liquid = 0

  GRate* colorDensity;	// density->color lookup map
  int*   colorIds;	        // color id lookup map
  GCoord cellVol;		// volume of lattice cell

  id bac1List; // create list of agents on the fly for each step for faster access
  id bac2List; // one list for each type of agent, saves superfluous questions
  id epsList;
}
// messages are explained in the .m file

-init;
-setHaveInitMcr;
+create;
-addSiteParams;
-landGrandma: (Site*) site;
-(GRate**) allocFloatLattice;
-(char**) allocCharLattice;
-fillLattice: (GRate**) which WithFloats: (GRate) fc;
-fillLattice: (char**) which WithChars: (char) fv;
-printFloatLattice: (GRate**) which;
-printCharLattice: (char**) which;
-initColors;
-(int) getColorCellX: (int) xi Y: (int) yi;
-(int) getColorDensity: (GRate) den;
-copymom;
-copyTo: (DiffuseQuad*) newbie;
-(GCoord) getCarrierLayer;
-(BOOL) getPeriodicY;
-(GCoord) getLengthX;
-(GCoord) getLengthY;
-(GCoord) getLengthZ;
-(GCoord) getCellSize;
-(int) getNumSubs;
-(int) getNumSpecies;
-(GCoord) getOverlapWithCarrierAtX: (GCoord) xi Y: (GCoord) yi;
-(GRate) getExtConcn: (BacillusSub) which;
-(GRate) getConcnSub: (BacillusSub) this X: (GCoord) xi Y: (GCoord) yi;
-doleOut: (GRate) what To: (GRate**) latt X: (GCoord) xpos Y: (GCoord) ypos Radius: (GCoord) radius;
-(GRate) avgSub: (BacillusSub) this X: (GCoord) xpos Y: (GCoord) ypos Radius: (GCoord) radius;
-(GRate) calcPorosity;
-update_x_Matrices;
-update_c_Matrix;
-(int) findFront;
-update_cs_Matrix;
-update_d_Matrices; 
-update_r_Matrices;
-update_r_MatricesList;
-update_r_Matrix: (int) k;
-tridag;
-tridagPeriodic;
-recalcParams;
-adi: (int) k;
-adiPlate: (int) k;
-stepResources;
-setNumSubs: (int) i;
-setNumSpecies: (int) i;
-setCellSize: (GCoord) xi;
-setCellsX: (int) i;
-setCellsY: (int) i;
-(int) getCellsX;
-(int) getCellsY;
-setPlate: (int) p;
-setBoundaryLayer: (int) b;
-setCarrierLayer: (int) cl;
-setSliceThickness: (GCoord) st;
-setDiffC0: (GRate) xi;
-setDiffC1: (GRate) xi;
-setDiffC2: (GRate) xi;
-setDiffC3: (GRate) xi;
-setExtConcn0: (GRate) xi;
-setExtConcn1: (GRate) xi;
-setExtConcn2: (GRate) xi;
-setExtConcn3: (GRate) xi;
-setNumColors: (int) i;
-setHighColorDensity: (GRate) xi;
-setLowColorDensity: (GRate) xi;
-(GCoord) getMcrLXCellX: (int) xi Y: (int) yi;
-(GCoord) getMcrLYCellX: (int) xi Y: (int) yi;
-(GCoord) getMcrUXCellX: (int) xi Y: (int) yi;
-(GCoord) getMcrUYCellX: (int) xi Y: (int) yi;
-printv;
-printColorThresholds;
-printLatticeHet;
-printSubConcnHet;
-debug;

@end
