// Copyright (C) 1997 - 1999 Jan Kreft <Kreft@cardiff.ac.uk>

#import <objectbase.h>
#import <simtools.h>
#import "RandSource.h"
#import "Critter.h"
#import "Eps.h"
#import "LandRoot.h"
#import "GeckoControlSwarm.h"
#import "Stat.h"
#import "ResourceQuad.h"
#import "Diffuse.h"
#import "Base.h"

@implementation Eps

// recycle
Eps* recycleEpss = 0;

static LandRoot* landRoot;
static DiffuseQuad* actualSite;
static Quad* mobileRoot;
static Mcr* landMcr;

// global parameters valid for all Eps blobs, set by Protocol!
static GRate epsDensity;                // the density of the EPS (epsDryMass)
static GRate convertBack;		// when die, multiplier on mass->substrate
static GRate decayR;		        // rate of EPS decay

// geometry info we get from DiffuseQuad, set via Protocol
static BOOL   periodicY; // if we have a plate, we have periodicY boundaries
static BOOL   biofilm;   // should be set from protocol via modified "Plate" directive, make plate and periodicY independent
static GCoord carrierLayer; // height of the carrier
static GCoord lengthX;   // length of the non-periodic coordinate
static GCoord lengthY;   // length of the periodic coordinate
static GCoord lengthZ;   // length of the third dimension, also with periodic boundaries
static GCoord cellSize;  // length of a lattice cell (all sides equal)

-recycle
{
    [species tallyDeath: self];
    hook = recycleEpss;
    recycleEpss = self;
    [species decrementCensus];
    return nil;
}

-recycleClear
{
    Eps* togo;
    while (recycleEpss) {
	togo = recycleEpss;
	recycleEpss = recycleEpss->hook;
	[togo drop];
    }
    return self;
}

-reuse
{
    Eps* result = recycleEpss;
    recycleEpss = recycleEpss->hook;
    result->hook    = nil;
    return result;
}

+create 
{
    id result;

    if (recycleEpss) {
	result = [recycleEpss reuse];
    } else {
	result = [Eps createBegin: globalZone];
    }
    return result;					
}

-copymom {
  // create an agent matching mom in all heritable respects
  
  Eps* baby;
  Quad* newQuad;
  
  baby = [Eps create];
  baby->species   = species;
  //baby->color     = color; // inherit color
  [baby setGColor: [species getGColor]]; // set to default

  baby->kidsize   = kidsize; 
  
  [species incrementCensus];
  [species tallyBirth: self];
  
  newQuad = [Quad create];
  [newQuad setData: baby Type: critter];
  [newQuad setSibling: nil];
  [newQuad setChild: nil];
  [newQuad setParent: nil];
  [baby setQuad: newQuad];

  return baby;
}

-excreteEpsAmount: (ResQuant) e X: (GRate) xi Y: (GRate) yi Z: (GRate) zi {

  timeval_t now = [geckoControl getCurrentTime];
  Eps* baby;
  Quad* myRoot;

  // make the kid 
  baby = [self copymom];

  baby->birthday = [geckoControl getCurrentTime];

  baby->epsDryMass = e;
  baby->initVol = e / epsDensity;
  // EPS could fill the space completely, this is simulated here by treating EPS blobs as cubes
  // For shoving, our radius is half the length a cube of our volume would have
  // In other words, this radius is of the sphere bounded by the cube
  //baby->radius = pow( (baby->epsDryMass / epsDensity) , 1.0/3.0 ) / 2.0;
  baby->radius = sphere2radius( baby->epsDryMass / epsDensity );
  if ((baby->radius * CubertPiOver6) > cellSize)
    printLog("Warning, in eps at %d (baby->radius * CubertPiOver6) > cellSize\n", now); 

  baby->x      = xi;
  baby->y      = yi;
  baby->z      = zi;

  // out of bounds checking has been done already in Bacillus.m, hence not repeated here

  // hang on quad tree
  [baby getMcb: baby->myquad->mcr];
  myRoot = [self getRoot];
  [myRoot addNode: baby->myquad];

  [myquad growMcr];

  return self;
}

-liveGrandma: (Species*) s {
  // Initialize/reinit a default Eps which is as empty as possible
  species = s;

  [self addSpeciesParams];

  return self;
}

-setX: (GCoord) cx Y: (GCoord) cy Z: (GCoord) cz Radius: (GCoord) r {
  // this is called from Bacillus

  [self setX: cx Y: cy  Radius: r];
  z = cz;
  [self getMcb: myquad->mcr]; // have to repeat after changing z

  return self;
}

-setX: (GCoord) cx Y: (GCoord) cy Radius: (GCoord) r {
  // this is called in GeckoModelSwarm.m (-hangCritters: (Critter*)
  // grandma Num: (int) num) to make the very first real Eps (Grandma)
  // with r being grandma->kidsize

  // or, if initNum is zero, is only called by Bacillus with r being epsMass

  timeval_t now = [geckoControl getCurrentTime];

  x = cx;
  y = cy;
  // z comes later

  epsDryMass = r;
  initVol = r / epsDensity;
  //radius = pow( (epsDryMass / epsDensity) , 1.0/3.0 ) / 2.0;
  radius = sphere2radius( epsDryMass / epsDensity );

  // this is a good place for initializations since also called at
  // start of second run and only called at start of runs
  birthday = [geckoControl getCurrentTime];
  
  //[self setGColor: 8]; // we want a different grandma, but protocol-set color for all other eps blobs
  // 0 black, 6 yellow, 8 red, 9 orange, rest hard to see on blue background
  
  [self getMcb: myquad->mcr]; // after this, we have an mcr and can ask for our site
  landRoot   = [[geckoControl getModelSwarm] getLandRoot];
  actualSite = [landRoot firstQuadProducing: 1 Over: [myquad getMcr]];
  mobileRoot = [[geckoControl getModelSwarm] getMobileRoot];
  landMcr    = [actualSite getMcr];

  periodicY = [actualSite getPeriodicY];
  if (periodicY) biofilm = TRUE;
  carrierLayer = [actualSite getCarrierLayer];
  lengthX   = [actualSite getLengthX];
  lengthY   = [actualSite getLengthY];
  lengthZ   = [actualSite getLengthZ];
  cellSize  = [actualSite getCellSize];

  if ((radius * CubertPiOver6) > cellSize)
    printLog("Warning, in eps at %d (radius * CubertPiOver6) > cellSize\n", now); 

  // we can set z randomly in the range [0,lengthZ]
  z = lengthZ * [randSource evenRand];
  [self getMcb: myquad->mcr]; // have to repeat after changing z

  return self;
}

-setConvertBack: (GRate) cb {convertBack = cb; return self;}
-setEpsDensity: (GRate) d {epsDensity = d; return self;}
-setDecayR: (GRate) d {decayR = d * [[geckoControl getModelSwarm] getStepSize]; return self;}
-setKidSize: (GCoord) ks {kidsize = ks; return self;}

-addSpeciesParams {
  [species addParam: "ConvertBack" Vary: 0 DValue: convertBack
	   Message: M(setConvertBack:) Owner: self];
  
  [species addParam: "EpsDensity" Vary: 0 DValue: epsDensity
	   Message: M(setEpsDensity:) Owner: self];

  [species addParam: "DecayR" Vary: 0 DValue: decayR
	   Message: M(setDecayR:) Owner: self];
  
  [species addParam: "KidSize" Vary: 0 DValue: kidsize 
          Message: M(setKidSize:) Owner: self];

  return self;   
}

-move {
  // OK, so I lied.  In theory, it shoves neighbors.  In practice,
  // each finds neighbors to be shoved by.  Would be more efficient
  // doing as a batch once-through process, but first do it right
  // then make it faster if and when.  As of the moment, the distance
  // shoved is simply the vector sum of overlap radii.
  
  timeval_t now = [geckoControl getCurrentTime];
  Quad* myRoot;
  Critter* neighbor;
  Quad* nq;
  GCoord overlapRadius;
  
  static BOOL firsttime = 1;
  static Mcr* shovewin;		// equiv-area square
  
  if (firsttime) {
    firsttime = 0;
    shovewin = [Mcr create];
  }
  
  wannax = wannay = wannaz = 0.0;

  // pick around for neighbors inside radius
  myquad->visible = 0;
  [shovewin circleToBrickX: x Y: y Z: z Radius: radius];
  myRoot = [self getRoot];
  nq = [myRoot pickWin: shovewin Root: myRoot Invisibles: 0]; 
  while (nq) {
    // get one shove
    neighbor = nq->data;
    overlapRadius = radius;
    overlapRadius -= distance3D(x, y, z, neighbor->x, neighbor->y, neighbor->z) - neighbor->radius;
    if (overlapRadius > 0) {
      [self addDirectionX: x - neighbor->x Y: y - neighbor->y Z: z - neighbor->z Weight: overlapRadius];
    }
    
    // next
    nq = [nq pickWin: shovewin Root: myRoot Invisibles: 0]; 
  }
  
  myquad->visible = 1;

  if (biofilm) {
    // Then get a shove from the concrete wall called carrier as well (none if we don't sit on carrier)
    // carrier is always at the low x side
    overlapRadius = [actualSite getOverlapWithCarrierAtX: (x - radius) Y: y ];
    if (overlapRadius > 0) {
      [self addDirectionX: 1 Y: 0 Z: 0 Weight: overlapRadius];
    }
  }
  
  // now do it
  x += wannax;
  y += wannay;
  z += wannaz;

  // Out of bounds check
  if (biofilm) { // EPS should not be within the carrier nor in the topmost layer which is always bulk liquid
    if (x <= carrierLayer) {
      x = carrierLayer + EPSILON; printLog("Warning, in eps at %d x < cellSize\n", now);
    }
    if (x >= (lengthX-cellSize)) {
      x = lengthX - cellSize - EPSILON; printLog("Warning, in eps at %d x > (lengthX-cellSize)\n", now);
    }
  }
  if (periodicY) {
    if (y < 0)        y += lengthY;
    if (y >= lengthY) y -= lengthY;
  }
  // periodicZ always true so far
  if (z < 0)       z += lengthZ;
  if (z >= lengthZ) z -= lengthZ;
  
  [self getMcb: myquad->mcr];
  [myquad growMcr];
  
  return self;
}

// This is an unique identifier for Eps rather than a process number proper, bacteria will have a positive process #
-(int) getProcess { return -1; }

-(ResQuant) getDryMass {return epsDryMass;}

-(ResQuant) getVolume {
  // dummy method, this method is called by GeckoModelSwarm to calc "Biomass"
  return epsDryMass;
}

-checkDeath {
  
  if (radius <= 0.005) {
    [self die]; 
    return nil;
  } else {
    return self;
  }       
}

-die {

  timeval_t now = [geckoControl getCurrentTime];

  printLog("Death of eps at %d\n", now);
  
  [species tallyLossStarve: self Amount: epsDryMass]; 
  // should increase the appropriate sub in the sub matrix, don't want landroot to deal with this
  [landRoot payTaxX: x Y: y Resource: 1 Amount: (epsDryMass * convertBack) ];
  // note that recycle, called by delData, does the census
  [myquad delData];

  return nil;
}

-decay {
  // EPS decay is assumed to be catalyzed by a constant, concn-independent, activity per volume (space, not EPS)

  epsDryMass -= decayR * epsDryMass / initVol;

  return self;
}

-step {
  // Basically, EPS only moves. It's passive stuff. But maybe we want to recycle or age it later.
  // Note: returns nil if dies--higher level routine deletes, because
  // higher level routine is traversing the tree.

  if (decayR > 0.0) {
    // death if radius very small
    if (![self checkDeath]) return nil;
    [self decay];
  }
  
  // print after decay (change in size) and only if alive this step
  //if ([geckoControl getCurrentTime] % [geckoControl getStatCycle] == 0) [self printv];
  
  [self move]; // if this doesn't suffice, Diffuse will call doTheLocomotion
  
  return self;
}

-printv {
  // most printing is done from Diffuse while the growth step is frozen, here we only print when we try things out
  // some things should be printed for all cells, some not
  // some things should be printed often, some not

  static int lastRound = -1;
  static int lastRun = -1;
  timeval_t now = [geckoControl getCurrentTime];
  int thisRun = [geckoControl getNumRun];

  // this is printed only for one cell per round
  if ((lastRound != now) || (lastRun != thisRun)) {
    // this is printed at lower frequency ([geckoControl getStatCycle])
    if (now % [geckoControl getStatCycle] == 0) {

    }
    
    // this is printed every printstep

    lastRound = now;
    lastRun = thisRun;
  }
  
  // this is printed for all cells
  
  // this is printed at lower frequency ([geckoControl getStatCycle])
  if (now % [geckoControl getStatCycle] == 0) {

  }
  return self;
}

-printSpheresMatlab {
  // Replacement for snapshots, prints all spheres for import into Matlab, print Lattices separately
  char filename[256];
  FILE *sphereFile;
  static int lastRound = -1;
  static int lastRun = -1;
  timeval_t now = [geckoControl getCurrentTime];
  int thisRun = [geckoControl getNumRun];
  
  sprintf(filename, "eps_%03ld_%06ld.ml", thisRun, now);
  
  if ((lastRound != now) || (lastRun != thisRun)) { // do this only once at the beginning of each round
    sphereFile = fopen(filename, "w"); // overwrite, don't append
    fclose(sphereFile);
    lastRound = now;
    lastRun = thisRun;
  }

  sphereFile = fopen(filename, "a"); // append, don't overwrite

  fprintf(sphereFile, "%d\t", -1); // numerical identifier for species, for Eps always -1
  fprintf(sphereFile, "%g\t", x);  // x coordinate of center
  fprintf(sphereFile, "%g\t", y);  // y coordinate of center
  fprintf(sphereFile, "%g\t", z);  // z coordinate of center
  fprintf(sphereFile, "%g\n", radius);       // radius

  fclose(sphereFile);
  return self;
}

@end
