// Copyright (C) 1996 - 1999 Yale Institute for Biospheric Studies and Ginger Booth.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

// vfb, original, 960620:  copy from HeatbugBatchSwarm and fit to Gecko.

// The Gecko Control Swarm encapsulates the model swarm and batch and graphical 
// user interface swarms, and the user interface common to batch and graphics,
// and decides which run to run next.

#import <objectbase.h>
#import <objectbase/Swarm.h>
#import "GeckoModelSwarm.h"
#import "GeckoDisplaySwarm.h"
#import "Colors.h"
#import <space.h>
#import <activity.h>
#import <collections.h>
#import <simtoolsgui.h>

@interface ControlParams : Parameterized {
  @public
    int displaycycle;		          
    int statcycle;		          
    int timestep;		          
    int geckoRuns;			  
  int iniGeckoRuns; // the initial number of geckoRuns saved off
  int startTime;
  int startRun;
    BOOL ESRun;				  
    char* fileStem;
}

+create;
-setStatCycle: (int) i;
-setTimeStep: (int) i;
-setDisplayCycle: (int) i ;
-setGeckoRuns: (int) i ;
-setESRun: (int) i ;
-setFileStem: (char*) s;
@end

// what the hell, let's try a GUISwarm on top....see if works.
@interface GeckoControlSwarm : Swarm {
  const char* protocolFileName;		  
  FILE* protocolFile;
  
  ControlParams* settings;
  int displayFlag; 
  
  id reportActions;			  // schedule data structs
  id reportSchedule;
  id stopSchedule;
  
@public // for the moment...
  // id <ControlPanel> controlPanel;		  // if displaying....
  GeckoModelSwarm *   geckoModelSwarm;  // the Swarm we're running
  GeckoDisplaySwarm * geckoDisplaySwarm;// if displaying
  
  // presently not optional
  TraceGraph* popGraph;
  TraceGraph* biomassGraph;
  
  FILE* logFile;			  // running reports, .log
  FILE* bacFile;			  // running reports, for use by Bacillus, .bac
  FILE* porFile;			  // running reports, for use by Diffuse for porosities, .por
  FILE* diffFile;			  // running reports, for use by Diffuse for diffusion steps, .diff
  FILE* archiveFile;			  // reports protocol file contents, .out
@private
  BOOL wantToStop;
  BOOL setupFlag;			  // when it's safe to start doing Swarm inquiries.... 
  BOOL restartFlag;			  // when this is a restarted run
}

+createBegin: (id) aZone;
-setFileStem: (char*) stem;	// stem.log, stem.arc, stem.popGraph, etc.
-(char*) getFileStem;		// stem.log, stem.arc, stem.popGraph, etc.
-(timeval_t) getCurrentTime;    // swarm getCurrentTime unsafe until activity begins
-setDisplayFlag: (int) df;
-(BOOL) getDisplayFlag;
-setStatCycle: (int) sc;
-setGeckoRuns: (int) gr;
-(BOOL) getESRun;		
-setESRun;			// default off, turn on.

-setProtocol: (const char*) s;	// default off, turn on.
-(FILE*) getProtocol;		// nil if no protocol file

-setRestartFlag;
-setStartTime: (int) to;
-setStartRun: (int) ro;
-(BOOL) getRestartFlag;
-(int) getStartTime;
-(int) getStartRun;

-(int) getGeckoRuns;
-(int) getNumRun; // get the number of the current geckoRun
-(GeckoModelSwarm*) getModelSwarm;
-(FILE*) getLogFile;
-(FILE*) getBacFile;
-(FILE*) getPorFile;
-(FILE*) getDiffFile;
-(FILE*) getArchiveFile;
-(int) getStatCycle;

-buildObjects;
-buildActions;
-activateIn: (id) swarmContext;

-go;
-init;
-setWantToStop: (BOOL) swt;
-(BOOL) getWantToStop;
-stopRunning;
-echoInput;	// not yet

-printv;	
-printCensus;	
-println;	// basically, calls all the printlns, so > 1 line...

@end

extern GeckoControlSwarm * geckoControl;
