// Copyright (C) 1996 - 1999 Yale Institute for Biospheric Studies and Ginger Booth.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

// vfb, original, 960620:  copy from HeatbugObserverSwarm and fit to Gecko.

#import "GeckoDisplaySwarm.h"
#import <simtools.h>
#import "QuadDisplay.h"
#import "OtherDisplay.h"

@implementation GeckoDisplaySwarm

// createBegin: here we set up the default display parameters.
+createBegin: (id) aZone {
    GeckoDisplaySwarm * obj;
    
    // Superclass createBegin to allocate ourselves.
    obj = [super createBegin: aZone];
    
    // Fill in the relevant parameters (only one, in this case).
    obj->displaycycle = 1;
    
    // initialize to nil - createEnd will complain if not set by then
    obj->geckoModelSwarm 		= nil;
    
    // weirdly enough, no probe map.

    return obj;
}

-setModel: (GeckoModelSwarm *) gms	// mandatory
{ geckoModelSwarm = gms; return self; }

-setDisplaycycle: (int) dc 
{ displaycycle = dc; return self; }

-getControlPanel {return controlPanel;}
-getActionCache {return actionCache;}

-createEnd
{
	[super createEnd];
    if (geckoModelSwarm == nil)
        geckoError(InvalidCombination,
            "No Gecko world to display.\n");
    return self;
}

-makeWorldDisplay
{
    worldDisplay = [QuadDisplay createBegin: [self getZone]];
	[worldDisplay createEnd];
    [worldDisplay setWindowTitle: "Gecko"];
    [worldDisplay setcolormap];
    [worldDisplay setWidth: 700 Height: 663];
    [worldDisplay setX: -1 Y: 0];
    [worldDisplay setDisplayQuadFlag: 0];
    [worldDisplay setQuadToDisplay: [geckoModelSwarm getWorldRoot]];
    [worldDisplay pack];
    [worldDisplay drawTree];
    [worldDisplay drawSelf];
    return self;
}

-resetClear
{
    [worldDisplay erase];
    [worldDisplay setQuadToDisplay: [geckoModelSwarm getWorldRoot]];
    return self;
}

// Create the objects used in the display of the model.
-buildObjects {
    [super buildObjects];
    [self makeWorldDisplay];
    return self;
}  

// This is where the schedule is built (but not run!)
// Here we create a display schedule - this is used to display the
// state of the world and check for user input. This schedule should
// be thought of as independent from the model - in particular, you
// will also want to run the model without any display.
-buildActions {
    [super buildActions];
    
    // Create the group of display actions. 
    displayActions = [ActionGroup create: [self getZone]];
	[displayActions createActionTo: actionCache  message: M(doTkEvents)];
    [displayActions createActionTo: worldDisplay message: M(erase)];
    [displayActions createActionTo: worldDisplay message: M(drawTree)];
    [displayActions createActionTo: worldDisplay message: M(drawSelf)];
    
    // And the display schedule. Here we put a different repeat interval,
    // to allow for the display to run at a different rate than the world.
    displaySchedule = [Schedule createBegin: [self getZone]];
    [displaySchedule setRepeatInterval: displaycycle]; // note frequency!
    displaySchedule = [displaySchedule createEnd];
    [displaySchedule at: 0 createAction: displayActions];
  
    return self;
}  

// activateIn: - activate the schedules so they're ready to run.
// The swarmContext argument has to do with what we were activated *in*.
// In this case, the GeckoControlSwarm.
-activateIn: (id) swarmContext {
  // First, activate ourselves (just pass along the context).

  [super activateIn: swarmContext];

  // Now activate our schedule in ourselves. This arranges for the
  // execution of the schedule we built.
  [displaySchedule activateIn: self];
  
  // Activate returns the swarm activity - the thing that's ready to run.
  return [self getSwarmActivity];
}

// You could override the "go" method here if you want something special
// to happen when the model and observer actually start running. But
// the default GUISwarm go is probably good enough.

@end
