// vfb, add maxAge to tlg's, 970204
// vfb, add scaredyGrazers for Os, 970204
// vfb, fix tlg's, 961217.
// vfb, orig, 9611??

#import <objectbase.h>
#import "Base.h"
#import "Animal.h"
#import "CritterZoo.h"
#import "Grazer.h"


// TLG_Kris is original TimeLimitGrazer, kept in original implementation 
// because had already run experiments for Kris Johnson
@interface TLG_Kris : Grazer {
  int adultTime;	  // when we become adult (sets br at that time)
}

// routines I need to override 
+create;
-recycle;			
-reuse;			       
-recycleClear;                  // clear out recycling list
-copymom;			

-setAdultTime: (int) at;
-(int) getAdultTime;

-liveGrandma: (Species*) s;	// initialize/reinit a default grazer
-addSpeciesParams;

-(BOOL) reproduce;		// false if didn't -choose- to.

-printv;			
@end

// Note veer does nothing on ColorChoosyGrazers--they look before leaping.
@interface CCG_Kris : TLG_Kris {
    GRate smell_radius;		// multiplier on radius for smelling distance
    int numColors;		// number of colors we eat
    int* colorsToEat;		// array of colors to eat, no pref.
}

// routines I need to override 
+create;
-recycle;			
-reuse;			       
-recycleClear;                  // clear out recycling list
-copymom;			
-addSpeciesParams;

// call eatColor only on grandma
-eatColor: (int) c;
-(int) getNumColors;
-(int) getColor: (int) i;

-setSmellRadius: (GRate) sr;
-(GRate) getSmellRadius;

-(BOOL) attacks: (Animal*) other;

-printv;			
@end

@interface TimeLimitGrazer : Grazer {
  int adultTime;	  // when we become adult (sets br at that time)
  int maxAge;		  // if nonzero, dies at birthday+maxage, else ignored
}

// routines I need to override 
+create;
-recycle;			
-reuse;			       
-recycleClear;                  // clear out recycling list
-copymom;			

-setAdultTime: (int) at;
-(int) getAdultTime;
-setMaxAge: (int) ma;
-(int) getMaxAge;

-liveGrandma: (Species*) s;	// initialize/reinit a default grazer
-addSpeciesParams;

-(BOOL) reproduce;		// false if didn't -choose- to.
-step;

-printv;			
@end

#define MAXCHOOSYCOLORS 5

// Note veer does nothing on ColorChoosyGrazers--they look before leaping.
@interface ColorChoosyGrazer : TimeLimitGrazer {
    GRate smell_radius;		// multiplier on radius for smelling distance
    int numColors;		// number of colors we eat
    int* colorsToEat;		// array of colors to eat, no pref.
}

// routines I need to override 
+create;
-recycle;			
-reuse;			       
-recycleClear;                  // clear out recycling list
-copymom;			
-addSpeciesParams;

// call eatColor only on grandma
-eatColor: (int) c;
-(int) getNumColors;
-(int) getColor: (int) i;

-setSmellRadius: (GRate) sr;
-(GRate) getSmellRadius;

-(BOOL) attacks: (Animal*) other;

-printv;			
@end

// original, 970123, vfb, for Os experiment comparing effect of predation 
// (death) vs. predation risk (evasive behavior) on ecosystem dynamics.
// note:  it's possible to have a scaredyGrazer that's not afraid of anything.
// just set scaredof to nil (or blank on protocol)

// seasonal modifications, 970701, vfb, for same Os experiment:
// Iff there are seasons (geckoModelSwarm->settings > 0), then newborns are
// laid dormant. For esthetics, they're visible, but all deleted at the
// end of the season, and their number used as InitNum next season.

typedef enum { ScaredyHappy, ScaredyFroze, ScaredyRun, ScaredyEgg } 
    ScaredyState;

@interface ScaredyGrazer : TimeLimitGrazer {
  @private
    GCoord lastPredatorDistance;
    GCoord wannadistance;	// like wannax and wannay, maybe < radius
    ScaredyState mood;		// our fright/dormant-egg status	     
  @public
    GRate smell_radius;		// multiplier on radius for smelling predators
    GRate refuge_radius;	// multiplier on radius for finding refuge
    Species* refuge;		// what plant hides me from predators
    Species* scaredof;		// what species I'm scared of
    GRate    refugeGrowrate;	// growth rate eating refuge species
}

// routines I need to override or add
+create;
-recycle;			
-reuse;			       
-recycleClear;                  // clear out recycling list
-copymom;			
-(BOOL) reproduce;		// false if didn't -choose- to.
-addSpeciesParams;

-setSmellRadius: (GRate) sr;
-(GRate) getSmellRadius;
-setRefugeRadius: (GRate) rr;
-(GRate) getRefugeRadius;
-setRefuge: (Species*) safety;
-(Species*) getRefuge;
-setRefugeByName: (char*) safety;
-(char*) getRefugeByName;
-setScaredOf: (Species*) safety;
-(Species*) getScaredOf;
-setScaredOfByName: (char*) safety;
-(char*) getScaredOfByName;
-setRefugeGrowRate: (GRate) rgr;
-(GRate) getRefugeGrowRate;

// NOTE:  if ALL members of this species are dead, the last initNum gets reused.
-countAndDie;	// optional season--reset initNum, add to initNum, and die

-(BOOL) hiddenExposed: (Mcr*) exposed; // is this scaredyGrazer on a refuge plant?
-(BOOL) threatAtX: (GCoord*) tx Y: (GCoord*) ty Distance: (GCoord*) d;
-(BOOL) refugeAtX: (GCoord*) rx Y: (GCoord*) ry Distance: (GCoord*) d;

-move;
-hunt;
-step;

-printv;			
@end
