// orig, vfb, 951019.

// Mcr = "minimum containing rectangle".  This is just an arithmetic toolbox.

// Additions for 3D: Mcb = "minimum containing brick"

// Basically, Mcr's data structure is for 3D, but the standard methods
// ignore the third dimension (z) and just look at x and y. 3D methods
// can be spotted by their name, containing, e.g., "Mcb" or "Z:".

#import "Base.h"

typedef enum { disjoint, contains, containedby, intersect } McrRelation;

@interface Mcr: SwarmObject {
  @public
    GCoord lx;
    GCoord ly;
    GCoord lz;
    GCoord ux;
    GCoord uy;
    GCoord uz;
}

// create sets the mcr to null.
+create;
-(void) drop;
-setNullMcr;
-(BOOL) isNullMcr;

-setMcr: (Mcr*) m;  // set this mcr = m				
-setMcb: (Mcr *) m; // the 3D homologue
-(void) getMcrCenter: (GCoord *) x Y: (GCoord *) y;	
-(void) getMcbCenterX: (GCoord *) x Y: (GCoord *) y Z: (GCoord *) z; // the 3D homologue

// What, if any, intersection does this mcr have to argument mcr?
// Notes:  
//  	a. The null mcr is disjoint everything.  
//      b. If two mcr's are coincident, answer is self contains thatone.
//      c. If the mcr's do not intersect, don't use crossm.

-(McrRelation) intersectMcr: (Mcr*) m Intersection: (Mcr*) crossm;
-(McrRelation) intersectMcr: (Mcr*) m;

// unIntersect always sets uncrossm--the mcr of self's extensions beyond m.
// if self coincident with m, answer is containedby in the un-case
-(McrRelation) unIntersectMcr: (Mcr*) m unIntersection: (Mcr*) uncrossm;

// Like intersectMcr, but quicker--all 3 kinds of intersection = true.
-(BOOL) crossMcr: (Mcr*) m;

// Return intersecting -area- of this mcr with circle's equiv-area square.
-(GArea) crossCircleX: (GCoord) x Y: (GCoord) y Radius: (GCoord) r;

// Return intersecting -areas- of two circles' equiv-area squares.
// This routine does nothing with the mcr it's called with...oh, well.
// Eventually I'll know Objective-C better than this....
-(GArea) crossCircles: (GCoord) x1 Y1: (GCoord) y1 Radius1: (GCoord) r1
                       X2: (GCoord) x2 Y2: (GCoord) y2 Radius2: (GCoord) r2;
-(BOOL) ptInsideX: (GCoord) x Y: (GCoord) y; 

// The resulting square is the mcr returned as self.  Orig mcr ignored.
-circleToSquare: (GCoord) x Y: (GCoord) y Radius: (GCoord) r;
-circleToBrickX: (GCoord) x Y: (GCoord) y Z: (GCoord) z Radius: (GCoord) r; // 3D

// Same, but return an mcr, not equivalent-area square.
-circleToInscribedSquare: (GCoord) x Y: (GCoord) y Radius: (GCoord) r;
                    
// set mcr to prior mcr of itself, extended to include argument mcr.
-growMcr: (Mcr*) m;

// inverted mcr's have no area.
-(GArea) areaMcr;

-print;
-println;

@end
