// vfb, rewrite, 960116?
// vfb, orig, 951020

#import <limits.h>
#import <math.h>
#import "Mcr.h"
#import "Stat.h"

// Defining the methods for an Mcr.
@implementation Mcr

//creating a null Mcr.  
+create
{
    id result;
    result = [[Mcr createBegin: globalZone] createEnd];
    [result setNullMcr];
    return result;					
}

-(void) drop
{
    [super drop];
}

-setNullMcr
{
    lx = INT_MAX;
    ux = INT_MIN;
    ly = INT_MAX;
    uy = INT_MIN;
    lz = INT_MAX;
    uz = INT_MIN;
    return self;
}

-(BOOL) isNullMcr
{
  BOOL answer = 0; 		// until proven otherwise
  if (lx == INT_MAX)
    if (ux == INT_MIN)
      if (ly == INT_MAX)
	if (uy == INT_MIN)
	  if (lz == INT_MAX)
	    if (uz == INT_MIN)
	      answer = 1;
  
  return(answer);
}

-setMcr: (Mcr *) m
    // set this mcr = m
{
    lx = m->lx;    
    ly = m->ly;    
    ux = m->ux;    
    uy = m->uy;    
    return self;   
}

-setMcb: (Mcr *) m
    // set this mcb = m
{
    lx = m->lx;    
    ly = m->ly;    
    lz = m->lz;    
    ux = m->ux;    
    uy = m->uy;    
    uz = m->uz;    
    return self;   
}

-(void) getMcrCenter: (GCoord *) x Y: (GCoord *) y	
{
    *x = lx + (ux - lx) / 2;
    *y = ly + (uy - ly) / 2;
}

-(void) getMcbCenterX: (GCoord *) x Y: (GCoord *) y Z: (GCoord *) z
{
    *x = lx + (ux - lx) / 2;
    *y = ly + (uy - ly) / 2;
    *z = lz + (uz - lz) / 2;
}

-(McrRelation) intersectMcr: (Mcr *) m Intersection: (Mcr *) crossm
    // What, if any, intersection does this mcr have to argument mcr?
    //   Notes:  
    //   a. The null mcr is disjoint everything.  
    //   b. If two mcr's are coincident, answer is self contains thatone.
    //   c. If the mcr's do not intersect, crossm is possibly inverted.
{
    // construct the intersection crossm, then decide. 
    crossm->ux = (ux < m->ux) ? ux : m->ux;  
    crossm->lx = (lx > m->lx) ? lx : m->lx; 
    crossm->uy = (uy < m->uy) ? uy : m->uy;  
    crossm->ly = (ly > m->ly) ? ly : m->ly; 

    if (crossm->ux <= crossm->lx)
        return(disjoint);

    if (crossm->uy <= crossm->ly)
        return(disjoint);

    if ( (crossm->ux == m->ux) && (crossm->uy == m->uy) && 
         (crossm->lx == m->lx) && (crossm->ly == m->ly) )
         return(contains);

    if ( (crossm->ux == ux) && (crossm->uy == uy) && 
         (crossm->lx == lx) && (crossm->ly == ly) )
         return(containedby);

    return(intersect);    
}

-(McrRelation) unIntersectMcr: (Mcr*) m unIntersection: (Mcr*) uncrossm
    // unIntersect always sets uncrossm--the mcr of self's extensions beyond m.
    // if self coincident with m, answer is containedby
{
    int corners, last;
    Mcr* crossm = [Mcr create];
    McrRelation answer = [m intersectMcr: self Intersection: crossm];

    // note we swapped order calling intersectMcr, so who contains who backwards
    if (answer == disjoint) {
        [uncrossm setMcr: self];
    } else if (answer == contains) {
        [uncrossm setNullMcr];
        answer = containedby;
    } else if (answer == containedby) {
        [uncrossm setMcr: self];
        answer = contains;
        
    // all other cases (intersection) boil down to a question of whether
    // the crossm has 2 of self's corners.  If so, we can lop off self's mcr, 
    // else not.
    } else {
        corners = last = 0;
        if ((crossm->ux == ux) && (crossm->uy == uy)) {corners++; }
        if ((crossm->ux == ux) && (crossm->ly == ly)) {corners++; last=1;}
        if ((crossm->lx == lx) && (crossm->ly == ly)) {corners++; last=2;}
        if ((crossm->lx == lx) && (crossm->uy == uy)) {corners++; last=3;}
        if ((crossm->ux == ux) && (crossm->uy == uy)) {last=0; }

        [uncrossm setMcr: self];

        if (corners == 2) {
            if (last==0) {
                uncrossm->uy = crossm->ly;
            } else if (last==1) {
                uncrossm->ux = crossm->lx;
            } else if (last==2) {
                uncrossm->ly = crossm->uy;
            } else {
                uncrossm->lx = crossm->ux;
            }
        }
    }
    
    [crossm drop];
    return answer;
}

-(McrRelation) intersectMcr: (Mcr *) m
{
    McrRelation result;
    Mcr * crossm;
    
    crossm = [Mcr create];
    result = [self intersectMcr: m Intersection: crossm];
    [crossm drop];
    return result;
}

-(BOOL) crossMcr: (Mcr *) m
    // Like intersectMcr, but all 3 kinds of intersection = true.
{
    GCoord cux;
    GCoord clx;
    GCoord cuy;
    GCoord cly;
    
    // construct the intersection, then decide. 
    cux = (ux < m->ux) ? ux : m->ux;  
    clx = (lx > m->lx) ? lx : m->lx; 
    if (cux <= clx)
        return(0);
    
    cuy = (uy < m->uy) ? uy : m->uy;  
    cly = (ly > m->ly) ? ly : m->ly; 

    if (cuy <= cly)
        return(0);

    return(1);    
}

-circleToSquare: (GCoord) x Y: (GCoord) y Radius: (GCoord) r;
    // The resulting square is the mcr returned as self.  Orig mcr ignored.
{
    GCoord equivradius = SqrtPiOver2 * r;
    lx = x - equivradius;
    ly = y - equivradius;
    ux = x + equivradius;
    uy = y + equivradius;
    
    return self;
}

-circleToBrickX: (GCoord) x Y: (GCoord) y Z: (GCoord) z Radius: (GCoord) r;
    // The resulting cube is the mcr returned as self.  Orig mcr ignored.
{
    GCoord equivradius = CubertPiOver6 * r;
    lx = x - equivradius;
    ly = y - equivradius;
    lz = z - equivradius;
    ux = x + equivradius;
    uy = y + equivradius;
    uz = z + equivradius;
    
    return self;
}

-circleToInscribedSquare: (GCoord) x Y: (GCoord) y Radius: (GCoord) r
    // Same, but return an mcr, not equivalent-area square.
{
    lx = x - r;
    ly = y - r;
    ux = x + r;
    uy = y + r;
    
    return self;
}

-(GArea) crossCircleX: (GCoord) x Y: (GCoord) y Radius: (GCoord) r 
       // Return intersecting -area- of self mcr cross with circle's equivalent
      // area square.  Errors from using this approximation are a maximum of
     // about 4% in rare cases.  More commonly 0-2%.
     
      // A square of area equivalent to this circle has sides of length
     // radius * sqrt(pi).  Note it's possible for the two mcrs to cross, but 
    // for this mcr to have no common area with the target circle.  So check.
{
    GArea commonarea;
    Mcr * square = [Mcr create];
    Mcr * crossm = [Mcr create];

    [square circleToSquare: x Y: y Radius: r];
    
    if ( [self intersectMcr: square Intersection: crossm] == disjoint)
        commonarea = 0.0;
    else 
        commonarea = [crossm areaMcr];

    [crossm drop];
    [square drop];

    return(commonarea);
}

-(GArea) crossCircles: (GCoord) x1 Y1: (GCoord) y1 Radius1: (GCoord) r1
                       X2: (GCoord) x2 Y2: (GCoord) y2 Radius2: (GCoord) r2
    // Return intersecting -areas- of two circles' equiv-area squares.
    
    // Defense of weird constant "errpercent":
    // If we use the actual equivalent squares here, we'll get our circles
    // aligning to a square grid, which gives the misbehavior that circles
    // are more likely to settle to 8 neighbors than 5 (for instance).  Given
    // that the landscape quads are rectangular, this really pushes the circles
    // onto a square grid.  So here we construct the equivalent square overlap
    // if the squares were rotated into alignment for minimum overlap.  This
    // error won't cancel out on average, though.  So, we use a fudge factor
    // of .015 (that being somewhere between 0 and 2%, on the generous side)
    // to calculate only the intersecting area.  By inspection, this appears
    // to work correctly.
{
    // cross squares-rotated-to-min-overlap method
    const double errpercent = .015;

    double equivradius1 = (SqrtPiOver2 + errpercent) * r1;
    double equivradius2 = (SqrtPiOver2 + errpercent) * r2;
    double height;

    double d1 = (x2 - x1);
    double d2 = (y2 - y1);

    double dist = sqrt((d1 * d1) + (d2 * d2));
    double width = equivradius1 + equivradius2 - dist;

    if (width < 0.0) width = 0.0;
    
    height = equivradius1 < equivradius2 ? 
             equivradius1 : equivradius2;

    return (width * 2.0 * height);
}    
                   
-(BOOL) ptInsideX: (GCoord) x Y: (GCoord) y
{
    if (ux > x)
        if (lx < x)
            if (uy > y)
                if (ly < y)
                    return 1;
    return 0;
} 

-growMcr: (Mcr *) m
    // set mcr to prior mcr of itself, extended to include argument mcr.
{
    ux = (ux > m->ux) ? ux : m->ux;  
    lx = (lx < m->lx) ? lx : m->lx; 
    uy = (uy > m->uy) ? uy : m->uy;  
    ly = (ly < m->ly) ? ly : m->ly; 
    return(self);
}

-(GArea) areaMcr
    // inverted mcr's have no area.
{
    GCoord answer = 0;
    
    answer = (ux - lx);
    if (answer <= 0) 
        return(0);

    answer *= (uy - ly);
    if (answer <= 0) 
        return(0);

    return(answer);
}

-print
{
    printLogD("[(%5.2f,", lx);
    printLogD("%5.2f,", ly);
    printLogD("%5.2f),", lz);
    printLogD("(%5.2f,", ux);
    printLogD("%5.2f,", uy);
    printLogD("%5.2f)]", uz);
    return self;
}
-println
{  
    [self print];
    printLog("\n");
    return self;
}

@end
