// vfb, orig, 951221:


#import "Base.h"	
#import "Res.h"

// Discussion:  an agent must convert resources from its gullet to its
// stores in order to assimilate anything it's eaten/gotten.  If it's 1:1,
// that's a conversion from A, to A, rate 1.0.  In general, anything
// can be converted to anything, and this supports a conversion matrix.  But
// almost always, the conversion matrix isn't needed, so we cheap out whenever
// possible.  An agent's size is lock-step to its stores, but that logic must
// go uplevel to Critter, as we don't have a generic way to convert stores of
// an arbitrary number of resources to one radius.

@interface Metab: SwarmObject {
  @public	// we may regret this....
    ResList* gullet;
    ResList* stores;
    GRate**  convert;
    GRate*   taxrate;		// taxes are of form taxrate * biomass ^ taxexp
    GRate*   taxexp;
    
    // these are set to nullres if the story is complicated
    ResType  ieat;	  	  // what I eat
    ResType  iam;	          // what I convert it to
    GRate    efficiency;	  // if ieat and iam, conversion is simple, too.
}

+create;
-zero;

// need to add at least one of these for each resource we eat
-addConvertFrom: (ResType) from To: (ResType) to Rate: (GRate) rate;
-setConvertFrom: (ResType) rfrom To: (ResType) rto Rate: (GRate) newrate;

// need to add one of these for each resource we convert to
-addTax: (ResType) r Rate: (GRate) rate Exponent: (GRate) exponent;
-replaceTax: (ResType) r Rate: (GRate) rate Exponent: (GRate) exponent; // actually, same as addTax
-(GRate) getTaxRate: (ResType) r;
-(GRate) getTaxExponent: (ResType) r;
-(GRate) getConvertFrom: (ResType) rfrom To: (ResType) rto;

-copyTo: (Metab*) c;
-newcopy;		// returns pointer to newly created copy

-metabolize; 		// metabolize gullet -> stores 
-(ResList*) taxes; 	// debit taxes and return list to pay

-(ResQuant) getStores: (ResType) res; 
-(ResList*) getStores; 
-setStores: (ResList*) list; 
-setStores: (ResType) what Amount: (ResQuant) quant;
-(ResType) getIeat;
-(ResType) getIam;
-(BOOL) doIeat: (ResType) what;
-(BOOL) amI: (ResType) what;

-feed: (ResList*) dinner;
-debit: (ResList*) payout;
-feedRes: (ResType) res Amount: (ResQuant) quant;

-print;
-println;
-printv;
@end
