// vfb, orig, 9511??.

#import <objectbase.h>
#import <objectbase/SwarmObject.h>
#import <gui.h>
#import <tkobjc/Graph.h>

// A graph that fetches its data, draws it.  Actually traces anything of
// integral value, but defaults to population size.
@interface TraceGraph : SwarmObject {
  int numlines;					  // current.
  int numInts;					  // in census array
  int numDoubles;				  // in values array
  int maxTraces;				  // allocated
  unsigned** census;				  // pointers to values to plot
  double** values;				  // pointers to values to plot
  char** labels;				  // labels for the lines
  id <GraphElement> *lines;	      // the lines themselves
  char* wintitle;				  // default: "Population"
  char* yaxislabel;				  // default: none
  FILE* log;				  	  // default: none
  int displayFlag;			  	  // default: off
  int lastStep;					  // internal
 @public
  id <Graph> graph;			    	  // window for graph
}

+create;	
+createFat: (int) ntraces;	

// call track for each trace you want, then createEnd
-track: (unsigned*) censusvar Key: (char*) label Color: (char*) colorname;
-trackDouble: (double*) valuevar Key: (char*) label Color: (char*) colorname;

// want exact same graph, but start from 0 again.
-clear; 

// setting the labels is optional--really for use by non-population graphs
-setWinTitle: (char*) label;
-setYAxisLabel: (char*) label;
-setLabel: (int) i To: (char*) label;

// if you set a log file, a record is appended at each step
-logFile: (char*) filename;

-createEnd;

-display;

-step;						  // update & refresh

@end

// A graph that fetches its data, draws it.  Differs from TraceGraph in that
// the x-axis is not time.  Only does double vs. double.
@interface TangleGraph : TraceGraph {
  int numXDoubles;				  // in values array
  double** valuesX;				  // pointers to values to plot
  char* xaxislabel;				  // default: none
}
+create;	
// call track for each trace you want, then createEnd
-trackDoubleX: (double*) indepvar Y: (double*) depvar
          Key: (char*) label Color: (char*) colorname;
-setXAxisLabel: (char*) label;
-createEnd;
-step;
@end

/*  vfb, 980327, commented out due to const char* const * labels in
    the latest Histogram class...irksome, and I'm not using Fuzzies.

// Since all in this graph is expressed as a percent, we need to explicitly
// say what to use as onehundred % to scale everything else.  I intend to
// use different divisors for different items in the same graph.  Hence, the
// "data" and "onehundred" lists.
@interface FuzzyGraph : SwarmObject {		// a histogram for buckets
    int numlines;				  // current.
    unsigned** data;	  			  // pointers to values to plot
    unsigned** onehundred;			  // divisor for each data val
    Histogram* graph;				  // window for graph
    GraphElement **bars;	        	  // the bars themselves
    const char* const* labels; 	          // labels for the bars
    char** colors;				  // names of colors
    char* wintitle;				  // default: "Population Structure"
    char* yaxislabel;				  // default: "Percent"
    char* xaxislabel;				  // no default
    double* toplot;			          // derived values to plot
}

+create;

// call track for each bar you want, then createEnd
-track: (unsigned*) datavar Scale: (unsigned*) onehunvar 
   Key: (char*) label Color: (char*) colorname;

// setting the labels is optional
-setWinTitle: (char*) label;
-setXAxisLabel: (char*) label;
-setYAxisLabel: (char*) label;

-createEnd;

-step;			 			 // update & refresh

@end

*/