
#import "Base.h"
#import <collections/List.h>
			  // we inherit swarm objects

// Classes Param and Parameterized.

@interface Param: SwarmObject {
  @public
    char * 	label;
    ParamType 	paramType;	// LabelParam, IntParam, DoubleParam, StringParam- Base.h
    BOOL 	varyFlag;	// flag evol. strat to optimize this one
    int 	ival;
    double 	dval;
    char*	sval;
    SEL         message;	// what message to use to set this param
    id          owner;		// who to send the message to (normally)
};

-apply;				// call owner with message & current value

// Note that I/D here refers to the number you're passing in, not
// the number that gets stored.  It's converted to what the parameter
// thinks it wants.
-(ParamType) getParamType;
-(BOOL) getVary;

-setIValue: (int) i;
-(int) getIValue;
-setDValue: (double) d;
-(double) getDValue;

// these are synonyms for the above for normal Params
-(double) getPartDValue;   // only my value
-setPartDValue: (double) d; // only my value 
-(int) getPartIValue;      // only my value
-setPartIValue: (int) i; // only my value 

-setSValue: (char*) s;
-(char*) getSValue;
-(char*) getParamName;
-(id) getOwner;
-copymomZone: (id) z;// note: we never copy strings, just addresses.
-setLabel: (char*) l
    Vary: (BOOL) vflag
    Message: (SEL) m
    Owner: (id) o;

-print;
-println;
-printv;
-dump; // dump is reloadable, print is for looks
@end

// The value of an offset param is added to the value of the "offsetFrom"
// param, and -then- message is called to set the param's owner.
//    	[owner message: [offsetFrom getDValue] + dval];
// Example "BreedSize" offset might have value 4.0, offset from MinSize 3.0.
// results in 
// 	[owner setBreedSize: 3.0 + 4.0];
// Clearly, offsetFrom and OffsetParam must be of same type.

@interface OffsetParam: Param {
  @public
    Param*	offsetFrom;    
};
-apply;
-setOffsetFrom: (id) op;
-(id) getOffsetFrom;

-(double) getOffsetDValue; // offset's value, not mine
-(int) getOffsetIValue;    // offset's value, not mine
-(double) getPartDValue;   // only my value
-setPartDValue: (double) d; // only my value 
-(int) getPartIValue;      // only my value
-setPartIValue: (int) i; // only my value 
-setIValue: (int) i;	 // set overall value (me = i - offsetFrom)
-(int) getIValue;        // get overall value
-setDValue: (double) d;  // set overall value (me = d - offsetFrom)
-(double) getDValue;     // get overall value

-copymomZone: (id) z;
-print;
-dump;
-println;
-printv;
@end

@interface Parameterized: SwarmObject {
    char* 	 name;		// a label
    id <List> params;	// parameter list
};

+create;
-init;

-addLabelParam: (char*) label;

-addParam: (char*) label 	
    Vary: (BOOL) vflag
    DValue: (double) d
    Message: (SEL) m
    Owner: (id) o;

-addParam: (char*) label 	
    Vary: (BOOL) vflag
    DValue: (double) d
    Message: (SEL) m
    Owner: (id) o
    Offset:   (id) op;

-addParam: (char*) label 	
    Vary: (BOOL) vflag
    IValue: (int) i
    Message: (SEL) m
    Owner: (id) o
    Offset:   (id) op;

-addParam: (char*) label 	
    Vary: (BOOL) vflag
    IValue: (int) i
    Message: (SEL) m
    Owner: (id) o;

-addParam: (char*) label 	
    Vary: (BOOL) vflag
    SValue: (const char*) s
    Message: (SEL) m
    Owner: (id) o;

-(id) getParam: (char*) label;

// If Param of this name doesn't exist, but is optional, this adds it and
// returns a pointer to the newly added parameter.  For the base class, there
// are no optional parameters, but subclasses may have some.  If this behavior
// is desired, call getOptParam -instead- of getParam.
-(id) getOptParam: (char*) label; 

-(Parameterized*) getParameterized: (char*) label;

-(char*) getParamName;
-setName: (char*) label;
-(id) getParams;		// returns pointer to master copy
-setParam: (char*) label IValue: (int) i;
-setParam: (char*) label DValue: (double) d;
-setParam: (char*) label SValue: (char*) s;

-apply;	// apply all parameters to their owners

// Offset a parameter you've already added.  Necessary for offsetting from
// parameter of a different parameter list or overriding what the species
// added as default parameters.  Doesn't mean anything for string/label.
-offsetParam: (char*) label Off: (id) op;

// Current value is already the offset--don't subtract op.
-offsetParam: (char*) label AbsOff: (id) op;

// doesn't mean anything for string/label. 
-varyParam: (char*) label;

-print;
-println;
-printv;
// dump is reloadable, print is for looks -- note the @begin...@end is
// added higher in the object hierarchy--this is just a parameter list.
-dump; 

// parses/executes on current parameterized until next @begin / EOF.
// if @begin, string returned is rest of line after @begin
// if EOF, returns nil for string.
-(char*) loadToNextSection: (FILE*) fp;

// reiterates what was returned at last loadToNextSection
-(char*) loadWhere;

// single-word-step parsing (not executing)
-(char*) loadNextWord: (FILE*) fp;

// oops, got a word too many
-(char*) reloadln: (FILE*) fp;
                       
@end
