// vfb, orig, 951019

// With best regards to Kevin Crowder, Steve Krazit, and Jim Von Ehr.

#import "Quad.h"
#import "Critter.h"
#import "CritterZoo.h"
#import "Grazer.h"
#import "GrazerZoo.h"
#import "Carnie.h"
#import "CarnieZoo.h"
#import "GeckoControlSwarm.h"
#import "Evolve.h"
#import <activity.h>
#import "Base.h"

// The constant MaxPerQuad sets the threshold at which any quad (zone) is
// ideally subquadded.

// #define MaxPerQuad moved to Base.h

@implementation Quad

Quad* recycleQuads = nil;
Mcr* scratchwin = nil;
int pickTrace = 0;

-recycle
{
    sibling      = recycleQuads;
    child = parent = nil;
    type = invalid;
    recycleQuads = self;
    return nil;
}

-recycleClear
{
    Quad* togo;
    while (recycleQuads) {
        togo = recycleQuads;
        recycleQuads = recycleQuads->sibling;
        [togo->mcr drop];
        [togo drop];
    }
    return self;
}

-reuse
{
    Quad* result    = recycleQuads;
    recycleQuads    = recycleQuads->sibling;
    result->data    = nil;
    result->parent  = nil;
    result->sibling = nil;
    result->child   = nil;
    return result;
}

-init: (BOOL) newmcr
{
    type 	= aquad;
    visible 	= 1;
    data    	= nil;
    sibling     = child = parent = nil;

    if (newmcr) 
        mcr = [Mcr create];
    [mcr setNullMcr];
    
    return self;
}

+create
{
    Quad* newQuad;

    if (recycleQuads) {
        newQuad = [recycleQuads reuse];
        [newQuad init: 0];
    } else {
        newQuad = [Quad create: globalZone];
        [newQuad init: 1];
    }

    return newQuad;					
}

-setParent: (Quad *) p 
{
    parent = p;
    return self;
}

-setSibling: (Quad *) s 
{
    sibling = s;
    return self;
}

-setChild: (Quad *) c 
{
    child = c;
    return self;
}

-setData: d Type: (QuadType) t
{
    type = t;
    data = d;
    return self;
}

-setGColor: (int) c { color = c; return self; }

-(GCoord) getRadius
    // if not an agent, fake it.
{
    GCoord fake, fakey;
    if (isagent(type))
        return [data getRadius];
    fake = mcr->ux - mcr->lx;
    fakey = mcr->uy - mcr->ly;
    if (fakey > fake) fake = fakey;
    fake /= 2.0;
    return fake;
}

-(Mcr *) getMcr {return mcr;}

  //
 // OK, now for the multi-node routines....
//
-addNode: (Quad *) p
    // add node as first child of this parent--actually, to newbies collector
{
    Quad* where;
//GINGER - out? which out?  if no newbies, clean everywhere.
    // where        = [self findNewbies];
    where = self;
    p->sibling   = where->child;
    p->parent    = where;
    where->child = p;
    [p growMcr];
    return self;
}

-findNewbies
    // Return newbies collector under current quad--make newbies if none.
    // Note newbies must be added at the front of the tree if we don't want
    // to visit them again in the traversal they're born during.
{
    Quad* where;
    
    where = child;
    while ((where != nil) && (where->type != newbies))
        where = where->sibling;
    
    if (where == nil) {
        where          = [Quad create];
        where->sibling = child;
        where->parent  = self;
        where->type    = newbies;
        child          = where;
    }
    return where;
}

-diagnose
{
    Critter* c;
    Quad* children = child;
    while (children) {
        if (children->parent != self) {
            printLog("***ERROR--child/parent mismatch\n    ");
            [children printQuad];
            printLog("\n");
            if (!(children->data) && isagent(children->type))
                printLog("Quad says agent, but has no data\n");
            else {
                c = children->data;
                printLog("   "); 
                [c println];
            }
            printLog("    Reached from parent:"); 
            [self printQuad]; printLog("\n");
            printLog("    Says parent is:"); 
            [children->parent printQuad]; printLog("\n");
        }
        if (!(children->data) && isagent(children->type)) {
            printLog("***ERROR--Quad says agent, but has no data\n");
            [children printQuad];
            printLog("\n");
            printLog("    Reached from parent:"); 
            [self printQuad]; printLog("\n");
        }
        [children diagnose];
            
        children = children->sibling;
    }
    return self;
}

-delNode
      // Only this quad tree node.  Correct parent, siblings, children.
     // Lots of error checking.  If self has substructure, it squashes up
    // to the front of self's prior sibling chain. Doesn't actually delete
   // anything, just unplugs from tree.
{
    Quad * sis, * kid;
// GINGER--out
    Critter* c;

    if (self == nil)
        return(self);
    
    // fix the kids' parent
    if (child != nil) {
        kid = child;
        while (kid->sibling != nil) {
            kid->parent = parent;
            kid = kid->sibling;
        } 
        kid->parent = parent;
        kid->sibling = sibling;
    }
    
    // now, "child" is head of list, "kid" is end of list and its sibling set
    // or, if we had no children, child is nil and kid undefined

    sis = parent->child;

    if (child != nil) {
        if (sis == self) {
            parent->child = child;
        } else {
            while ((sis != nil) && (sis->sibling != self))
                sis = sis->sibling;
            if (!sis) {
                printLog("***ERROR--garbled tree detected in delNode (child)\n");
                [self printQuad];
                printLog("\n");
                c = (Critter*) data;
                [c println];
                if (!data)
                    printLog("    Quad data == nil\n");
                printLog("\n");
                printLog("    Parent:");
                [parent printQuad];
                printLog("\n");
                // [worldRoot diagnose];
            } else 
                sis->sibling = child;
        }
    } else {
        if (sis == self)
            parent->child = sibling;
        else {
            while ((sis != nil) && (sis->sibling != self))
                sis = sis->sibling; 
            if (sis == nil) {
                printLog("***ERROR--garbled tree detected in delNode (nochild)\n");
                [self printQuad];
                printLog("\n");
                c = (Critter*) data;
                [c println];
                if (!data)
                    printLog("    Quad data == nil\n");
                printLog("\n");
                printLog("    Parent:");
                [parent printQuad];
                printLog("\n");
                // [worldRoot diagnose];
            } else
                sis->sibling = sibling;
        }
    }
    
    return(self);
} 

-delData
{
    if (data != nil) [data recycle];
    data = nil;
    [self delNode];
    [self recycle];
    return nil;
}

-delDataQuick
{
    if (data != nil) [data recycle];
    data = nil;
    [self recycle];
    return nil;
}

-delTree
    // doesn't delete the Quad it's called with, only its children
{
    Quad* thislevel = self;
    Quad* delthis, *delnext;

    if (thislevel->child) {
        delthis = thislevel->child;
        thislevel->child = nil;
        do {
            delnext = delthis->sibling;
            if (delthis != nil) [delthis delTree];
            if (delthis != nil) [delthis delDataQuick];
            delthis = delnext;
        } while (delthis);
    }
    return self;
}

-growMcr
    // grows mcr's above this node, as needed	
{
    Quad* thisone;

    // if current mcr contains new mcr, get out.
    if (parent == nil) 
        return self;

    if ([parent->mcr intersectMcr: mcr] == contains) {
        return self;
    } 
    
    // otherwise, have work to do.
    thisone = parent;
    if (thisone != nil) do {
        [thisone->mcr growMcr: mcr];
        thisone = thisone->parent;
    } while ( (thisone != nil) && 
              ([thisone->mcr intersectMcr: mcr] != contains) );

    return self;
}

-(Quad *) pickWin: (Mcr *) pickwin 
          Root: (Quad*) rootPick Invisibles: (BOOL) pickinvis
     // We only pick leaves.
{
    Quad* lastPick = self;
    
    if (lastPick == nil) 
        return(nil);
    
    do {    
            // go down?
        if ((lastPick->child) && ([pickwin crossMcr: lastPick->mcr])) {
            lastPick = lastPick->child; 
    
            // stop here?
        } else if (lastPick == rootPick) {

            return(nil);
        
            // go across?
        } else if (lastPick->sibling) {
            lastPick = lastPick->sibling; 
    
            // pop up?
        } else {
            do {
                lastPick = lastPick->parent;
                if ( (lastPick == rootPick) || (lastPick == nil) )
            
                        return(nil);
                    
            } while (lastPick->sibling == nil);
            lastPick = lastPick->sibling;
        }
            // are we there yet?
        if (   ( ispickable(lastPick->type) )
             && ( [pickwin crossMcr: lastPick->mcr] ) 
             && ( pickinvis || (lastPick->visible) )    ) {

            return(lastPick);

        }
    } while (1);
    
    return(0); // not that we can reach here....
}	

-(Quad *) pickWin: (Mcr *) pickwin Root: (Quad*) rootPick
    // calls other with invisibles = false
{
    return [self pickWin: pickwin Root: rootPick Invisibles: 0];
}

-(Quad *) pick
     // We only pick leaves.
{
    return [self pickWin: mcr Root: self Invisibles: 0];
}

-(Quad *) pickRadius: (GCoord) pickR X: (GCoord) pickX Y: (GCoord) pickY
          Root: (Quad*) pickRoot
{
    Quad* answer;
    Critter* candidate;
    GCoord dist;
    
    if (scratchwin == nil) scratchwin = [Mcr create];
    [scratchwin circleToInscribedSquare: pickX Y: pickY Radius: pickR];
    answer = self;
    
    while ((answer = 
           [answer pickWin: scratchwin Root: pickRoot Invisibles: 0])) {
        candidate = answer->data;
        if (candidate) {
            dist = distance(pickX, pickY, [candidate getX], [candidate getY]);
            dist -= pickR;
            dist -= [candidate getRadius];
            if (dist <= 0) {
                return answer;
            }
        }
    }    
    return 0;
}

-cleanup
{
    if (child == nil) {
        self = [self delNode];
        [self recycle];
    }
    return self;
}

-(void) subQuad 
    // recursively subquad children of this quad
{
    int count = 0;
    int lrcount, llcount, urcount, ulcount;
    GCoord pivotx, pivoty, kidx, kidy;
    Quad* kid, * newparent, * nextkid;
    Quad* ulquad, * llquad, * urquad, * lrquad;
    
    // count children, calculating mcr.
    
    kid = self->child;

    if (kid == nil) {
        return; 
    }
    
    do {
        count++;
        self->mcr = [self->mcr growMcr: kid->mcr];
    } while ((kid = kid->sibling) != nil);

    // do we have few enough members of the quad to quit?
    if (count <= MaxPerQuad)
        return;

    // drat, more work to do.  
    // Allocate 4 subquads and thread together.
    llquad = [Quad create]; 
    lrquad = [Quad create]; 
    ulquad = [Quad create];
    urquad = [Quad create]; 
    [llquad setParent: self]; [llquad setSibling: lrquad];
    [lrquad setParent: self]; [lrquad setSibling: ulquad];
    [ulquad setParent: self]; [ulquad setSibling: urquad];
    [urquad setParent: self]; [urquad setSibling: nil];

    // divvy up the kids into grandkids.
    [self->mcr getMcrCenter: &pivotx Y: &pivoty];

    kid = self->child;
    urcount = ulcount = llcount = lrcount = 0;
    
    if (kid != nil) do {
        [kid->mcr getMcrCenter: &kidx Y: &kidy];
        
        // If a kid is coincident on the pivot, we have to choose whether
        // to go above or below.  If we always go below (for instance), then
        // it's possible (though hopefully uncommon) to get into an infinite 
        // recursion due to >MaxPerQuad coincident leaves.
        
        // We need another tie-breaker....

        if (kidx > pivotx) {
            if (kidy > pivoty) { 
                newparent = urquad; urcount++;
            } else if (kidy < pivoty) {
                newparent = lrquad; lrcount++;
            } else if (urcount < lrcount) {
                newparent = urquad; urcount++;
            } else {
                newparent = lrquad; lrcount++;
            }
        } else {
            if (kidy > pivoty) {
                newparent = ulquad; ulcount++;
            } else if (kidy < pivoty)  {
                newparent = llquad; llcount++;
            } else if (ulcount < llcount) {
                newparent = ulquad; ulcount++;
            } else {
                newparent = llquad; llcount++;
            }
        }
        
        nextkid = kid->sibling;
        [kid setParent: newparent];
        [kid setSibling: newparent->child];
        [newparent setChild: kid];
       
        count++;
    } while ((kid = nextkid) != nil);

    [self setChild: llquad];
    
    // And subquad.  Note that some subQuads may nix themselves.
    // Must do it here rather on sub-recursion, lest we delete our root.
    [urquad subQuad];
    [urquad cleanup];
    [lrquad subQuad];
    [lrquad cleanup];
    [ulquad subQuad];
    [ulquad cleanup];
    [llquad subQuad];
    [llquad cleanup];
}

-squashQuad
    // flatten subtree into simple child list
{
    Quad* thisone, * nextone, * list;

    thisone = [self pickWin: mcr Root: self Invisibles: 1];
    list = nil;
    while (thisone) {
        nextone = [thisone pickWin: mcr Root: self Invisibles: 1];
        [thisone delNode];
        thisone->parent = self;
        thisone->sibling = list;
        list    = thisone;
        thisone = nextone;
    }
    [self delTree];
    child = list;
    return self;
}

-(void) printQuad
{
    printLog("Quad mcr = ");
    [self->mcr print];
    printLog(", type = ");
    if (type == aquad) 
        printLog("aquad.");
    else if (type == critter) {
        printLog("\n\t\t\t");
        [data println];
    } else if (type == feed) 
        printLog("feed.");
    else if (type == diffuse) 
        printLog("diffuse.");
    else if (type == newbies) 
        printLog("newbies.");
    else if (type == landscape) 
        printLog("landscape.");
    else if (type == sessile) 
        printLog("sessile.");
    else if (type == mobile) 
        printLog("mobile.");
    else 
        printLog("UNKNOWN.");
}
-println {[self printQuad]; printLog("\n"); return self;}

-(void) printTree 
    // call printTree, not the one with depth.
{
    int quads = 0;
    int leaves = 0;
    [self printQuad];
    printLog("\n");
    [child printTree: 2 Quads: &quads Leaves: &leaves];
    printLog("Tree contains %d quads and %d leaves.\n", quads, leaves);
}

-(void) headcount: (char*) message
     // just print number of agents in tree
{
    int qcount = 0;
    int lcount = 0;
    [self maxDepth: 0 Quads: &qcount Leaves: &lcount];
    printLog("%s %d\n", message, lcount);
}

-(void) printTree: (int) depth Quads: (int*) qcount Leaves: (int*) lcount
    // Just traverse the tree and print it--debug routine.  The depth
    // parameter controls indentation only.
    // These trees can be vast--don't do both directions recursively....
{
    int i;
    Quad* thislevel = self;
    
    do {
        for (i=0; i<depth; i++) 
            printLog(" ");
        if (type == aquad) *qcount += 1; else *lcount += 1;
        [thislevel printQuad]; 
        printLog("\n");
        [thislevel->child printTree: depth+2 Quads: qcount Leaves: lcount];
    } while ((thislevel = thislevel->sibling) != nil);  
}

-(int) maxDepthSibs: (int) depth Quads: (int*) qcount Leaves: (int*) lcount
    // Collect some stats about the tree.
{
    int thisdepth;
    Quad* thislevel = self;
    int mydepth = depth; // going down
    
    do {
        if (ispickable(thislevel->type)) *lcount += 1; else *qcount += 1;
        thisdepth = [thislevel->child maxDepthSibs: depth+1 
                     Quads: qcount Leaves: lcount];
        if (thisdepth > mydepth) mydepth = thisdepth;
    } while ((thislevel = thislevel->sibling) != nil);  

    return(mydepth);
}

-(int) maxDepth: (int) depth Quads: (int*) qcount Leaves: (int*) lcount
    // Collect some stats about the tree.
{
    if (ispickable(type)) *lcount += 1; else *qcount += 1;
    depth = [child maxDepthSibs: depth+1 Quads: qcount Leaves: lcount];
    return(depth);
}

-substep
    // Call step for each data (leaf) node....
{
    Quad* thislevel = self;
    Quad* next;
    
    do {
        next = thislevel->sibling; // because we maybe kill thislevel or breed...
        if (isagent(thislevel->type)) {
            if ([thislevel->data step]) {
                if (thislevel->child)
                    [thislevel->child substep];
            }
        } else if (thislevel->child) {
            [thislevel->child substep];
        }
    } while ((thislevel = next) != nil); 

    return self; 
}

-step
    // Call step for each data (leaf) node....
{
    [child substep];
    return self; 
}

-stepOn: (int) roundth;
{
    if ([geckoControl getCurrentTime] % roundth == 0)
        [self step];
    return self;
}


@end
