#import <random.h>
#import <sys/time.h>
#import "RandSource.h"
extern int gettimeofday();

// vfb, 980407:  replaced old random.c from Popsim with Swarm
// random library wrapper.

@implementation RandSource

-makeFromSeed
{
	generator = [MT19937gen create: globalZone
			  setStateFromSeed: randomSeed];
	evenGen = [UniformDouble create: globalZone
			  setGenerator: generator
			  setDoubleMin: 0.0L
			  setMax: 1.0L];
	gaussGen = [NormalDist create: globalZone
			  setGenerator: generator
			  setMean: 0.0
			  setVariance: 1.0];
	return self;				
}

+create
{   
	struct timeval tp;
	id obj = [RandSource createBegin: globalZone];
	RandSource* robj = obj;

	if (gettimeofday(&tp, (struct timezone *)0) == -1){
	    robj->randomSeed = 123456789;
	}
	
	robj->randomSeed = tp.tv_sec;

	[obj makeFromSeed];
	return obj;
}

-reSeed: (unsigned) seed 
{
	randomSeed = seed;
	if (generator) {
		[generator drop];
		[evenGen drop];
		[gaussGen drop];
	}
	[self makeFromSeed];
	return self;
}

-setGen: newGenerator 
{
	if (generator) {
		[generator drop];
		[evenGen drop];
		[gaussGen drop];
	}
	generator = newGenerator;
	evenGen = [UniformDouble create: globalZone
			  setGenerator: generator
			  setDoubleMin: 0.0L
			  setMax: 1.0L];
	gaussGen = [NormalDist create: globalZone
			  setGenerator: generator
			  setMean: 0.0
			  setVariance: 1.0];

	return self;
}
-(double) evenRand {return [evenGen getDoubleSample];}
-(double) gaussRand {return [gaussGen getDoubleSample];}
-(unsigned long long int) getCurrentCount { return [generator getCurrentCount];}

-saveRngState {
  
  int rngStateSize;
  id  rngState;
  int status;

  char filename[256];
  FILE *file;
  timeval_t now = [geckoControl getCurrentTime];
  int thisRun = [geckoControl getNumRun];

  rngStateSize = [generator getStateSize];
  rngState = (id) [[self getZone] alloc: rngStateSize];  
  [generator putStateInto: (void *) rngState];

  sprintf(filename, "bac_%03ld_%06ld.rng", thisRun, now);

  file = fopen(filename, "w");
  status = fwrite(rngState, rngStateSize, 1, file);
  fclose(file);

  [[self getZone] free: rngState];
  
  return self;
}

-readRngState {
  
  int rngStateSize;
  id  rngState;
  int status;

  char filename[256];
  FILE *file;
  timeval_t now = [geckoControl getCurrentTime];
  int thisRun = [geckoControl getNumRun];

  rngStateSize = [generator getStateSize];
  rngState = (id) [[self getZone] alloc: rngStateSize];  

  sprintf(filename, "bac_%03ld_%06ld.rng", thisRun, now);

  file = fopen(filename, "r");
  if (file != NULL) {
    printLog("Warning: We now read in the rngState from file %s\n", filename);
    status = fread(rngState, rngStateSize, 1, file);
    fclose(file);
    if (status == 1) {
      [generator setStateFrom: (void *) rngState];
    } else {
      geckoError(SourceMessage, "\n\nERROR: Could not read rngState file %s properly!\n\n\n", filename);    
    }
  } else {
    geckoError(SourceMessage, "\n\nERROR: Could not open rngState file %s for reading!\n\n\n", filename);    
  }
  
  [[self getZone] free: rngState];
  
  return self;
}

@end
