// vfb, orig, 960909:

#import "Param.h"
#import "Stat.h"
#import "Species.h"
#import "Colors.h"
#import "GeckoControlSwarm.h"
#import <simtools.h>
#import <ctype.h>
@implementation Species;

-init
{
    [super init];
    census = 0;
    lastalive = 0;
    biomass = 0;
    dbiomass = 0;
    grandma = nil;
    birthday = 0;
    deathday = 0;
    birthArea = nil;
    initNum = 0;
    ageStats = nil;
    ageGraph = nil;
    specificGraph = nil;
    stageGraph = nil;
    sizeGraph = nil;
    fertGraph = nil;
    mortGraph = nil;
    stageStats = nil;
    sizeStats = nil;
    bioRateStats = nil;
    bioTransStats = nil;
    popFlag = bioFlag = 0;
    seasonMessage = 0;
    return self;
}

+create
{
    id result = [[Species createBegin: globalZone] createEnd];
    result = [result init];
    return result;					
}

-createEnd 
    // after **everything**, finish up the graphs
{
    // give species last chance to intervene
    if (grandma) [grandma statsFinalize];

    if (ageGraph) ageGraph = [ageGraph createEnd];
    if (stageGraph) stageGraph = [stageGraph createEnd];
    if (sizeGraph) sizeGraph = [sizeGraph createEnd];
    if (specificGraph) specificGraph = [specificGraph createEnd];
    if (mortGraph) mortGraph = [mortGraph createEnd];
    if (fertGraph) fertGraph = [fertGraph createEnd];
    if (bioRateGraph) bioRateGraph = [bioRateGraph createEnd];
    if (bioTransGraph) bioTransGraph = [bioTransGraph createEnd];
    if (bioTheftGraph) bioTheftGraph = [bioTheftGraph createEnd];
    if (bioGrowthGraph) bioGrowthGraph = [bioGrowthGraph createEnd];
    if ([geckoControl getDisplayFlag]) {
        if (ageGraph)  [ageGraph display];
        if (stageGraph) [stageGraph display];
        if (sizeGraph) [sizeGraph display];
        if (specificGraph) [specificGraph display];
        if (ageGraph)  [ageGraph display];
        if (fertGraph) [fertGraph display];
        if (mortGraph) [mortGraph display];
        if (bioRateGraph) [bioRateGraph display];
        if (bioTransGraph) [bioTransGraph display];
        if (bioTheftGraph) [bioTheftGraph display];
        if (bioGrowthGraph) [bioGrowthGraph display];
    }
    return self;
}

-setName: (char*) s { name = s; return self; }
-setMobile: (int) i { if (i) mobile = 1; else mobile = 0; return self; }
-setBioScale: (double) d { bioscale = d; return self; }
-setType: (char*) s { return self; } // Yes, this is a no-op, type can't be set

-make: (char*) n WithGrandma: (id) g Color:   (int) c
                 Birthday:   (int) b InitNum: (int) i
                 Mobile: (BOOL) m    BioScale: (double) bs
                 
    // "create" a grandma of correct type, then this
    // rely on class of grandma to dynamically select correct message
{
    int wordlen, w;
    const char* classname;
    char* copyclassname;

    name     = n;
    color    = c;
    birthday = b;
    initNum  = i;
    mobile   = m;
    bioscale = bs;
    ageStats = nil;

    classname = object_get_class_name(g);
    wordlen = 0;
    while (!isspace(UCHAR(classname[wordlen]))) wordlen++;
    wordlen++;
    copyclassname = [[self getZone] alloc: wordlen + 1];
    for (w=0; w <= wordlen; w++) 
        copyclassname[w] = classname[w];

    [self addParam: "Type" Vary: 0 SValue: copyclassname
          Message: M(setType:) Owner: self];
    [self addParam: "Mobile" Vary: 0 IValue: mobile
          Message: M(setMobile:) Owner: self];
    [self addParam: "BioScale" Vary: 0 DValue: bioscale
          Message: M(setBioScale:) Owner: self];
        
    [self addParam: "Birthday" Vary: 0 IValue: b 
          Message: M(setBothBirthday:) Owner: self];
    [self addParam: "Deathday" Vary: 0 IValue: 0 
          Message: M(setDeathDay:) Owner: self];
    [self addParam: "InitNum" Vary: 0 IValue: i 
          Message: M(setBothInitNum:) Owner: self];
    [self addParam: "Color" Vary: 0 IValue: c 
          Message: M(setGColor:) Owner: self];

    // remaining parameters added by grandma
    grandma = [g liveGrandma: self];

    return self;
}

-(Param*) getOptParam: (char*) label
{
    id result;
    char* opt1="BirthAreaLX";
    char* opt2="BirthAreaLY";
    char* opt3="BirthAreaUX";
    char* opt4="BirthAreaUY";
    char* opt5="OnSeason";
    char* opt6="AgesBuckets";
    char* opt9="SpecificBuckets";

    // simple case--one of the required parameters
    if ((result = [self getParam: label])) {
        return result;
    }
    
    // OK, now consider whether it's one of our optional params --  
    if ( !strcmp(label, opt1) ||
         !strcmp(label, opt2) ||
         !strcmp(label, opt3) ||
         !strcmp(label, opt4) ) {
         
       birthArea = [Mcr create];
        [self addParam: opt1 Vary: 0 DValue: 0 
            Message: M(setBALX:) Owner: self];
        [self addParam: opt2 Vary: 0 DValue: 0 
            Message: M(setBALY:) Owner: self];
        [self addParam: opt3 Vary: 0 DValue: 0 
            Message: M(setBAUX:) Owner: self];
        [self addParam: opt4 Vary: 0 DValue: 0 
            Message: M(setBAUY:) Owner: self];
    } else if ( !strcmp(label, opt5) ) {
        [self addParam: opt5 Vary: 0 SValue: ""
            Message: M(setOnSeason:) Owner: self];
    } else if ( !strcmp(label, opt6) ) {
        [self addParam: opt6 Vary: 0 IValue: 0
            Message: M(setAgesBuckets:) Owner: self];
    } else if ( !strcmp(label, opt9) ) {
        [self addParam: opt9 Vary: 0 IValue: 0
            Message: M(setSpecificBuckets:) Owner: self];
    } else {
        return nil; 
    }
    
    return [self getParam: label];
} 

-setBALX: (GCoord) d { birthArea->lx = d; return self;}
-setBALY: (GCoord) d { birthArea->ly = d; return self;}
-setBAUX: (GCoord) d { birthArea->ux = d; return self;}
-setBAUY: (GCoord) d { birthArea->uy = d; return self;}

-setOnSeason: (char*) s 
    // Note that the season message doesn't execute on me (species), but on 
    // grandma's descendents (agents).
{ 
    SEL sel;

    sel = sel_get_uid( s );
    if ( sel ) {
        if ( [grandma respondsTo: sel] ) {
	    seasonMessage = sel;
        } else {
            geckoError(InvalidCombination, 
                "Protocol error, season command '%s' not recognized.\n", s);
        }
    } else {
        geckoError(InvalidCombination,
            "Protocol error, season command '%s' not recognized.\n", s);
    }
    return self;
}

-onSeason
    // onSeason is called every World-season'th timestep, on sites and species.
    // Here, we simply check if we have a seasonMessage to perform, and if so,
    // traverse the world looking for members of this species, and call the
    // seasonMessage on each.
    // THEN, reset birthday for this species as now+original birthday.
{
    GeckoModelSwarm* model = geckoControl->geckoModelSwarm;
    id one;
    Quad* root;
    Quad* quad;
    
    // all the quick routes out....
    if (census == 0) 
        return self;
    if (seasonMessage == 0)
        return self;

    if (mobile) 
        root = model->mobileRoot;
    else
        root = model->plantRoot;
    
    quad = [root pickWin: root->mcr Root: root Invisibles: 0];
    while (quad) {
        one = quad->data;
        quad = [quad pickWin: root->mcr Root: root Invisibles: 0];
        if ([one getSpecies] == self) {
            [one perform: seasonMessage];
        }
    }
    
    birthday = [geckoControl getCurrentTime] + ( birthday % model->settings->season );
    
    return self;
}

-observePop
{
    popFlag = 1;
    [geckoControl->popGraph track: &census Key: name Color: colorName(color)];
    return self;
}

-observeBiomass
{
    bioFlag = 1;
    [geckoControl->biomassGraph trackDouble: &biomass Key: name 
            Color: colorName(color)];
    return self;
}

-setAgesBuckets: (int) i {
    if (ageStats == nil)
        ageStats = [AgeStructure createAges: i];
    return self;
}

-makeMortAndFertMax: (int) max {
    char* namebuf;
    int num;
    
    if (max > 20)
        num = max;
    else
        num = 20;
    if (fertGraph == nil) {
        namebuf = [globalZone alloc: 50];
        fertGraph = [TraceGraph createFat: num];
        sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
                name, "fertGraph");
        [fertGraph logFile: namebuf];
        sprintf(namebuf, "%s%s", name, " Fertility");
        [fertGraph setWinTitle: namebuf];
    }
    if (mortGraph == nil) {
        namebuf = [globalZone alloc: 50];
        mortGraph = [TraceGraph createFat: num];
        sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
                name, "mortGraph");
        [mortGraph logFile: namebuf];
        sprintf(namebuf, "%s%s", name, " Mortality");
        [mortGraph setWinTitle: namebuf];
    }
    return self;
}

-observeAgesOnly
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;

    if (ageGraph != nil)
        return self;
    
    if (ageStats == nil) {
        [self setAgesBuckets: [grandma getDefaultAgeCount]];
    }
    ageGraph = [TraceGraph createFat: [ageStats getMaxAge]];
    namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "ageGraph");
    [ageGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Age Structure");
    [ageGraph setWinTitle: namebuf];
    
    for (i=0; i < [ageStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "%% Age %d", i);
        bucket = (GenericBucket*) ageStats->bucket[i];
        [ageGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_AGEPERCENT] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-observeAges
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;
    
    [self observeAgesOnly];

    if ((mortGraph == nil) || (fertGraph == nil))
        [self makeMortAndFertMax: [ageStats getMaxAge] * 2];
    
    for (i=0; i < [ageStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Age %d", i);
        bucket = (GenericBucket*) ageStats->bucket[i];
        [fertGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_MX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Age %d", i);
        [mortGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_PX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-setSizesBuckets: (int) i {
    if (sizeStats == nil)
        sizeStats = [SizeStructure createSizes: i];
    return self;
}

-observeSizesOnly
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;

    if (sizeGraph != nil)
        return self;
    
    if (sizeStats == nil) {
        [self setSizesBuckets: [grandma getDefaultSizes]];
    }
    sizeGraph = [TraceGraph createFat: [sizeStats getMaxAge]];
    namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "sizeGraph");
    [sizeGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Size Structure");
    [sizeGraph setWinTitle: namebuf];
    
    for (i=0; i < [sizeStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "%% Size %d", i);
        bucket = (GenericBucket*) sizeStats->bucket[i];
        [sizeGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_AGEPERCENT] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-observeSizes
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;
    
    [self observeSizesOnly];

    if ((mortGraph == nil) || (fertGraph == nil))
        [self makeMortAndFertMax: [sizeStats getMaxAge] * 2];
    
    for (i=0; i < [sizeStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Size %d", i);
        bucket = (GenericBucket*) sizeStats->bucket[i];
        [fertGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_MX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Size %d", i);
        [mortGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_PX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-setStagesBuckets: (int) i {
    if (stageStats == nil)
        stageStats = [StageStructure createStages: i];
    return self;
}

-observeStagesOnly
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;

    if (stageGraph != nil)
        return self;
    
    if (stageStats == nil) {
        [self setStagesBuckets: [grandma getDefaultStages]];
    }
    stageGraph = [TraceGraph createFat: [stageStats getMaxAge]];
    namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "stageGraph");
    [stageGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Stage Structure");
    [stageGraph setWinTitle: namebuf];
    
    for (i=0; i < [stageStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "%% Stage %d", i);
        bucket = (GenericBucket*) stageStats->bucket[i];
        [stageGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_AGEPERCENT] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-observeStages
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;
    
    [self observeStagesOnly];

    if ((mortGraph == nil) || (fertGraph == nil))
        [self makeMortAndFertMax: [stageStats getMaxAge] * 2];
    
    for (i=0; i < [stageStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Stage %d", i);
        bucket = (GenericBucket*) stageStats->bucket[i];
        [fertGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_MX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Stage %d", i);
        [mortGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_PX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-setSpecificBuckets 
{
    if (specificStats == nil)
        specificStats = [SpecificStructure 
            createBuckets: [grandma getSpecificNumBuckets]
            Counters:      [grandma getSpecificCounts]
            Sums:          [grandma getSpecificSums]
            Derived:       [grandma getSpecificDeriveds] ];
    return self;
}

-observeSpecific
{
    int i, j, nd;
    int lablen = 40;
    char* namebuf;
    GenericBucket* bucket;

    if (specificGraph != nil)
        return self;
    
    if (specificStats == nil) {
        [self setSpecificBuckets];
    }
    
    // If species has taken responsibility for this, we're done.
    // Note species might also take some responsibility, but return
    // false to let us add traces.
    if ([grandma observeSpecific: specificStats])
        return self;
    
    // OK, let's provide default traces
    specificGraph = [TraceGraph createFat: 3 * [specificStats getMaxBucket]];
    namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "specificGraph");
    [specificGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Specific Structure");
    [specificGraph setWinTitle: namebuf];
    
    for (i=0; i < [specificStats getMaxBucket]; i++) {
        nd = [specificStats->bucket[i] getNumDerived];
        for (j=0; j < nd; j++) {
            namebuf = [globalZone alloc: lablen];
            sprintf(namebuf, "Specific %d %d", i, j);
            bucket = (GenericBucket*) specificStats->bucket[i];
            [specificGraph trackDouble: 
                    &( bucket->derived[j])
                    Key: namebuf 
                    Color: arbColorStrings[(nd*i + j) % arbColorMax]];
        }
    }
    return self;
}

-observeBiomassRates
{
    char* namebuf = [globalZone alloc: 50];

    bioRateStats = [BiomassRates create];

    bioRateGraph = [TraceGraph create];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "bioRateGraph");
    [bioRateGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Biomass Rates");
    [bioRateGraph setWinTitle: namebuf];
    [bioRateGraph trackDouble: &(bioRateStats->netGrowth)
                Key: "NetGrowth" Color: ColorNetGrowth];
    [bioRateGraph trackDouble: &(bioRateStats->growth)
                Key: "Growth" Color: ColorGrowth];
    [bioRateGraph trackDouble: &(bioRateStats->lossSteal)
                Key: "LossSteal" Color: ColorLossSteal];
    [bioRateGraph trackDouble: &(bioRateStats->lossStarve)
                Key: "LossStarve" Color: ColorLossStarve];
    [bioRateGraph trackDouble: &(bioRateStats->losses)
                Key: "Losses" Color: ColorLosses];
    [bioRateGraph trackDouble: &(bioRateStats->netRatio)
                Key: "NetRatio" Color: ColorNetRatio];
            
    return self;
}

-observeBiomassTransfersSilent
{
    bioTransStats = [BiomassTransfers create];
    return self;
}

-observeBiomassTransfers
{
    char* namebuf = [globalZone alloc: 50];

    if (!bioTransStats)
        [self observeBiomassTransfersSilent];

    bioTransGraph = [TraceGraph create];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "bioTransGraph");
    [bioTransGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Biomass Transfers");
    [bioTransGraph setWinTitle: namebuf];
    [bioTransGraph trackDouble: &(bioTransStats->netGrowth)
                Key: "NetGrowth" Color: ColorNetGrowth];
    [bioTransGraph trackDouble: &(bioTransStats->growth)
                Key: "Growth" Color: ColorGrowth];
    [bioTransGraph trackDouble: &(bioTransStats->lossSteal)
                Key: "LossSteal" Color: ColorLossSteal];
    [bioTransGraph trackDouble: &(bioTransStats->lossStarve)
                Key: "LossStarve" Color: ColorLossStarve];
    [bioTransGraph trackDouble: &(bioTransStats->losses)
                Key: "Losses" Color: ColorLosses];
    [bioTransGraph trackDouble: &(bioTransStats->netRatio)
                Key: "NetRatio" Color: ColorNetRatio];
    [bioTransGraph trackDouble: &(bioTransStats->biomass)
                Key: "Biomass" Color: colorName(color)];
            
    return self;
}

-observeBiomassTheft
{
    char* namebuf = [globalZone alloc: 50];

    if (!bioTransStats)
        [self observeBiomassTransfersSilent];

    bioTheftGraph = [TangleGraph create];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "bioTheftGraph");
    [bioTheftGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Biomass Theft");
    [bioTheftGraph setWinTitle: namebuf];
    [bioTheftGraph setXAxisLabel: "Biomass"];
    [bioTheftGraph setYAxisLabel: "Biomass Theft"];
    [bioTheftGraph trackDoubleX: &(bioTransStats->biomass)
                Y: &(bioTransStats->lossSteal)
                Key: "LossSteal" Color: colorName(color)];
    return self;
}

-observeBiomassGrowth
{
    char* namebuf = [globalZone alloc: 50];

    if (!bioTransStats)
        [self observeBiomassTransfersSilent];

    bioGrowthGraph = [TangleGraph create];
    sprintf(namebuf, "%s%s", name, ".bioGrowthGraph");
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "bioGrowthGraph");
    [bioGrowthGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Biomass Growth");
    [bioGrowthGraph setWinTitle: namebuf];
    [bioGrowthGraph setXAxisLabel: "Biomass"];
    [bioGrowthGraph setYAxisLabel: "Biomass Growth"];
    [bioGrowthGraph trackDoubleX: &(bioTransStats->biomass)
                Y: &(bioTransStats->growth)
                Key: "Growth" Color: colorName(color)];
    return self;
}

-tallyGrowth: (id) instance Amount: (ResQuant) b 
{ 
    [bioTransStats tallyGrowth: instance Amount: b]; 
    [bioRateStats tallyGrowth: instance Amount: b]; 
    [specificStats tallyGrowth: instance Amount: b]; 
    return self;
}

-tallyLossSteal: (id) instance Amount: (ResQuant) b 
{ 
    [bioTransStats tallyLossSteal: instance Amount: b]; 
    [bioRateStats tallyLossSteal: instance Amount: b]; 
    [specificStats tallyLossSteal: instance Amount: b]; 
    return self;
}
-tallyLossStarve: (id) instance Amount: (ResQuant) b 
{ 
    [bioTransStats tallyLossStarve: instance Amount: b]; 
    [bioRateStats tallyLossStarve: instance Amount: b]; 
    [specificStats tallyLossStarve: instance Amount: b]; 
    return self;
}
-zeroBiomassRates 
{
    [bioTransStats clear]; 
    [bioRateStats clear]; 
    return self;
}

-biomassRateUpdate
{
    [self biomassTransferUpdate];
    if (bioRateStats == nil) 
        return self;
    
    [bioRateStats calcNet: biomass];
    if (bioRateGraph) {
        [bioRateGraph step];
    }
    return self;
}

-biomassTransferUpdate
{
    if (bioTransStats == nil) 
        return self;
    
    [bioTransStats calcNet: biomass];
    if (bioTransGraph) 
        [bioTransGraph step];
    if (bioTheftGraph)
        [bioTheftGraph step];
    if (bioGrowthGraph)
        [bioGrowthGraph step];
    
    return self;
}

-tallyAlive: (id) instance
{
    Critter* c = instance;
    if (ageStats) [ageStats tallyAlive: c];
    if (sizeStats) [sizeStats tallyAlive: c];
    if (stageStats) [stageStats tallyAlive: c];
    if (specificStats) [specificStats tallyAlive: c];
    return self;
}

-tallySpecific: (id) aCritter EventType: (int) anEvent
{
    if (specificStats) 
        [specificStats tallySpecific: aCritter EventType: anEvent];
    return self;   
}

-tallyDeath: (id) instance
{
    Critter* c = instance;
    if (ageStats) [ageStats tallyDeath: c];
    if (sizeStats) [sizeStats tallyDeath: c];
    if (stageStats) [stageStats tallyDeath: c];
    if (specificStats) [specificStats tallyDeath: c];
    return self;
}

-tallyBirth: (id) instance
{
    Critter* c = instance;
    if (birthday == [geckoControl getCurrentTime]) return self;
    
    if (ageStats) [ageStats tallyBirthToMom: c];
    if (sizeStats) [sizeStats tallyBirthToMom: c];
    if (stageStats) [stageStats tallyBirthToMom: c];
    if (specificStats) [specificStats tallyBirthToMom: c];
    return self;
}

-xUpdate: (id) xStats graph: (TraceGraph*) xGraph
{
    if (!xStats)
        return self;
    [xStats calc];
    if (xGraph) {
        [xGraph step];
    }
    return self;
}

-ageUpdate
{
    if (!ageStats) return self;
    [self xUpdate: ageStats graph: ageGraph];
    return self;
}

-sizeUpdate
{
    if (!sizeStats) return self;
    
    [self xUpdate: sizeStats graph: sizeGraph];
    return self;
}

-stageUpdate
{
    if (!stageStats) return self;
    [self xUpdate: stageStats graph: stageGraph];
    return self;
}

-specificUpdate
{
    if (!specificStats)
        return self;
    [grandma calcSpecific: specificStats];
    if (specificGraph) {
        [specificGraph step];
    }
    return self;
}

-allStatsReset
{
    if (fertGraph) 
        [fertGraph step];
    if (mortGraph) 
        [mortGraph step];
    if (ageStats) [ageStats clear];
    if (sizeStats) [sizeStats clear];
    if (stageStats) [stageStats clear];
    if (specificStats) [specificStats clear];
    if (bioRateStats) [bioRateStats clear];
    if (bioTransStats) [bioTransStats clear];
    return self;
}

-setBirthArea: (Mcr*) where
    ResetMessage: (SEL) m
    Owner: (id) o
{
    birthArea = where;
    resetAreaMessage = m;
    birthAreaOwner = o;
    return self;
}

-setDeathDay: (int) d { deathday = d; return self; }

-deathdayCheck
    // go kill all instances of this species in the world
{
    GeckoModelSwarm* model = geckoControl->geckoModelSwarm;
    id one;
    Quad* root, *quad;
    // all the quick routes out....
    if (deathday == 0) 
        return self;
    if ([geckoControl getCurrentTime] < deathday) 
        return self;
    if (census == 0) 
        return self;

    // here we'll eventually check for other conditions--peak, etc.
    // not yet.
    
    if (mobile) 
        root = model->mobileRoot;
    else
        root = model->plantRoot;
    
    quad = [root pickWin: root->mcr Root: root Invisibles: 0];
    while (quad) {
        one = quad->data;
        quad = [quad pickWin: root->mcr Root: root Invisibles: 0];
        if ([one getSpecies] == self) {
            [one die];
        }
    }
    return self;
}

-birthdayCheck
{
    GeckoModelSwarm* model = geckoControl->geckoModelSwarm;
    if ([geckoControl getCurrentTime] == birthday) {
        if (birthArea) {
            [birthAreaOwner perform: resetAreaMessage];
        }
        if (mobile)
            [model hangCritters: grandma Num: initNum 
                   Root: model->mobileRoot Where: birthArea];
        else
            [model hangCritters: grandma Num: initNum 
                   Root: model->plantRoot Where: birthArea];
        if (bioRateStats) {
            [bioRateStats clear];
        }
        if (bioTransStats) {
            [bioTransStats clear];
        }
    }
    return self;
}


-resetClear 
{
    // season/birthday reset
    birthday = saveBirthday;
    initNum  = saveInitNum;
    
    [self allStatsReset];
    if (sizeGraph) {
        [sizeGraph clear];
    }
    if (ageGraph) {
        [ageGraph clear];
    }
    if (stageGraph) {
        [stageGraph clear];
    }
    if (specificGraph) {
        [specificGraph clear];
    }
    if (mortGraph) {
        [mortGraph clear];
    }
    if (fertGraph) {
        [fertGraph clear];
    }
    if (bioRateGraph) {
        [bioRateGraph clear];
    }
    if (bioTransGraph) {
        [bioTransGraph clear];
    }
    if (bioTheftGraph) {
        [bioTheftGraph clear];
    }
    if (bioGrowthGraph) {
        [bioGrowthGraph clear];
    }

    // for some reason, this already seems correct, but let's be careful....
    census    = 0;
    lastalive = 0;
    biomass   = 0;
    dbiomass   = 0;
    return self;
}

-setGrandma: (id) g { grandma = g; return self; }
-(id) getGrandma { return grandma; }
-(unsigned) getCensus {return census;}
-(unsigned) getLastAlive
{ 
    if (census) 
        lastalive = [geckoControl getCurrentTime]; 
    return lastalive;
}
-(unsigned) getBirthday {return birthday;}

-incrementCensus 
{
    census++; 
    lastalive = [geckoControl getCurrentTime];
    return self;
}

-decrementCensus 
{
    census--; 
    if (census > 0)
        lastalive = [geckoControl getCurrentTime];
    return self;
}

-(double) getBiomass {return biomass;}
-setBiomass: (double) b {dbiomass = b; return self;}
-addBiomass: (double) b {dbiomass += b; return self;}
-zeroBiomass {biomass = 0; dbiomass = 0; return self;}
-scaleBiomass {biomass = dbiomass / bioscale; return self;}

-setBirthday: (unsigned) b {
  birthday = b + [geckoControl getStartTime];
  return self;
}

-setBothBirthday: (unsigned) b {
  birthday     = b + [geckoControl getStartTime];
  saveBirthday = b + [geckoControl getStartTime]; 
  return self;
}

-setInitNum:  (unsigned) i {initNum = i; return self;}
-setBothInitNum:  (unsigned) i {initNum = i; saveInitNum = i; return self;}
-(unsigned) getInitNum {return initNum;}

-(int) getGColor {return color;}

-setGColor: (int) c
{
    color = c;
    [(Critter*) grandma setGColor: c];
    return self;
}

// these are for use by CompositeSpecies...I want multiple inheritance...
-(id) getStageStats {return stageStats;}
-(id) getAgeStats {return ageStats;}
-(id) getSizeStats {return sizeStats;}
-(id) getSpecificStats {return specificStats;}
-(BiomassRates*) getBioRateStats {return bioRateStats;}
-(BiomassTransfers*) getBioTranStats {return bioTransStats;}
-(BiomassTransfers*) getBioTransStats {return bioTransStats;}
-(double) getBioscale {return bioscale;}
-(TraceGraph*) getMortGraph { return mortGraph;}
-(TraceGraph*) getFertGraph { return fertGraph;}
-(TraceGraph*) getSpecificGraph { return specificGraph;}
-setSpecificGraph:  (TraceGraph*) sGraph {specificGraph = sGraph; return self;}
-setSpecificStats:  (id) specific
{specificStats = specific; return self;}

-print
{
    printLog("%s", name); return self;    
}

-println
{
    printLog("Species %s census=%d  biomass= ", name, census);
    printLogD("%lf \n", biomass);
    return self;    
}

-printv
{
    [self println];
    printLog("Species params: \n"); 
    [super printv];
    printLog("Species grandma: \n"); 
    [grandma printv];
    return self;
}

-printCensus 
    // no eoln.
{
    printLog("%s:%d ", name, census);
    return self;
}

-dump
    // reloadable
{
    printBoth("@begin Species %s\n", name);

    [super dump];

    if (popFlag)
        printBoth("\tobservePop\n");
    if (bioFlag)
        printBoth("\tobserveBiomass\n");
    if (ageGraph) {
        printBoth("\tAgesBuckets\t%d\n", [ageStats getMaxAge]);
        printBoth("\tobserveAges\n");
    }
    if (sizeGraph)
        printBoth("\tobserveSizes\n");
    if (stageGraph)
        printBoth("\tobserveStages\n");
    if (specificGraph)
        printBoth("\tobserveSpecific\n");
    if (bioRateGraph)
        printBoth("\tobserveBiomassRates\n");
    if (bioTransGraph)
        printBoth("\tobserveBiomassTransfers\n");
        
    printBoth("@end Species %s\n\n", name);
    return self;
}
@end

@implementation CompositeSpecies;

int lastTimeUpdatedBasics = -1;

-init
{
    [super init];
    census = 0;
    lastalive = 0;
    biomass = 0;
    members = [List create: [self getZone]];
    stageStats = nil;
    bioRateStats = nil;
    bioTransStats = nil;
    popFlag = bioFlag = 0;
    lastTimeUpdatedBasics = -1;

    [self addParam: "Color" Vary: 0 IValue: 1 
          Message: M(setGColor:) Owner: self];
    return self;
}

+create
{
    id result = [[CompositeSpecies createBegin: globalZone] createEnd];
    result = [result init];
    return result;					
}

-createEnd 
    // after **everything**, finish up the graphs
{
    if (ageGraph) ageGraph = [ageGraph createEnd];
    if (stageGraph) stageGraph = [stageGraph createEnd];
    if (sizeGraph) sizeGraph = [sizeGraph createEnd];
    if (specificGraph) specificGraph = [specificGraph createEnd];
    if (mortGraph) mortGraph = [mortGraph createEnd];
    if (fertGraph) fertGraph = [fertGraph createEnd];
    if (bioRateGraph) bioRateGraph = [bioRateGraph createEnd];
    if (bioTransGraph) bioTransGraph = [bioTransGraph createEnd];
    if (bioTheftGraph) bioTheftGraph = [bioTheftGraph createEnd];
    if (bioGrowthGraph) bioGrowthGraph = [bioGrowthGraph createEnd];
    if ([geckoControl getDisplayFlag]) {
        if (ageGraph)  [ageGraph display];
        if (stageGraph) [stageGraph display];
        if (sizeGraph) [sizeGraph display];
        if (specificGraph) [specificGraph display];
        if (ageGraph)  [ageGraph display];
        if (fertGraph) [fertGraph display];
        if (mortGraph) [mortGraph display];
        if (bioRateGraph) [bioRateGraph display];
        if (bioTransGraph) [bioTransGraph display];
        if (bioTheftGraph) [bioTheftGraph display];
        if (bioGrowthGraph) [bioGrowthGraph display];
    }
    return self;
}

-addSpecies:  (Species*) s {
    [members addLast: s]; 
    bioscale = [s getBioscale];
    return self; 
}    

-(Param*) getOptParam: (char*) label
{
    id result;
    char* opt6="AgesBuckets";

    // simple case--one of the required parameters
    if ((result = [self getParam: label])) {
        return result;
    }
    
    // OK, now consider whether it's one of our optional params --  
    if ( !strcmp(label, opt6) ) {
        [self addParam: opt6 Vary: 0 IValue: 0
            Message: M(setAgesBuckets:) Owner: self];
    } else {
        return nil; 
    }
    
    return [self getParam: label];
} 

-setAgesBuckets: (int) i {
    if (ageStats == nil)
        ageStats = [AgeStructure createAges: i];
    return self;
}

-setSizesBuckets: (int) i {
    if (sizeStats == nil)
        sizeStats = [SizeStructure createSizes: i];
    return self;
}

-makeMortAndFertMax: (int) max {
    char* namebuf;
    int num;
    
    if (max > 20)
        num = max;
    else
        num = 20;
    if (fertGraph == nil) {
        namebuf = [globalZone alloc: 50];
        fertGraph = [TraceGraph createFat: num];
        sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
                name, "fertGraph");
        [fertGraph logFile: namebuf];
        sprintf(namebuf, "%s%s", name, " Fertility");
        [fertGraph setWinTitle: namebuf];
    }
    if (mortGraph == nil) {
        namebuf = [globalZone alloc: 50];
        mortGraph = [TraceGraph createFat: num];
        sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
                name, "mortGraph");
        [mortGraph logFile: namebuf];
        sprintf(namebuf, "%s%s", name, " Mortality");
        [mortGraph setWinTitle: namebuf];
    }
    return self;
}

-observeAgesOnly {
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;
    id  index;
    Species* s;

    if (ageGraph != nil)
        return self;
    
    if (ageStats == nil) {
        s = [members first];
        [self setAgesBuckets: [[s getGrandma] getDefaultAgeCount]];
    }
    
    index = [members begin: globalZone];
    // note:  if member species already has ageStats, this doesn't override
    while ( (s = [index next] ) ) {
        [s setAgesBuckets: [ageStats getMaxAge]];
    }
    [index drop];
    
    ageGraph = [TraceGraph createFat: [ageStats getMaxAge]];
    namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "ageGraph");
    [ageGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Age Structure");
    [ageGraph setWinTitle: namebuf];
    
    for (i=0; i < [ageStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "%% Age %d", i);
        bucket = (GenericBucket*) ageStats->bucket[i];
        [ageGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_AGEPERCENT] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-observeAges
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;

    if (ageGraph != nil)
        return self;
    
    [self observeAgesOnly];
    
    if ((mortGraph == nil) || (fertGraph == nil))
        [self makeMortAndFertMax: [ageStats getMaxAge] * 2];
    for (i=0; i < [ageStats getMaxAge]; i++) {
        bucket = (GenericBucket*) ageStats->bucket[i];
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Age %d", i);
        [fertGraph trackDouble: &(bucket->derived[AGE_DERIVE_MX])
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Age %d", i);
        [mortGraph trackDouble: &(bucket->derived[AGE_DERIVE_PX])
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-observeSizesOnly
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;
    id  index;
    Species* s;

    if (sizeGraph != nil)
        return self;
    
    if (sizeStats == nil) {
        s = [members first];
        [self setSizesBuckets: [[s getGrandma] getDefaultSizes]];
    }
    index = [members begin: globalZone];
    // note:  if member species already has sizeStats, this doesn't override
    while ( (s = [index next] ) ) {
        [s setSizesBuckets: [sizeStats getMaxAge]];
    }
    [index drop];
    
    sizeGraph = [TraceGraph createFat: [sizeStats getMaxAge]];
    namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "sizeGraph");
    [sizeGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Size Structure");
    [sizeGraph setWinTitle: namebuf];
    
    for (i=0; i < [sizeStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "%% Size %d", i);
        bucket = (GenericBucket*) sizeStats->bucket[i];
        [sizeGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_AGEPERCENT] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-observeSizes
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;
    
    [self observeSizesOnly];

    if ((mortGraph == nil) || (fertGraph == nil))
        [self makeMortAndFertMax: [sizeStats getMaxAge] * 2];
    
    for (i=0; i < [sizeStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Size %d", i);
        bucket = (GenericBucket*) sizeStats->bucket[i];
        [fertGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_MX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Size %d", i);
        [mortGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_PX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-setStagesBuckets: (int) i {
    if (stageStats == nil)
        stageStats = [StageStructure createStages: i];
    return self;
}

-observeStagesOnly
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;
    id  index;
    Species* s;

    if (stageGraph != nil)
        return self;
    
    if (stageStats == nil) {
        s = [members first];
        [self setStagesBuckets: [[s getGrandma] getDefaultStages]];
    }
    index = [members begin: globalZone];
    // note:  if member species already has stageStats, this doesn't override
    while ( (s = [index next] ) ) {
        [s setStagesBuckets: [stageStats getMaxAge]];
    }
    [index drop];
    
    stageGraph = [TraceGraph createFat: [stageStats getMaxAge]];
    namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "stageGraph");
    [stageGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Stage Structure");
    [stageGraph setWinTitle: namebuf];
    
    for (i=0; i < [stageStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "%% Stage %d", i);
        bucket = (GenericBucket*) stageStats->bucket[i];
        [stageGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_AGEPERCENT] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

-observeStages
{
    int i;
    int lablen = 11;
    char* namebuf;
    GenericBucket* bucket;
    
    [self observeStagesOnly];

    if ((mortGraph == nil) || (fertGraph == nil))
        [self makeMortAndFertMax: [stageStats getMaxAge] * 2];
    
    for (i=0; i < [stageStats getMaxAge]; i++) {
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Stage %d", i);
        bucket = (GenericBucket*) stageStats->bucket[i];
        [fertGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_MX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
        namebuf = [globalZone alloc: lablen];
        sprintf(namebuf, "Stage %d", i);
        [mortGraph trackDouble: 
                  &( bucket->derived[AGE_DERIVE_PX] )
                  Key: namebuf 
                  Color: arbColorStrings[i % arbColorMax]];
    }
    return self;
}

// The grandma used is the first in the list of members.  Hopefully they match.
-setSpecificBuckets: (id) grandma 
{
    if (specificStats == nil)
        specificStats = [SpecificStructure 
            createBuckets: [grandma getSpecificNumBuckets]
            Counters:      [grandma getSpecificCounts]
            Sums:          [grandma getSpecificSums]
            Derived:       [grandma getSpecificDeriveds] ];
    return self;
}

-observeSpecific
{
    int i, j, nd;
    int lablen = 40;
    char* namebuf;
    GenericBucket* bucket;
    id  index;
    Species* s;

    if (specificGraph != nil)
        return self;
    
    if (specificStats == nil) {
        s = [members first];
        [self setSpecificBuckets: [s getGrandma]];
    }
    index = [members begin: globalZone];
    while ( (s = [index next] ) ) {
        [s setSpecificBuckets];
    }
    [index drop];
    
    // If species has taken responsibility for this, we're done.
    // Note species might also take some responsibility, but return
    // false to let us add traces.
    if ([[s getGrandma] observeSpecific: specificStats])
        return self;
    
    // OK, let's provide default traces
    specificGraph = [TraceGraph createFat: [specificStats getMaxBucket]];
    namebuf = [globalZone alloc: 50];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "specificGraph");
    [specificGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Specific Structure");
    [specificGraph setWinTitle: namebuf];
    
    for (i=0; i < [specificStats getMaxBucket]; i++) {
        nd = [specificStats->bucket[i] getNumDerived];
        for (j=0; j < nd; j++) {
            namebuf = [globalZone alloc: lablen];
            sprintf(namebuf, "Specific %d %d", i, j);
            bucket = (GenericBucket*) specificStats->bucket[i];
            [specificGraph trackDouble: 
                    &( bucket->derived[j])
                    Key: namebuf 
                    Color: arbColorStrings[(nd*i + j) % arbColorMax]];
        }
    }
    return self;
}

-(char*) getName { return name; }
-setName: (char*) s { name = s; return self; }
-(int) getGColor { return color; }
-setGColor: (int) c { color = c; return self; }

-observePop
{
    popFlag = 1;
    [geckoControl->popGraph track: &census Key: name Color: colorName(color)];
    return self;
}

-observeBiomass
{
    bioFlag = 1;
    [geckoControl->biomassGraph trackDouble: &biomass Key: name
        Color: colorName(color)];
    return self;
}

-observeBiomassRates
{
    char* namebuf = [globalZone alloc: 50];

    bioRateStats = [BiomassRates create];

    bioRateGraph = [TraceGraph create];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "bioRateGraph");
    [bioRateGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Biomass Rates");
    [bioRateGraph setWinTitle: namebuf];
    [bioRateGraph trackDouble: &(bioRateStats->netGrowth)
                Key: "NetGrowth" Color: ColorNetGrowth];
    [bioRateGraph trackDouble: &(bioRateStats->growth)
                Key: "Growth" Color: ColorGrowth];
    [bioRateGraph trackDouble: &(bioRateStats->lossSteal)
                Key: "LossSteal" Color: ColorLossSteal];
    [bioRateGraph trackDouble: &(bioRateStats->lossStarve)
                Key: "LossStarve" Color: ColorLossStarve];
    [bioRateGraph trackDouble: &(bioRateStats->losses)
                Key: "Losses" Color: ColorLosses];
    [bioRateGraph trackDouble: &(bioRateStats->netRatio)
                Key: "NetRatio" Color: ColorNetRatio];
    return self;
}

-observeBiomassTransfersSilent
{
    id  index;
    Species* s;
    
    bioTransStats = [BiomassTransfers create];
    index = [members begin: globalZone];
    while ( (s = [index next] ) ) {
        if (![s getBioTransStats])
            [s observeBiomassTransfersSilent];
    }
    [index drop];

    return self;
}

-observeBiomassTransfers
{
    char* namebuf = [globalZone alloc: 50];

    if (!bioTransStats)
        [self observeBiomassTransfersSilent];

    bioTransGraph = [TraceGraph create];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "bioTransGraph");
    [bioTransGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Biomass Transfers");
    [bioTransGraph setWinTitle: namebuf];
    [bioTransGraph trackDouble: &(bioTransStats->netGrowth)
                Key: "NetGrowth" Color: ColorNetGrowth];
    [bioTransGraph trackDouble: &(bioTransStats->growth)
                Key: "Growth" Color: ColorGrowth];
    [bioTransGraph trackDouble: &(bioTransStats->lossSteal)
                Key: "LossSteal" Color: ColorLossSteal];
    [bioTransGraph trackDouble: &(bioTransStats->lossStarve)
                Key: "LossStarve" Color: ColorLossStarve];
    [bioTransGraph trackDouble: &(bioTransStats->losses)
                Key: "Losses" Color: ColorLosses];
    [bioTransGraph trackDouble: &(bioTransStats->netRatio)
                Key: "NetRatio" Color: ColorNetRatio];
    [bioTransGraph trackDouble: &(bioTransStats->biomass)
                Key: "Biomass" Color: colorName(color)];
    return self;
}

-observeBiomassTheft
{
    char* namebuf = [globalZone alloc: 50];

    if (!bioTransStats)
        [self observeBiomassTransfersSilent];

    bioTheftGraph = [TangleGraph create];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "bioTheftGraph");
    [bioTheftGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Biomass Theft");
    [bioTheftGraph setWinTitle: namebuf];
    [bioTheftGraph setXAxisLabel: "Biomass"];
    [bioTheftGraph setYAxisLabel: "Biomass Theft"];
    [bioTheftGraph trackDoubleX: &(bioTransStats->biomass)
                Y: &(bioTransStats->lossSteal)
                Key: "Theft" Color: colorName(color)];
    return self;
}

-observeBiomassGrowth
{
    char* namebuf = [globalZone alloc: 50];

    if (!bioTransStats)
        [self observeBiomassTransfersSilent];

    bioGrowthGraph = [TangleGraph create];
    sprintf(namebuf, "%s.%s.%s", [geckoControl getFileStem], 
            name, "bioGrowthGraph");
    [bioGrowthGraph logFile: namebuf];
    sprintf(namebuf, "%s%s", name, " Biomass Growth");
    [bioGrowthGraph setWinTitle: namebuf];
    [bioGrowthGraph setXAxisLabel: "Biomass"];
    [bioGrowthGraph setYAxisLabel: "Biomass Growth"];
    [bioGrowthGraph trackDoubleX: &(bioTransStats->biomass)
                Y: &(bioTransStats->growth)
                Key: "Growth" Color: colorName(color)];
    return self;
}

-onSeason 
{
    return self;
}

-allStatsReset
{
    return self;
}

-resetClear 
{
    [self allStatsReset];
    if (ageGraph) [ageGraph clear];
    if (sizeGraph) [sizeGraph clear];
    if (stageGraph) [stageGraph clear];
    if (specificGraph) [specificGraph clear];
    if (mortGraph) [mortGraph clear];
    if (fertGraph) [fertGraph clear];
    if (bioRateGraph) [bioRateGraph clear];
    if (bioTransGraph) [bioTransGraph clear];
    if (bioTheftGraph) [bioTheftGraph clear];
    if (bioGrowthGraph) [bioGrowthGraph clear];

    // for some reason, this already seems correct, but let's be careful....
    census    = 0;
    lastalive = 0;
    biomass   = 0;
    return self;
}

-(unsigned) getCensus { return census; }
-(double) getBiomass { return biomass; }
-(unsigned) getLastAlive {return lastalive;}
-setBiomass: (unsigned) b {biomass=b; return self;}
-addBiomass: (unsigned) b {biomass+=b; return self;}
-zeroBiomass {biomass=0; return self; }
-scaleBiomass {biomass /= bioscale; return self;}

-baseUpdate
{
    id  index;
    Species* s;
    census    = 0;
    lastalive = 0;
    biomass   = 0;
    index = [members begin: globalZone];
    while ( (s = [index next] ) ) {
        census  += [s getCensus];
        biomass += [s getBiomass];
    }
    [index drop];
    if (census > 0)
        lastalive = [geckoControl getCurrentTime];
    return self;
}

-xUpdate: (id) xStats graph: (TraceGraph*) xGraph GetMessage: (SEL) m
{
    int max;
    id  index;
    Species* s;
    id substages;

    [self baseUpdate];

    if (xStats == nil)
        return self;
    
    max = [xStats getMaxAge];
        
    [xStats clear];
    
    index = [members begin: globalZone];
    
    while ( (s = [index next] ) ) {
        substages = [s perform: m];
        [xStats addBucket: substages];
    }
    [index drop];
    
    [xStats calc];
     
    if (xGraph) {
        [xGraph step];
        [mortGraph step];
        [fertGraph step];
    } 
    return self;
}
-ageUpdate
{
    [self xUpdate: ageStats graph: ageGraph GetMessage: M(getAgeStats)];
    return self;
}
-sizeUpdate
{
    [self xUpdate: sizeStats graph: sizeGraph GetMessage: M(getSizeStats)];
    return self;
}
-stageUpdate
{
    [self xUpdate: stageStats graph: stageGraph GetMessage: M(getStageStats)];
    return self;
}

-specificUpdate
{
    int max;
    id  index;
    Species* s;
    id substages;

    [self baseUpdate];

    if (specificStats == nil)
        return self;
    
    max = [specificStats getMaxBucket];
        
    [specificStats clear];
    
    index = [members begin: globalZone];
    
    while ( (s = [index next] ) ) {
        substages = [s getSpecificStats];
        [specificStats addBucket: substages];
    }
    [[s getGrandma] calcSpecific: specificStats];
    
    [index drop];
    
    if (specificGraph) [specificGraph step];
    return self;
}

-biomassRateUpdate
{
    id  index;
    Species* s;
    BiomassRates* subrates;
    
    [self baseUpdate];

    [self biomassTransferUpdate];
        
    if (bioRateStats == nil) 
        return self;
    if (biomass == 0)
        return self;

    [bioRateStats clear];

    index = [members begin: globalZone];
    while ( (s = [index next] ) ) {
        subrates = [s getBioRateStats];
        bioRateStats->growth += subrates->growth * [s getBiomass];
        bioRateStats->lossSteal += subrates->lossSteal * [s getBiomass];
        bioRateStats->lossStarve += subrates->lossStarve * [s getBiomass];
    }
    [index drop];
    bioRateStats->growth     /= biomass;
    bioRateStats->lossSteal  /= biomass;
    bioRateStats->lossStarve /= biomass;

    bioRateStats->losses = bioRateStats->lossSteal + bioRateStats->lossStarve;
    bioRateStats->netGrowth = bioRateStats->growth - bioRateStats->losses;
    bioRateStats->netRatio = bioRateStats->netGrowth - bioRateStats->lastNetGrowth;   
    
    if (bioRateGraph) {
        [bioRateGraph step];
    }

    return self;
}

-biomassTransferUpdate
{
    id  index;
    Species* s;
    BiomassTransfers* subrates;
    
    [self baseUpdate];
    
    if (bioTransStats == nil) 
        return self;
    if (biomass == 0)
        return self;

    [bioTransStats clear];

    index = [members begin: globalZone];
    while ( (s = [index next] ) ) {
        subrates = [s getBioTransStats];
        bioTransStats->growth += subrates->growth;
        bioTransStats->lossSteal += subrates->lossSteal;
        bioTransStats->lossStarve += subrates->lossStarve;
        bioTransStats->biomass += subrates->biomass;
    }
    [index drop];

    bioTransStats->losses = bioTransStats->lossSteal + bioTransStats->lossStarve;
    bioTransStats->netGrowth = bioTransStats->growth - bioTransStats->losses;
    bioTransStats->netRatio = bioTransStats->netGrowth - bioTransStats->lastNetGrowth;   
    
    if (bioTransGraph) {
        [bioTransGraph step];
    }
    if (bioGrowthGraph) {
        [bioGrowthGraph step];
    }
    if (bioTheftGraph) {
        [bioTheftGraph step];
    }

    return self;
}

-allSpeciesUpdate
{
    [self baseUpdate];
    return self;
}

// no-ops
-birthdayCheck {return self;}
-deathdayCheck {return self;}
-(unsigned) getBirthday { return 0;}

// these are for use by specific stats....
-(id) getStageStats {return stageStats;}
-(id) getAgeStats {return ageStats;}
-(id) getSizeStats {return sizeStats;}
-(id) getSpecificStats {return specificStats;}
-(BiomassRates*) getBioRateStats {return bioRateStats;}
-(BiomassTransfers*) getBioTranStats {return bioTransStats;}
-(BiomassTransfers*) getBioTransStats {return bioTransStats;}
-(double) getBioscale {return bioscale;}
-(TraceGraph*) getMortGraph { return mortGraph;}
-(TraceGraph*) getFertGraph { return fertGraph;}
-setSpecificGraph:  (TraceGraph*) sGraph {specificGraph = sGraph; return self;}
-setSpecificStats:  (id) specific
{specificStats = specific; return self;}

-print
{
    printLog("%s", name); return self;    
}

-println
{
    printLog("Composite %s: census=%d  biomass=", name, census); 
    printLogD("%lf\n", biomass); 
    return self;    
}

-printv
{
    [self println];
    return self;
}

-printCensus 
    // no eoln.
{
    printLog("%s:%d ", name, census);
    return self;
}

-dump
    // reloadable
{
    id  index;
    Species* s;
    
    printBoth("@begin CompositeSpecies %s\n", name);
    index = [members begin: globalZone];
    while ( (s = [index next] ) ) {
        printBoth("\tMember\t\t%s\n", [s getParamName]);
    }
    [params forEach: M(dump)];
    
    if (popFlag)
        printBoth("\tobservePop\n");
    if (bioFlag)
        printBoth("\tobserveBiomass\n");
    if (ageGraph) {
        printBoth("\tAgesBuckets\t%d\n", [ageStats getMaxAge]);
        printBoth("\tobserveAges\n");
    }
    if (stageGraph)
        printBoth("\tobserveStages\n");
    if (sizeGraph)
        printBoth("\tobserveSizes\n");
    if (specificGraph)
        printBoth("\tobserveSpecific\n");
    if (bioRateGraph)
        printBoth("\tobserveBiomassRates\n");
    if (bioTransGraph)
        printBoth("\tobserveBiomassTransfers\n");
    
    [index drop];
    printBoth("@end CompositeSpecies %s\n\n", name);
    return self;
}

@end

